Package: fastglm
Type: Package
Title: Fast and Stable Fitting of Generalized Linear Models using
        'RcppEigen'
Version: 0.1.1
Authors@R: c(
    person("Jared", "Huling", , email = "jaredhuling@gmail.com", c("aut", "cre")),
    person("Douglas", "Bates", , , c("cph")),
    person("Dirk", "Eddelbuettel", , , c("cph")),
    person("Romain", "Francois", , , c("cph")),
    person("Yixuan", "Qiu", , , c("cph")),
    person("Noah", "Greifer", role=c("ctb"), 
           email = "noah.greifer@gmail.com",
	       comment=c(ORCID="0000-0003-3067-7154"))
    )
Description: Fits generalized linear models efficiently using 'RcppEigen'. The iteratively reweighted least squares 
    implementation utilizes the step-halving approach of Marschner (2011) <doi:10.32614/RJ-2011-012> to help safeguard
    against convergence issues.
BugReports: https://github.com/jaredhuling/fastglm/issues
URL: https://github.com/jaredhuling/fastglm,
        https://jaredhuling.org/fastglm/
License: GPL (>= 2)
Encoding: UTF-8
Depends: R (>= 4.0.0)
Imports: Rcpp (>= 0.12.13), bigmemory, methods, Matrix, Formula
LinkingTo: Rcpp, RcppEigen, BH, bigmemory
Suggests: glm2, knitr, rmarkdown, testthat (>= 3.0.0), sandwich,
        microbenchmark, MASS, statmod, logistf, brglm2, pscl, speedglm,
        tweedie, ggplot2
Config/testthat/edition: 3
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2026-06-05 13:55:11 UTC; huling
Author: Jared Huling [aut, cre],
  Douglas Bates [cph],
  Dirk Eddelbuettel [cph],
  Romain Francois [cph],
  Yixuan Qiu [cph],
  Noah Greifer [ctb] (ORCID: <https://orcid.org/0000-0003-3067-7154>)
Maintainer: Jared Huling <jaredhuling@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-07 14:40:07 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-07 15:00:52 UTC; unix
Archs: fastglm.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,fastglm_hurdle)
S3method(coef,fastglm_zi)
S3method(deviance,fastglm)
S3method(family,fastglm)
S3method(fastglm,default)
S3method(logLik,fastglm)
S3method(logLik,fastglm_hurdle)
S3method(logLik,fastglm_nb)
S3method(logLik,fastglm_zi)
S3method(predict,fastglm)
S3method(print,fastglm)
S3method(print,fastglm_hurdle)
S3method(print,fastglm_zi)
S3method(residuals,fastglm)
S3method(sandwich::bread,fastglm)
S3method(sandwich::bread,fastglmFit)
S3method(sandwich::estfun,fastglm)
S3method(sandwich::estfun,fastglmFit)
S3method(sandwich::vcovHC,fastglm)
S3method(sandwich::vcovHC,fastglmFit)
S3method(stats::vcov,fastglm)
S3method(stats::vcov,fastglmFit)
S3method(summary,fastglm)
S3method(summary,fastglmFit)
S3method(vcov,fastglm_hurdle)
S3method(vcov,fastglm_zi)
export(fastglm)
export(fastglmPure)
export(fastglm_control)
export(fastglm_fit)
export(fastglm_hurdle)
export(fastglm_nb)
export(fastglm_streaming)
export(fastglm_zi)
export(negbin)
import(methods)
import(stats)
importClassesFrom(Matrix,dgCMatrix)
importFrom(Rcpp,evalCpp)
importFrom(bigmemory,as.big.matrix)
importFrom(bigmemory,is.big.matrix)
importFrom(methods,new)
useDynLib(fastglm, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

[![version](http://www.r-pkg.org/badges/version/fastglm)](https://cran.r-project.org/package=fastglm)
[![downloads](http://cranlogs.r-pkg.org/badges/grand-total/fastglm)](https://cran.r-project.org/package=fastglm)

# fastglm

The *fastglm* package is a **fast** and **stable** alternative to
`stats::glm()` for fitting generalized linear models. It is built on
*RcppEigen* and is fully compatible with R’s `family` objects: the
downstream methods you expect (`summary()`, `vcov()`, `predict()`,
`coef()`, `residuals()`, `logLik()`) all work exactly as they do for a
`glm`.

Beyond standard GLMs, *fastglm* provides dedicated fitting functions for
negative-binomial regression, hurdle and zero-inflated count models, and
Firth bias-reduced GLMs, all of which reuse the same C++ IRLS solver.

### Features

- Six decomposition methods for the IRLS weighted least-squares step:
  column-pivoted QR (default, rank-revealing), unpivoted QR, LLT
  Cholesky, LDLT Cholesky, full-pivoted QR, and bidiagonal
  divide-and-conquer SVD.
- Robust convergence via step-halving safeguard following Marschner
  (2011), with better-initialized starting values than `glm()` or
  `glm2()`.
- Large scale data features: Sparse design matrices
  (`Matrix::dgCMatrix`), on-disc `big.matrix` objects (`bigmemory`), and
  a streaming callback interface (`fastglm_streaming()`) for fitting on
  data that does not fit in memory.
- Firth bias-reduced GLMs (`firth = TRUE`) implementing the AS_mean
  adjustment of Kosmidis and Firth (2009, 2021) for all standard
  families, all six decomposition methods, and all three large-data
  paths.
- Negative-binomial regression (`fastglm_nb()`) with joint
  `(beta, theta)` MLE entirely in C++.
- Hurdle count models (`fastglm_hurdle()`) with Poisson or NB count
  components and a binary zero/non-zero component.
- Zero-inflated count models (`fastglm_zi()`) with Poisson or NB count
  distributions, fit by an EM algorithm in C++.
- Inference: `vcov()`, `summary()`, `predict(se.fit = TRUE)`, and
  compatibility with `sandwich::vcovHC()` and `sandwich::vcovCL()` for
  robust standard errors.

## Installation

Install from CRAN:

``` r
install.packages("fastglm")
```

or the development version from GitHub:

``` r
pak::pak("jaredhuling/fastglm")
```

## Fitting a GLM

The main function is `fastglm()`. It takes a numeric design matrix `x`,
a response `y`, and an R `family` object:

``` r
library(fastglm)

data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp), data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

fit <- fastglm(x, y, family = binomial(link = "cloglog"))
summary(fit)
#> 
#> Call:
#> fastglm.default(x = x, y = y, family = binomial(link = "cloglog"))
#> 
#> Coefficients:
#>                Estimate Std. Error z value Pr(>|z|)    
#> (Intercept)    -4.29199    0.29777 -14.414  < 2e-16 ***
#> agegp.L         3.30677    0.63454   5.211 1.88e-07 ***
#> agegp.Q        -1.35525    0.57310  -2.365    0.018 *  
#> agegp.C         0.20296    0.43333   0.468    0.640    
#> agegp^4         0.15056    0.29318   0.514    0.608    
#> agegp^5        -0.23425    0.17855  -1.312    0.190    
#> unclass(tobgp)  0.33276    0.07285   4.568 4.93e-06 ***
#> unclass(alcgp)  0.80384    0.07538  10.664  < 2e-16 ***
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> (Dispersion parameter for binomial family taken to be 1)
#> 
#>     Null deviance: 61.609  on 87  degrees of freedom
#> Residual deviance: 96.950  on 80  degrees of freedom
#> AIC: 228
#> 
#> Number of Fisher Scoring iterations: 6
```

`fastglm()` operates on a pre-built design matrix. To use a formula and
a data frame, pass `fastglm_fit` as the fitting function to base
`glm()`:

``` r
fit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) + unclass(alcgp),
            data = esoph, family = binomial(link = "cloglog"),
            method = fastglm_fit)
```

A third, minimal-use function, `fastglmPure()`, returns only the
coefficient vector and working quantities, skipping dispersion, AIC, and
null-deviance computation. Use this when calling *fastglm* from another
package and you only need the coefficients.

## Decomposition methods

The IRLS algorithm reduces every iteration to a weighted least-squares
problem. *fastglm* supports six different matrix decompositions for
solving that WLS step, all from *RcppEigen* (Bates and Eddelbuettel,
2013); the choice trades off speed against numerical stability and
rank-revealing behavior:

| `method` | decomposition                                           |
|----------|---------------------------------------------------------|
| `0`      | column-pivoted Householder QR (default; rank-revealing) |
| `1`      | unpivoted Householder QR                                |
| `2`      | LLT Cholesky                                            |
| `3`      | LDLT Cholesky                                           |
| `4`      | full-pivoted Householder QR                             |
| `5`      | bidiagonal divide-and-conquer SVD                       |

The default (`method = 0`) is the safe choice: it is rank-revealing, so
it handles aliased or collinear columns gracefully. The Cholesky methods
(`2` and `3`) are roughly 3–4x faster but assume full column rank.

``` r
set.seed(123)
n <- 5000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rbinom(n, 1, plogis(x %*% rnorm(p) * 0.05))

system.time(f0 <- fastglm(x, y, family = binomial()))                 # default QR
#>    user  system elapsed 
#>   0.004   0.000   0.005
system.time(f2 <- fastglm(x, y, family = binomial(), method = 2))     # LLT
#>    user  system elapsed 
#>   0.002   0.000   0.002
```

## Speed

*fastglm* runs the same IRLS algorithm as `glm.fit()` but executes the
per-iteration WLS solve in C++ via *RcppEigen*, which is often
substantially faster than the compiled-R + LAPACK path that `glm.fit()`
uses. The gap widens with sample size because the R-side overhead in
`glm.fit()` is fixed per iteration:

``` r
library(microbenchmark)
library(ggplot2)

set.seed(123)
n.obs  <- 10000
n.vars <- 100

x <- matrix(rnorm(n.obs * n.vars, sd = 3), n.obs, n.vars)
y <- 1 * (drop(x[, 1:25] %*% runif(25, -0.1, 0.1)) > rnorm(n.obs))

ct <- microbenchmark(
    glm.fit          = glm.fit(x, y, family = binomial()),
    fastglm_QR       = fastglm(x, y, family = binomial(), method = 0),
    fastglm_LLT      = fastglm(x, y, family = binomial(), method = 2),
    fastglm_LDLT     = fastglm(x, y, family = binomial(), method = 3),
    times = 25L
)

autoplot(ct, log = FALSE) +
    ggplot2::stat_summary(fun = median, geom = 'point', size = 2) +
    ggplot2::theme_bw()
```

<img src="man/figures/README-glm_bench-1.png" alt="" width="100%" style="display: block; margin: auto;" />

Coefficient estimates agree with `glm.fit()` to floating-point
precision:

``` r
gl <- glm.fit(x, y, family = binomial())
c(fastglm_QR   = max(abs(coef(gl) - coef(fastglm(x, y, family = binomial(), method = 0)))),
  fastglm_LLT  = max(abs(coef(gl) - coef(fastglm(x, y, family = binomial(), method = 2)))),
  fastglm_LDLT = max(abs(coef(gl) - coef(fastglm(x, y, family = binomial(), method = 3)))))
#>   fastglm_QR  fastglm_LLT fastglm_LDLT 
#> 1.360023e-15 1.249001e-15 1.165734e-15
```

## Stability

*fastglm* does not compromise computational stability for speed. It uses
a step-halving safeguard following Marschner (2011) and starts from
better-initialized values than `glm()` or `glm2::glm2()`, so it tends to
converge in cases where the standard IRLS algorithm fails. As an
example, consider a Gamma model with a `sqrt` link — a mild response
misspecification combined with a badly misspecified link. In such
scenarios the standard IRLS algorithm tends to have convergence issues:

``` r
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- (exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000))) + 0.1

gfit1 <- glm(y ~ x, family = Gamma(link = "sqrt"), method = fastglm_fit)
gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt"))

## fastglm converges with a higher likelihood
c(fastglm_converged = gfit1$converged, glm_converged = gfit2$converged)
#> fastglm_converged     glm_converged 
#>              TRUE             FALSE
c(fastglm_logLik = logLik(gfit1), glm_logLik = logLik(gfit2))
#> fastglm_logLik     glm_logLik 
#>      -16046.66      -16704.05
```

See `vignette("fastglm", package = "fastglm")` for the full comparison,
including `glm2::glm2()` and `speedglm`.

## Native C++ families

For the most commonly used `family`/`link` combinations, *fastglm*
dispatches `variance()`, `mu.eta()`, `linkinv()`, and `dev.resids()` to
inline C++ implementations rather than calling back into R once per IRLS
iteration. The covered combinations are:

- gaussian (identity, log, inverse)
- binomial (logit, probit, cloglog, log)
- poisson (log, identity, sqrt)
- Gamma (log, inverse, identity)
- inverse.gaussian (1/mu^2, log, identity, inverse)

Detection is automatic: if the `family` object matches one of the above,
the native path is used; otherwise *fastglm* falls back to the
R-callback path. The C++ native approach is meaningfully faster on large
`n` because it eliminates the per-iteration calls to R for each of the
four family functions.

## Sparse, big.matrix, and streaming designs

For designs that are sparse, that live on disc, or that have to be built
from a parquet / *arrow* / *DuckDB* source, *fastglm* provides three
large-data paths that share a common streaming kernel and produce
identical results:

- **`Matrix::dgCMatrix`**: pass directly to `fastglm()`. Useful for
  one-hot encoded categoricals and high-dimensional sparse designs.
- **`bigmemory::big.matrix`**: pass directly to `fastglm()`. The matrix
  is read in row-blocks and never fully materialized in memory.
- **`fastglm_streaming(chunk_callback, n_chunks, family)`**: a
  user-supplied closure yields one row-block per call. The right path
  for fitting on a parquet dataset, *DuckDB* query, or any external
  columnar store.

A short example of the streaming computation approach:

``` r
n <- 4000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.4, -0.2, 0.3)))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))
#> [1] 1.042751e-07
```

See `vignette("large-data-fastglm", package = "fastglm")` for a detailed
walk-through of all three paths.

## Extended models

### Negative-binomial regression

`fastglm_nb()` fits negative-binomial regression with the dispersion
`theta` estimated jointly with the regression coefficients, in the
spirit of `MASS::glm.nb()`. The joint `(beta, theta)` MLE runs entirely
in C++; IRLS for `beta`, Brent’s method for `theta`:

``` r
set.seed(123)
n <- 5000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y <- MASS::rnegbin(n, mu = mu, theta = 2)

f_nb <- fastglm_nb(X, y)
c(coef = coef(f_nb), theta = f_nb$theta)
#>    coef.x1    coef.x2    coef.x3    coef.x4      theta 
#>  0.4926199  0.3862258 -0.2200450  0.2991296  1.9923309
```

### Hurdle models

`fastglm_hurdle()` fits a two-part count model: a binary regression for
whether `y > 0`, plus a zero-truncated Poisson or NB regression on the
positive subset. The two parts factorize and both are fit by the same
C++ IRLS solver. This is the same model as `pscl::hurdle()` (Zeileis,
Kleiber, and Jackman, 2008). Different designs for the count and zero
parts are specified via the `Formula` package’s two-RHS syntax:

``` r
set.seed(123)
n <- 5000
x1 <- rnorm(n);  x2 <- rnorm(n)
lam <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
yt <- integer(n)
for (i in seq_len(n)) {
    repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
}
y <- ifelse(is_pos == 1, yt, 0L)

f_h <- fastglm_hurdle(y ~ x1 + x2, data = data.frame(y, x1, x2), dist = "poisson")
coef(f_h)
#> count_(Intercept)          count_x1          count_x2  zero_(Intercept) 
#>         0.6793788         0.4162547        -0.3137229        -0.3767285 
#>           zero_x1           zero_x2 
#>         0.4447325         0.1999711
```

### Zero-inflated models

`fastglm_zi()` fits a zero-inflated Poisson or NB regression, a binary
inflation component overlaid on the original count distribution, fit by
an EM algorithm in C++ with closed-form posterior responsibilities and
an analytical observed-information `vcov`. This is the same model as
`pscl::zeroinfl()`:

``` r
set.seed(123)
n <- 5000
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z <- rbinom(n, 1, plogis(eta_z))
y <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))

f_zi <- fastglm_zi(y ~ x1 + x2, data = data.frame(y, x1, x2), dist = "poisson")
coef(f_zi)
#> count_(Intercept)          count_x1          count_x2  zero_(Intercept) 
#>         0.6647396         0.3965153        -0.3407168        -0.3750258 
#>           zero_x1           zero_x2 
#>         0.4512788         0.1819040
```

### Firth bias-reduced GLMs

Setting `firth = TRUE` activates the general mean-bias reduction of
Kosmidis and Firth (2009, 2021). This extends Firth’s (1993) original
logistic penalty to arbitrary GLM families, producing finite estimates
even under separation and removing the leading $O(1/n)$ bias from
maximum likelihood estimates:

``` r
data(sex2, package = "logistf")
X <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y <- sex2$case

f_firth <- fastglm(X, y, family = binomial(), firth = TRUE)
coef(f_firth)
#> (Intercept)         age          oc         vic        vicl         vis 
#>  0.12025405 -1.10598133 -0.06881673  2.26887465 -2.11140819 -0.78831695 
#>         dia 
#>  3.09601183
```

Firth bias reduction works with all six decomposition methods, all
standard R families and link functions, and all three large-data paths
(sparse, big.matrix, streaming). See
`vignette("firth-fastglm", package = "fastglm")` for verification
against `logistf::logistf()` and `brglm2::brglmFit()`.

## Inference

The fitted object stores the unscaled covariance directly, so `vcov()`
and `summary()` work as expected. Heteroskedasticity-consistent and
cluster-robust covariance matrices are available via
`sandwich::vcovHC()` and `sandwich::vcovCL()`, *fastglm* registers
methods on those generics, so loading *sandwich* is all that is
required:

``` r
library(sandwich)
V_hc <- vcovHC(fit, type = "HC0")
V_cl <- vcovCL(fit, cluster = cluster, type = "HC1")
```

Results are numerically identical to `sandwich` applied to a `glm` fit
to floating-point precision. `predict()` supports `se.fit = TRUE`:

``` r
predict(fit, newdata = xnew, type = "response", se.fit = TRUE)
```

## Benchmarks

A comprehensive benchmarking study is available in
`vignette("benchmarks-fastglm", package = "fastglm")`, comparing
*fastglm* against the canonical reference implementations across
standard GLMs (`glm.fit`, `glm2`, `speedglm`), negative-binomial
regression (`MASS::glm.nb`), Firth bias-reduced GLMs (`brglm2`,
`logistf`), and hurdle / zero-inflated count regressions
(`pscl::hurdle`, `pscl::zeroinfl`).

The following summary plot shows the speedup *fastglm* delivers over the
canonical reference for each model class, as a function of sample size.
The reference for the standard GLMs is the fastest among `glm.fit`,
`glm2`, and `speedglm`, so the comparison is conservative. Larger is
better:

<figure>
<img src="vignettes/benchmarks-figs/unnamed-chunk-18-1.png"
alt="Speedup of fastglm vs canonical reference implementations across model classes" />
<figcaption aria-hidden="true">Speedup of fastglm vs canonical reference
implementations across model classes</figcaption>
</figure>

Across all model classes the same picture holds: *fastglm* matches the
canonical reference implementation to floating-point precision, and the
runtime gap grows with sample size. By $n = 10^5$ the speedup is
generally an order of magnitude or more. For models with an outer
iteration (NB joint MLE, hurdle/ZI with NB), the gap is widest, since
the entire outer loop is in C++ in *fastglm* and entirely in R in the
reference implementations.

## References

- Firth, D. (1993). Bias reduction of maximum likelihood estimates.
  *Biometrika*, 80(1), 27–38.

- Kosmidis, I. and Firth, D. (2009). Bias reduction in exponential
  family nonlinear models. *Biometrika*, 96(4), 793–804.

- Kosmidis, I. and Firth, D. (2021). Jeffreys-prior penalty, finiteness
  and shrinkage in binomial-response generalized linear models.
  *Biometrika*, 108(1), 71–82.

- Marschner, I. C. (2011). glm2: Fitting generalized linear models with
  convergence problems. *The R Journal*, 3(2), 12–15.

- Bates, D. and Eddelbuettel, D. (2013). Fast and elegant numerical
  linear algebra using the RcppEigen package. *Journal of Statistical
  Software*, 52(5), 1–24.

- Zeileis, A., Kleiber, C., and Jackman, S. (2008). Regression models
  for count data in R. *Journal of Statistical Software*, 27(8), 1–25.

<!-- README.md is generated from README.Rmd. Please edit that file -->

[![version](http://www.r-pkg.org/badges/version/fastglm)](https://cran.r-project.org/package=fastglm)

# Overview of *fastglm*

The *fastglm* package is a re-write of `glm()` using *RcppEigen*
designed to be computationally efficient and algorithmically stable.

# Installing the *fastglm* package

*fastglm* can be installed using *pak*:

``` r
pak::pak("fastglm")
# pak::pak("jaredhuling/fastglm") #development version
```

or by cloning and building using `R CMD INSTALL`.

# Quick Usage Overview

Load the package:

``` r
library(fastglm)
```

A (not comprehensive) comparison with `glm.fit()` and `speedglm.wfit()`:

``` r
library(speedglm)
library(microbenchmark)
library(ggplot2)

set.seed(123)
n.obs  <- 10000
n.vars <- 100
x <- matrix(rnorm(n.obs * n.vars, sd = 3), n.obs, n.vars)
Sigma <- 0.99 ^ abs(outer(1:n.vars, 1:n.vars, FUN = "-"))
x <- MASS::mvrnorm(n.obs, mu = runif(n.vars, min = -1), Sigma = Sigma)

y <- 1 * ( drop(x[,1:25] %*% runif(25, min = -0.1, max = 0.10)) > rnorm(n.obs))

ct <- microbenchmark(
    glm.fit =         {gl1 <- glm.fit(x, y, family = binomial())},
    speedglm.eigen  = {sg1 <- speedglm.wfit(y, x, intercept = FALSE,
                                            family = binomial())},
    speedglm.chol   = {sg2 <- speedglm.wfit(y, x, intercept = FALSE, 
                                            family = binomial(), method = "Chol")},
    speedglm.qr     = {sg3 <- speedglm.wfit(y, x, intercept = FALSE,
                                            family = binomial(), method = "qr")},
    fastglm.qr.cpiv = {gf1 <- fastglm(x, y, family = binomial())},
    fastglm.qr      = {gf2 <- fastglm(x, y, family = binomial(), method = 1)},
    fastglm.LLT     = {gf3 <- fastglm(x, y, family = binomial(), method = 2)},
    fastglm.LDLT    = {gf4 <- fastglm(x, y, family = binomial(), method = 3)},
    fastglm.qr.fpiv = {gf5 <- fastglm(x, y, family = binomial(), method = 4)},
    times = 25L
)

autoplot(ct, log = FALSE) +
    stat_summary(fun.y = median, geom = 'point', size = 2)
```

<img src="man/figures/README-gen_data-1.png" width="100%" style="display: block; margin: auto;" />

``` r
# comparison of estimates
c(glm_vs_fastglm_qrcpiv = max(abs(coef(gl1) - gf1$coef)),
  glm_vs_fastglm_qr     = max(abs(coef(gl1) - gf2$coef)),
  glm_vs_fastglm_qrfpiv = max(abs(coef(gl1) - gf5$coef)),
  glm_vs_fastglm_LLT    = max(abs(coef(gl1) - gf3$coef)),
  glm_vs_fastglm_LDLT   = max(abs(coef(gl1) - gf4$coef)))
```

    ## glm_vs_fastglm_qrcpiv     glm_vs_fastglm_qr glm_vs_fastglm_qrfpiv 
    ##          8.881784e-15          9.936496e-15          1.204592e-14 
    ##    glm_vs_fastglm_LLT   glm_vs_fastglm_LDLT 
    ##          1.202649e-13          2.076672e-13

``` r
# now between glm and speedglm
c(glm_vs_speedglm_eigen = max(abs(coef(gl1) - sg1$coef)),
  glm_vs_speedglm_Chol  = max(abs(coef(gl1) - sg2$coef)),
  glm_vs_speedglm_qr    = max(abs(coef(gl1) - sg3$coef)))
```

    ## glm_vs_speedglm_eigen  glm_vs_speedglm_Chol    glm_vs_speedglm_qr 
    ##          1.259048e-12          1.259048e-12          1.320777e-12

# Stability

The *fastglm* package does not compromise computational stability for
speed. In fact, for many situations where `glm()` and even
`glm2::glm2()` do not converge, `fastglm()` does converge.

As an example, consider the following data scenario, where the response
distribution is (mildly) misspecified, but the link function is quite
badly misspecified. In such scenarios, the standard IRLS algorithm tends
to have convergence issues. The `glm2()` package was designed to handle
such cases, however, it still can have convergence issues. The
`fastglm()` package uses a similar step-halving technique as `glm2()`,
but it starts at better initialized values and thus tends to have better
convergence properties in practice.

``` r
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- (exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000)) ) + 0.1


system.time(gfit1 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         method = fastglm_fit))
```

    ##    user  system elapsed 
    ##   0.261   0.005   0.267

``` r
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt")))
```

    ##    user  system elapsed 
    ##   0.387   0.025   0.412

``` r
system.time(gfit3 <- glm2::glm2(y ~ x, family = Gamma(link = "sqrt")))
```

    ##    user  system elapsed 
    ##   0.267   0.019   0.288

``` r
system.time(gfit4 <- speedglm(y ~ x, family = Gamma(link = "sqrt")))
```

    ##    user  system elapsed 
    ##   0.067   0.012   0.079

``` r
## speedglm appears to diverge
system.time(gfit5 <- speedglm(y ~ x, family = Gamma(link = "sqrt"),
                              maxit = 500))
```

    ##    user  system elapsed 
    ##   1.215   0.195   1.411

``` r
## Note that fastglm() returns estimates with the
## largest likelihood

c(fastglm     = logLik(gfit1),
  glm         = logLik(gfit2),
  glm2        = logLik(gfit3),
  speedglm    = logLik(gfit4),
  speedglm500 = logLik(gfit5))
```

    ##     fastglm         glm        glm2    speedglm speedglm500 
    ##   -16030.81   -16704.05   -16046.66   -36423.35   -70953.78

``` r
rbind(fastglm     = coef(gfit1)[1:5],
      glm         = coef(gfit2)[1:5],
      glm2        = coef(gfit3)[1:5],
      speedglm    = coef(gfit4)[1:5],
      speedglm500 = coef(gfit5)[1:5])
```

    ##             (Intercept)        x1            x2         x3         x4
    ## fastglm      1.42925957 0.1258714  5.311209e-03 -0.1293960  0.2389508
    ## glm          1.43116799 0.1251936 -6.896739e-05 -0.1281857  0.2366473
    ## glm2         1.42686381 0.1242616 -9.860241e-05 -0.1254873  0.2361301
    ## speedglm    -4.56190855 5.0654671 -4.844225e+00 -0.9470766 -5.8183704
    ## speedglm500 -0.00159227 1.3962876  8.377002e+00 -1.2283816  0.1281007

``` r
## check convergence of fastglm and #iterations
# 1 means converged, 0 means not converged
c(gfit1$converged, gfit1$iter)
```

    ## [1]  1 15

``` r
## now check convergence for glm()
c(gfit2$converged, gfit2$iter)
```

    ## [1]  0 25

``` r
## check convergence for glm2()
c(gfit3$converged, gfit3$iter)
```

    ## [1]  1 19

``` r
## check convergence for speedglm()
c(gfit4$convergence, gfit4$iter, gfit5$convergence, gfit5$iter)
```

    ## [1]   0  25   0 500

``` r
## increasing number of IRLS iterations for glm() does not help that much
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         maxit = 1000))
```

    ##    user  system elapsed 
    ##  10.559   0.966  11.538

``` r
gfit2$converged
```

    ## [1] FALSE

``` r
gfit2$iter
```

    ## [1] 1000

``` r
logLik(gfit1)
```

    ## 'log Lik.' -16030.81 (df=102)

``` r
logLik(gfit2)
```

    ## 'log Lik.' -26232.17 (df=102)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE, echo=TRUE--------------------------------------------------
# devtools::install_github("jaredhuling/fastglm")

## ----echo=TRUE----------------------------------------------------------------
library(fastglm)

## ----echo = TRUE--------------------------------------------------------------
data(esoph)
x <- model.matrix(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp)
                                         + unclass(alcgp), data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

gfit1 <- fastglm(x = x, y = y, family = binomial(link = "cloglog"))

summary(gfit1)

## ----fig.show='hold'----------------------------------------------------------
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- (exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000)) ) + 0.1


system.time(gfit1 <- fastglm(cbind(1, x), y, family = Gamma(link = "sqrt")))

system.time(gfit2 <- glm(y~x, family = Gamma(link = "sqrt")) )

system.time(gfit3 <- glm2::glm2(y~x, family = Gamma(link = "sqrt")) )

## Note that fastglm() returns estimates with the
## largest likelihood
logLik(gfit1)
logLik(gfit2)
logLik(gfit3)

coef(gfit1)[1:5]
coef(gfit2)[1:5]
coef(gfit3)[1:5]

## check convergence of fastglm
gfit1$converged
## number of IRLS iterations
gfit1$iter

## now check convergence for glm()
gfit2$converged
gfit2$iter

## check convergence for glm2()
gfit3$converged
gfit3$iter


## increasing number of IRLS iterations for glm() does not help that much
system.time(gfit2 <- glm(y~x, family = Gamma(link = "sqrt"), control = list(maxit = 100)) )

gfit2$converged
gfit2$iter

logLik(gfit1)
logLik(gfit2)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)

## -----------------------------------------------------------------------------
library(fastglm)
set.seed(1)

## -----------------------------------------------------------------------------
library(Matrix)
n <- 5000

## a 50-level factor that interacts with a continuous covariate.
g    <- factor(sample.int(50, n, replace = TRUE))
xnum <- rnorm(n)
Xd   <- model.matrix(~ g * xnum)        # 5000 x 100, mostly zeros
Xs   <- as(Xd, "CsparseMatrix")

object.size(Xd)
object.size(Xs)

beta_true <- rnorm(ncol(Xd)) / 4
y <- rbinom(n, 1, plogis(Xd %*% beta_true))

## fit using the dense and sparse representations
fit_dense  <- fastglm(Xd, y, family = binomial(), method = 2)
fit_sparse <- fastglm(Xs, y, family = binomial(), method = 2)

max(abs(coef(fit_sparse) - coef(fit_dense)))

## ----error = TRUE-------------------------------------------------------------
try({
fastglm(Xs, y, family = binomial(), method = 0)
})

## -----------------------------------------------------------------------------
library(bigmemory)
n <- 20000
p <- 20
X  <- matrix(rnorm(n * p), n, p)
Xb <- as.big.matrix(X)               # in-memory big.matrix; could equally be filebacked

y  <- rbinom(n, 1, plogis(X %*% rnorm(p) * 0.05))

fit_dense <- fastglm(X,  y, family = binomial(), method = 2)
fit_big   <- fastglm(Xb, y, family = binomial(), method = 2)

max(abs(coef(fit_big) - coef(fit_dense)))

## -----------------------------------------------------------------------------
Sys.setenv(FASTGLM_CHUNK_ROWS = "2000")
fit_big_small <- fastglm(Xb, y, family = binomial(), method = 2)
Sys.unsetenv("FASTGLM_CHUNK_ROWS")

max(abs(coef(fit_big_small) - coef(fit_big)))

## ----error = TRUE-------------------------------------------------------------
try({
fastglm(Xb, y, family = binomial(), method = 0)
})

## -----------------------------------------------------------------------------
n <- 10000
p <- 5
X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.5, -0.3, 0.4, -0.2)))

## simulate a "data source" that yields the design matrix in 5 row-blocks
chunk_size <- 2000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 5, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))

## -----------------------------------------------------------------------------
n <- 8000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
ofs <- runif(n, -0.1, 0.1)
pw  <- runif(n,  0.5, 1.5)
y   <- rpois(n, exp(X %*% c(0.2, 0.4, -0.2, 0.3) + ofs))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx],
         offset = ofs[idx], weights = pw[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 8, family = poisson())
fit_full   <- fastglm(X, y, family = poisson(), offset = ofs, weights = pw, method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))

## ----eval = FALSE-------------------------------------------------------------
# library(arrow)
# ds      <- open_dataset("path/to/data.parquet")
# batches <- as.list(Scanner$create(ds)$ScanBatches())
# 
# chunks <- function(k) {
#     tbl <- as.data.frame(batches[[k]])
#     list(X = model.matrix(~ x1 + x2 + x3, data = tbl),
#          y = tbl$y)
# }
# 
# fit <- fastglm_streaming(chunks,
#                          n_chunks = length(batches),
#                          family   = binomial())

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)

## ----eval = FALSE-------------------------------------------------------------
# fastglm(x, y, family = binomial(), firth = TRUE)

## -----------------------------------------------------------------------------
library(fastglm)

## -----------------------------------------------------------------------------
data(sex2, package = "logistf")
X <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y <- sex2$case

fit_f <- fastglm(X, y, family = binomial(), firth = TRUE)
fit_l <- logistf::logistf(case ~ age + oc + vic + vicl + vis + dia,
                          data = sex2, pl = FALSE, plconf = NULL,
                          control = logistf::logistf.control(
                              lconv = 1e-12, gconv = 1e-12, xconv = 1e-12,
                              maxit = 200L))

cbind(fastglm = unname(coef(fit_f)), logistf = unname(coef(fit_l)))
max(abs(unname(coef(fit_f)) - unname(coef(fit_l))))

## -----------------------------------------------------------------------------
mb <- microbenchmark::microbenchmark(
    fastglm = fastglm(X, y, family = binomial(), firth = TRUE),
    logistf = logistf::logistf(case ~ age + oc + vic + vicl + vis + dia,
                               data = sex2, pl = FALSE, plconf = NULL),
    times = 10L
)
print(mb, unit = "ms")

## -----------------------------------------------------------------------------
library(brglm2)
set.seed(123)
n <- 500
x1 <- rnorm(n);  x2 <- rnorm(n)
X <- cbind(1, x1, x2)
eta_true <- 0.5 + 0.3 * x1 - 0.2 * x2

## -----------------------------------------------------------------------------
y_bin <- rbinom(n, 1, plogis(eta_true))
df <- data.frame(y = y_bin, x1 = x1, x2 = x2)

for (lnk in c("logit", "probit", "cloglog")) {
    fam <- binomial(lnk)
    f <- fastglm(X, y_bin, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = df, family = fam, method = "brglmFit",
             type = "AS_mean", control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("binomial %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}

## -----------------------------------------------------------------------------
y_pois <- rpois(n, exp(eta_true))

for (lnk in c("log", "sqrt")) {
    fam <- poisson(lnk)
    f <- fastglm(X, y_pois, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y_pois, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("poisson  %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}

## -----------------------------------------------------------------------------
y_gam <- rgamma(n, shape = 5, rate = 5 / exp(eta_true))

for (lnk in c("log", "inverse")) {
    fam <- Gamma(lnk)
    f <- fastglm(X, y_gam, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y_gam, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("Gamma    %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}

## -----------------------------------------------------------------------------
y_gauss <- eta_true + rnorm(n, 0, 0.5)

for (lnk in c("identity", "log")) {
    fam <- gaussian(lnk)
    y0 <- if (lnk == "log") pmax(y_gauss, 0.01) else y_gauss
    f <- fastglm(X, y0, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y0, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("gaussian %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}

## -----------------------------------------------------------------------------
mu_ig <- exp(eta_true)
y_ig <- statmod::rinvgauss(n, mean = mu_ig, shape = 5)
y_ig[y_ig <= 0] <- 0.01

f <- fastglm(X, y_ig, family = inverse.gaussian("log"), firth = TRUE, tol = 1e-10)
b <- glm(y ~ x1 + x2, data = data.frame(y = y_ig, x1 = x1, x2 = x2),
         family = inverse.gaussian("log"), method = "brglmFit", type = "AS_mean",
         control = list(epsilon = 1e-10, maxit = 200))
cat(sprintf("inv.gauss log      max|diff| = %.2e\n",
            max(abs(unname(coef(f)) - unname(coef(b))))))

## -----------------------------------------------------------------------------
f <- fastglm(X, y_bin, family = binomial(), firth = TRUE, tol = 1e-10)
b <- glm(y ~ x1 + x2, data = df, family = binomial(), method = "brglmFit",
         type = "AS_mean", control = list(epsilon = 1e-10, maxit = 200))

cbind(fastglm_SE = f$se, brglm2_SE = summary(b)$coefficients[, "Std. Error"])
max(abs(f$se - summary(b)$coefficients[, "Std. Error"]))

## -----------------------------------------------------------------------------
f <- fastglm(X, y_bin, family = binomial(), firth = TRUE)
c(deviance = f$deviance,
  log.det.XtWX = f$log.det.XtWX,
  penalized.deviance = f$penalized.deviance)

## -----------------------------------------------------------------------------
set.seed(123)
n <- 10000;  p <- 5
x1 <- rnorm(n);  x2 <- rnorm(n); x3 <- rnorm(n)
x4 <- rnorm(n);  x5 <- rnorm(n)
X <- cbind(1, x1, x2, x3, x4, x5)
eta <- 0.5 + 0.3 * x1 - 0.2 * x2 + 0.1 * x3

families <- list(
    `binomial logit`   = list(fam = binomial("logit"),
                              y = rbinom(n, 1, plogis(eta))),
    `binomial probit`  = list(fam = binomial("probit"),
                              y = rbinom(n, 1, plogis(eta))),
    `binomial cloglog` = list(fam = binomial("cloglog"),
                              y = rbinom(n, 1, plogis(eta))),
    `poisson log`      = list(fam = poisson("log"),
                              y = rpois(n, exp(eta))),
    `Gamma log`        = list(fam = Gamma("log"),
                              y = rgamma(n, shape = 5, rate = 5 / exp(eta))),
    `gaussian identity`= list(fam = gaussian("identity"),
                              y = eta + rnorm(n, 0, 0.5))
)

df <- data.frame(y = 0, x1 = x1, x2 = x2, x3 = x3, x4 = x4, x5 = x5)
results <- list()
for (nm in names(families)) {
    fi <- families[[nm]]
    df$y <- fi$y
    mb <- microbenchmark::microbenchmark(
        fastglm = fastglm(X, fi$y, family = fi$fam, firth = TRUE),
        brglm2  = glm(y ~ x1 + x2 + x3 + x4 + x5, data = df,
                      family = fi$fam, method = "brglmFit", type = "AS_mean"),
        times = 10L
    )
    s <- summary(mb, unit = "ms")
    cat(sprintf("%-20s fastglm=%7.2f ms  brglm2=%7.2f ms  speedup=%.1fx\n",
                nm, s$median[1], s$median[2], s$median[2] / s$median[1]))
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)

## -----------------------------------------------------------------------------
library(fastglm)

## -----------------------------------------------------------------------------
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

fit <- fastglm(x, y, family = binomial(link = "cloglog"))
summary(fit)

## -----------------------------------------------------------------------------
fit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) + unclass(alcgp),
            data = esoph, family = binomial(link = "cloglog"),
            method = fastglm_fit)

## -----------------------------------------------------------------------------
set.seed(123)
n <- 5000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rbinom(n, 1, plogis(x %*% rnorm(p) * 0.05))

system.time(f0 <- fastglm(x, y, family = binomial()))                 # default
system.time(f2 <- fastglm(x, y, family = binomial(), method = 2))     # LLT
system.time(f3 <- fastglm(x, y, family = binomial(), method = 3))     # LDLT

## -----------------------------------------------------------------------------
fit <- fastglm(x, y, family = binomial(), method = 2)

V    <- vcov(fit)
dim(V)

## standard errors agree with the diagonal of vcov()
all.equal(sqrt(diag(V)), unname(coef(summary(fit))[, "Std. Error"]))

## -----------------------------------------------------------------------------
library(sandwich)

V_hc <- vcovHC(fit, type = "HC0")
V_hc[1:3, 1:3]

cluster <- rep(seq_len(20), length.out = n)
V_cl    <- vcovCL(fit, cluster = cluster, type = "HC1")
V_cl[1:3, 1:3]

## -----------------------------------------------------------------------------
xnew <- x[1:5, , drop = FALSE]
predict(fit, newdata = xnew, type = "response", se.fit = TRUE)

## -----------------------------------------------------------------------------
n <- 4000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.4, -0.2, 0.3)))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))

## ----eval = TRUE--------------------------------------------------------------
library(microbenchmark)

set.seed(7)
n <- 50000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rpois(n, exp(x %*% rnorm(p) * 0.05 + 1))

## force the R-callback path by giving the family object an unrecognized name
disguise <- function(fam) { fam$family <- paste0(fam$family, "*"); fam }

mb <- microbenchmark(
    native   = fastglm(x, y, family = poisson(),            method = 2),
    callback = fastglm(x, y, family = disguise(poisson()),  method = 2),
    times = 5L
)
print(mb)

skip_if_not_installed("statmod")
skip_if_not_installed("tweedie")

sim_tweedie <- function(n, var.power, link.power, seed = 31) {
    set.seed(seed)
    p <- 3
    X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    beta <- c(1, 0.5, -0.3)
    eta  <- as.vector(X %*% beta)
    mu   <- if (link.power == 0) exp(eta)
            else if (link.power == 1) eta
            else eta^(1 / link.power)
    y    <- tweedie::rtweedie(n, mu = mu, phi = 1.2, power = var.power)
    list(X = X, y = y, beta = beta)
}

test_that("native Tweedie log-link matches glm() across var.power values", {
    for (vp in c(1.3, 1.5, 1.7, 1.9)) {
        d   <- sim_tweedie(500, var.power = vp, link.power = 0, seed = 31 + 10 * vp)
        fam <- statmod::tweedie(var.power = vp, link.power = 0)
        gfit <- glm(d$y ~ d$X[, -1], family = fam)
        ffit <- fastglm(d$X, d$y,    family = fam, method = 2)
        expect_equal(unname(coef(ffit)), unname(coef(gfit)),
                     tolerance = 1e-7,
                     info = paste0("Tweedie log, var.power = ", vp))
        expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-7,
                     info = paste0("Tweedie log dev, var.power = ", vp))
        expect_equal(ffit$dispersion, summary(gfit)$dispersion,
                     tolerance = 1e-6)
    }
})

test_that("native Tweedie sqrt link matches glm()", {
    d   <- sim_tweedie(500, var.power = 1.5, link.power = 0.5, seed = 41)
    fam <- statmod::tweedie(var.power = 1.5, link.power = 0.5)
    gfit <- glm(d$y ~ d$X[, -1], family = fam,
                start = c(sqrt(mean(d$y)), 0, 0))
    ffit <- fastglm(d$X, d$y, family = fam, method = 2,
                    start = c(sqrt(mean(d$y)), 0, 0))
    expect_equal(unname(coef(ffit)), unname(coef(gfit)), tolerance = 1e-6)
    expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-6)
})

test_that("native Tweedie p=2 (Gamma limit) matches Gamma-log", {
    set.seed(47)
    n  <- 400
    X  <- cbind(1, matrix(rnorm(n * 2), n, 2))
    mu <- exp(X %*% c(0.5, 0.3, -0.2))
    y  <- rgamma(n, shape = 2, rate = 2 / mu)
    fam_tw <- statmod::tweedie(var.power = 2, link.power = 0)
    fam_g  <- Gamma(link = "log")
    f_tw   <- fastglm(X, y, family = fam_tw, method = 2)
    f_g    <- fastglm(X, y, family = fam_g,  method = 2)
    expect_equal(unname(coef(f_tw)), unname(coef(f_g)), tolerance = 1e-6)
})

test_that("Tweedie native path matches R-callback fallback", {
    disguise <- function(fam) { fam$family <- paste0(fam$family, "_dis"); fam }
    d   <- sim_tweedie(400, var.power = 1.5, link.power = 0, seed = 53)
    fam <- statmod::tweedie(var.power = 1.5, link.power = 0)
    f_native <- fastglm(d$X, d$y, family = fam,           method = 2)
    f_cb     <- fastglm(d$X, d$y, family = disguise(fam), method = 2)
    expect_equal(unname(coef(f_native)), unname(coef(f_cb)), tolerance = 1e-7)
    expect_equal(f_native$deviance, f_cb$deviance, tolerance = 1e-7)
})

#' Fit a negative-binomial GLM with simultaneous (beta, theta) MLE
#'
#' `fastglm_nb()` fits a negative-binomial regression model, jointly
#' maximising the likelihood over the regression coefficients `beta` and the
#' NB2 dispersion `theta`. It is the fastglm analogue of [MASS::glm.nb()],
#' built on top of the native NB family kernel introduced in 0.0.6 so that
#' all numerical loops -- IRLS, the inner theta MLE Brent root-find, and
#' the outer (beta, theta) alternation -- run entirely in C++.
#'
#' @param x design matrix (numeric matrix, dgCMatrix not yet supported here).
#' @param y non-negative integer response vector.
#' @param weights optional prior weights vector of length `length(y)`.
#' @param offset optional offset vector of length `length(y)`.
#' @param start optional starting values for `beta`.
#' @param init.theta optional starting value for `theta`. If `NULL`, uses
#'   the method-of-moments estimator from a Poisson pilot fit.
#' @param link character, one of `"log"` (default), `"sqrt"`, `"identity"`.
#' @param method integer; `0..5`, see [fastglm()].
#' @param tol convergence tolerance for the IRLS inner loop.
#' @param maxit maximum number of inner-loop IRLS iterations.
#' @param outer.maxit maximum number of `(beta, theta)` outer iterations.
#' @param outer.tol convergence tolerance for the outer loop on the
#'   sup-norm of the beta update plus the relative change in `theta`.
#' @param theta.tol Brent tolerance for the inner theta MLE.
#' @param theta.maxit max iterations for the inner theta MLE.
#'
#' @returns A list of class `c("fastglm_nb", "fastglm")` with the usual
#'   fastglm components plus `theta`, `SE.theta`, `iter.theta`, and
#'   `twologlik` (twice the maximised NB log-likelihood).
#'
#' @examples
#' set.seed(1)
#' n <- 500
#' x <- cbind(1, matrix(rnorm(n * 2), n, 2))
#' eta <- x %*% c(0.3, 0.5, -0.2)
#' mu  <- exp(eta)
#' if (requireNamespace("MASS", quietly = TRUE)) {
#'   y   <- MASS::rnegbin(n, mu = mu, theta = 2)
#'   fit <- fastglm_nb(x, y)
#'   c(theta = fit$theta, fit$coefficients)
#' }
#'
#' @export
fastglm_nb <- function(x, y,
                       weights     = NULL,
                       offset      = NULL,
                       start       = NULL,
                       init.theta  = NULL,
                       link        = c("log", "sqrt", "identity"),
                       method      = 2L,
                       tol         = 1e-8,
                       maxit       = 100L,
                       outer.maxit = 25L,
                       outer.tol   = 1e-7,
                       theta.tol   = 1e-8,
                       theta.maxit = 100L)
{
    link <- match.arg(link)
    if (!is.matrix(x)) stop("'x' must be a numeric matrix.", call. = FALSE)
    if (!is.numeric(y) || any(y < 0))
        stop("'y' must be non-negative.", call. = FALSE)
    n <- length(y)
    if (nrow(x) != n)
        stop("nrow(x) and length(y) disagree.", call. = FALSE)

    if (is.null(weights)) weights <- rep(1, n)
    if (is.null(offset))  offset  <- rep(0, n)
    weights <- as.numeric(weights); offset <- as.numeric(offset)

    # Pilot Poisson fit to seed mu / eta and (if needed) theta.
    pilot <- fastglmPure(x, y,
                         family  = poisson(link = link),
                         weights = weights,
                         offset  = offset,
                         start   = start,
                         method  = as.integer(method),
                         tol     = tol,
                         maxit   = maxit)
    mu_init  <- as.numeric(pilot$fitted.values)
    eta_init <- as.numeric(pilot$linear.predictors)
    beta_init <- if (!is.null(start)) as.numeric(start) else as.numeric(pilot$coefficients)

    # Theta init: caller-supplied, or MoM (signalled to C++ via theta <= 0).
    init_th <- if (!is.null(init.theta)) {
        if (!is.numeric(init.theta) || length(init.theta) != 1L ||
            !is.finite(init.theta) || init.theta <= 0)
            stop("'init.theta' must be a positive finite scalar.", call. = FALSE)
        init.theta
    } else {
        -1
    }

    # The C++ driver uses the native NB kernels for {log,sqrt,identity}, but
    # still requires an R family object for the callback fallback signature.
    fam_dummy <- if (requireNamespace("MASS", quietly = TRUE)) {
        MASS::negative.binomial(theta = max(init_th, 1), link = link)
    } else {
        negbin(theta = max(init_th, 1), link = link)
    }

    fc <- switch(link,
                 "log"      = 17L,
                 "sqrt"     = 18L,
                 "identity" = 19L)

    res <- fit_glm_nb(
        x       = x,
        y       = as.numeric(y),
        weights = weights,
        offset  = offset,
        start   = beta_init,
        mu_init = mu_init,
        eta_init = eta_init,
        var_fun         = fam_dummy$variance,
        mu_eta_fun      = fam_dummy$mu.eta,
        linkinv_fun     = fam_dummy$linkinv,
        dev_resids_fun  = fam_dummy$dev.resids,
        valideta_fun    = fam_dummy$valideta,
        validmu_fun     = fam_dummy$validmu,
        type        = as.integer(method),
        tol         = as.double(tol),
        maxit       = as.integer(maxit),
        fam_code    = fc,
        init_theta  = as.double(init_th),
        theta_tol   = as.double(theta.tol),
        theta_maxit = as.integer(theta.maxit),
        outer_maxit = as.integer(outer.maxit),
        outer_tol   = as.double(outer.tol)
    )

    # Cosmetic post-processing
    cnames <- colnames(x)
    if (is.null(cnames)) cnames <- paste0("x", seq_len(ncol(x)))
    names(res$coefficients) <- cnames
    names(res$se)           <- cnames
    rownames(res$cov.unscaled) <- colnames(res$cov.unscaled) <- cnames

    final_family <- if (requireNamespace("MASS", quietly = TRUE)) {
        MASS::negative.binomial(theta = res$theta, link = link)
    } else {
        negbin(theta = res$theta, link = link)
    }
    res$family        <- final_family
    res$y             <- as.numeric(y)
    res$x             <- x
    res$prior.weights <- weights
    res$dispersion    <- 1                       # NB likelihood is fully specified
    res$df.null       <- n - 1L
    res$null.deviance <- sum(final_family$dev.resids(
        as.numeric(y),
        rep(if (sum(weights) > 0) sum(weights * as.numeric(y))/sum(weights) else mean(y), n),
        weights))
    res$call <- match.call()
    class(res) <- c("fastglm_nb", "fastglm")
    res
}

#' @export
logLik.fastglm_nb <- function(object, ...)
{
    df <- length(object$coefficients) + 1L          # +1 for theta
    structure(object$twologlik / 2, df = df,
              nobs = length(object$y), class = "logLik")
}

skip_if_not_installed("MASS")

# Match MASS::glm.nb(y ~ x[,-1]) on simulated NB data.

test_that("fastglm_nb matches MASS::glm.nb across realistic theta", {
    for (th_true in c(0.5, 2.0)) {
        set.seed(2024 + th_true * 10)
        n  <- 500
        X  <- cbind(1, matrix(rnorm(n * 2), n, 2))
        mu <- exp(X %*% c(0.3, 0.5, -0.2))
        y  <- MASS::rnegbin(n, mu = mu, theta = th_true)

        m  <- MASS::glm.nb(y ~ X[, -1])
        f  <- fastglm_nb(X, y)

        expect_equal(unname(coef(f)), unname(coef(m)),
                     tolerance = 1e-7,
                     info = paste0("theta_true = ", th_true))
        expect_equal(f$theta, m$theta, tolerance = 1e-6)
        expect_equal(f$SE.theta, m$SE.theta, tolerance = 1e-4)
        expect_equal(f$twologlik, m$twologlik, tolerance = 1e-6)
        # Coefficient SE should match too.
        expect_equal(unname(f$se), unname(sqrt(diag(vcov(m)))),
                     tolerance = 1e-6)
    }
})

test_that("fastglm_nb matches MASS::glm.nb on the quine dataset", {
    skip_if_not(requireNamespace("MASS", quietly = TRUE))
    quine <- tryCatch({
        utils::data("quine", package = "MASS", envir = environment())
        get("quine", envir = environment())
    }, error = function(e) NULL)
    skip_if(is.null(quine), "quine dataset unavailable")
    f <- fastglm_nb(model.matrix(~ Sex + Age, data = quine), quine$Days)
    m <- MASS::glm.nb(Days ~ Sex + Age, data = quine)
    expect_equal(unname(coef(f)), unname(coef(m)), tolerance = 1e-6)
    expect_equal(f$theta, m$theta, tolerance = 1e-5)
    expect_equal(f$twologlik, m$twologlik, tolerance = 1e-6)
})

test_that("init.theta is honored", {
    set.seed(2025)
    n  <- 300
    X  <- cbind(1, matrix(rnorm(n * 2), n, 2))
    y  <- MASS::rnegbin(n, mu = exp(X %*% c(0.3, 0.5, -0.2)), theta = 1.5)
    f1 <- fastglm_nb(X, y)
    f2 <- fastglm_nb(X, y, init.theta = 5)
    expect_equal(f1$theta, f2$theta, tolerance = 1e-6)
    expect_equal(unname(coef(f1)), unname(coef(f2)), tolerance = 1e-7)
})

test_that("fastglm_nb returns a fastglm-classed object with the expected slots", {
    set.seed(33)
    n  <- 200
    X  <- cbind(1, matrix(rnorm(n * 2), n, 2))
    y  <- MASS::rnegbin(n, mu = exp(X %*% c(0.3, 0.4, -0.2)), theta = 2)
    f  <- fastglm_nb(X, y)
    expect_s3_class(f, "fastglm_nb")
    expect_s3_class(f, "fastglm")
    expect_named(f$coefficients)
    expect_true(is.finite(f$theta))
    expect_true(is.finite(f$SE.theta))
    expect_true(f$converged)
    # vcov / SE work via the standard fastglm S3 dispatch.
    v <- vcov(f)
    expect_equal(dim(v), rep(ncol(X), 2))
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)

## -----------------------------------------------------------------------------
library(fastglm)
suppressPackageStartupMessages({
    library(MASS)        # glm.nb, rnegbin
    library(pscl)        # hurdle, zeroinfl, bioChemists
    library(microbenchmark)
})

## -----------------------------------------------------------------------------
data(quine)
X <- model.matrix(~ Eth + Sex + Age + Lrn, data = quine)
y <- quine$Days

fit_f <- fastglm_nb(X, y)
fit_m <- MASS::glm.nb(Days ~ Eth + Sex + Age + Lrn, data = quine)

c(theta_fastglm = fit_f$theta, theta_glm.nb = fit_m$theta)

max(abs(unname(coef(fit_f)) - unname(coef(fit_m))))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_m)))

## -----------------------------------------------------------------------------
set.seed(1)
n  <- 5e4
X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y  <- MASS::rnegbin(n, mu = mu, theta = 2)
df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])

mb_nb <- microbenchmark(
    fastglm_nb = fastglm_nb(X, y),
    glm.nb     = MASS::glm.nb(y ~ x1 + x2 + x3, data = df),
    times = 5L
)
print(mb_nb)

## -----------------------------------------------------------------------------
fit_joint <- fastglm_nb(X, y)
fit_known <- fastglm(X, y, family = negbin(theta = fit_joint$theta, link = "log"))
max(abs(unname(coef(fit_known)) - unname(coef(fit_joint))))

## -----------------------------------------------------------------------------
data(bioChemists, package = "pscl")
fit_f <- fastglm_hurdle(art ~ ., data = bioChemists, dist = "poisson")
fit_p <- pscl::hurdle (art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))

## -----------------------------------------------------------------------------
coef(fit_f, model = "count")
coef(fit_f, model = "zero")

## -----------------------------------------------------------------------------
set.seed(11)
n  <- 4000
x1 <- rnorm(n);  x2 <- rnorm(n)
lam    <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
yt     <- integer(n)
for (i in seq_len(n)) {
    repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
}
y  <- ifelse(is_pos == 1, yt, 0L)
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_hurdle <- microbenchmark(
    fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_hurdle)

## -----------------------------------------------------------------------------
fit_f <- fastglm_zi(art ~ ., data = bioChemists, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 300L)
fit_p <- pscl::zeroinfl(art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))

## -----------------------------------------------------------------------------
set.seed(21)
n  <- 3000
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z     <- rbinom(n, 1, plogis(eta_z))
y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
df    <- data.frame(y = y, x1 = x1, x2 = x2)

mb_zi <- microbenchmark(
    fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_zi)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  echo = TRUE,
  eval = TRUE
)

## ----eval = FALSE, echo=TRUE--------------------------------------------------
# pak::pak("fastglm")
# # pak::pak("jaredhuling/fastglm") #development version

## ----echo=TRUE----------------------------------------------------------------
library(fastglm)

## ----echo = TRUE--------------------------------------------------------------
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)

y <- cbind(esoph$ncases, esoph$ncontrols)

gfit1 <- fastglm(x = x, y = y, family = binomial(link = "cloglog"))

summary(gfit1)

## -----------------------------------------------------------------------------
gfit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) +
                 unclass(alcgp), data = esoph,
             family = binomial(link = "cloglog"),
             method = fastglm_fit)

summary(gfit2, refit = FALSE)

## ----fig.show='hold'----------------------------------------------------------
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000)) + 0.1


# fastglm
system.time(gfit1 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         method = fastglm_fit))

# glm
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt")))

# glm2
system.time(gfit3 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         method = glm2::glm.fit2))

## Note that fastglm() returns estimates with the
## largest likelihood
logLik(gfit1)
logLik(gfit2)
logLik(gfit3)

coef(gfit1)[1:5]
coef(gfit2)[1:5]
coef(gfit3)[1:5]

## check convergence of fastglm
gfit1$converged
## number of IRLS iterations
gfit1$iter

## now check convergence for glm()
gfit2$converged
gfit2$iter

## check convergence for glm2()
gfit3$converged
gfit3$iter


## increasing number of IRLS iterations for glm() does not help that much
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         control = list(maxit = 100)))

gfit2$converged
gfit2$iter

logLik(gfit1)
logLik(gfit2)

make_glm_data <- function(n = 500, p = 5, seed = 1L,
                          response = c("binomial", "gaussian", "poisson", "gamma_log"))
{
    response <- match.arg(response)
    set.seed(seed)
    X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    colnames(X) <- c("(Intercept)", paste0("x", seq_len(p - 1)))
    beta_true <- c(0.2, 0.5, -0.3, 0.4, -0.2)[seq_len(p)]
    eta <- as.vector(X %*% beta_true)
    y <- switch(response,
        binomial   = rbinom(n, 1, plogis(eta)),
        gaussian   = eta + rnorm(n, sd = 0.5),
        poisson    = rpois(n, exp(eta)),
        gamma_log  = rgamma(n, shape = 2, rate = 2 / exp(eta)))
    list(X = X, y = y, beta_true = beta_true, n = n, p = p)
}

family_for <- function(name) {
    switch(name,
        binomial  = binomial(),
        gaussian  = gaussian(),
        poisson   = poisson(),
        gamma_log = Gamma(link = "log"))
}

test_that("vcov(fastglm) matches vcov(glm) across all decomposition methods", {
    d <- make_glm_data(n = 500, p = 5, response = "binomial")
    gl <- glm.fit(d$X, d$y, family = binomial())

    for (m in 0:5) {
        gf <- fastglm(d$X, d$y, family = binomial(), method = m)
        expect_equal(unname(coef(gf)), unname(gl$coefficients), tolerance = 1e-6,
                     info = sprintf("method = %d coefficients", m))
        v_glm <- summary(glm(y ~ . - 1,
                             data = data.frame(d$X, y = d$y),
                             family = binomial()))$cov.scaled
        expect_equal(unname(vcov(gf)), unname(v_glm), tolerance = 1e-6,
                     info = sprintf("method = %d vcov", m))
    }
})

test_that("vcov(fastglm) for gaussian, poisson, gamma agrees with glm", {
    for (resp in c("gaussian", "poisson", "gamma_log")) {
        d <- make_glm_data(n = 500, p = 4, response = resp)
        fam <- family_for(resp)
        df <- data.frame(d$X, y = d$y)
        gl_full <- glm(y ~ . - 1, data = df, family = fam)
        gf <- fastglm(d$X, d$y, family = fam, method = 2)
        expect_equal(unname(coef(gf)), unname(coef(gl_full)), tolerance = 1e-6,
                     info = resp)
        expect_equal(unname(vcov(gf)), unname(vcov(gl_full)), tolerance = 1e-6,
                     info = resp)
    }
})

test_that("predict(se.fit = TRUE) matches predict.glm", {
    d <- make_glm_data(n = 300, p = 4, response = "binomial")
    df <- data.frame(d$X, y = d$y)
    gl <- glm(y ~ . - 1, data = df, family = binomial())
    gf <- fastglm(d$X, d$y, family = binomial(), method = 2)

    new_X <- d$X[1:50, , drop = FALSE]
    p_glm  <- predict(gl, newdata = data.frame(new_X), se.fit = TRUE)
    p_fast <- predict(gf, newdata = new_X, se.fit = TRUE)

    expect_equal(unname(p_fast$fit),    unname(p_glm$fit),    tolerance = 1e-6)
    expect_equal(unname(p_fast$se.fit), unname(p_glm$se.fit), tolerance = 1e-6)
})

test_that("vcov.fastglmFit no longer requires a refit", {
    d <- make_glm_data(n = 200, p = 4, response = "gaussian")
    df <- data.frame(d$X, y = d$y)
    gl <- glm(y ~ . - 1, data = df, family = gaussian(), method = fastglm_fit)
    expect_s3_class(gl, "fastglmFit")
    v <- vcov(gl)
    v_ref <- vcov(glm(y ~ . - 1, data = df, family = gaussian()))
    expect_equal(unname(v), unname(v_ref), tolerance = 1e-6)
})

test_that("fastglm_streaming with K chunks matches single-shot fastglm", {
    d <- make_glm_data(n = 1000, p = 5, response = "binomial")
    K <- 4
    chunk_size <- d$n / K
    chunks <- function(k) {
        idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
        list(X = d$X[idx, , drop = FALSE], y = d$y[idx])
    }
    f_stream <- fastglm_streaming(chunks, n_chunks = K, family = binomial(),
                                  tol = 1e-12)
    f_full   <- fastglm(d$X, d$y, family = binomial(), method = 2, tol = 1e-12)

    expect_equal(unname(coef(f_stream)), unname(coef(f_full)), tolerance = 1e-6)
    expect_lt(max(abs(f_stream$cov.unscaled - f_full$cov.unscaled)) /
              max(abs(f_full$cov.unscaled)), 1e-5)
})

test_that("streaming honors offset and prior weights", {
    set.seed(3)
    n <- 1000; K <- 5
    X <- cbind(1, matrix(rnorm(n * 3), n, 3))
    eta <- X %*% c(0.1, 0.4, -0.2, 0.3)
    ofs <- runif(n, -0.1, 0.1)
    pw  <- runif(n, 0.5, 1.5)
    yp <- rpois(n, exp(eta + ofs))

    chunk_size <- n / K
    chunks <- function(k) {
        idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
        list(X = X[idx, , drop = FALSE], y = yp[idx],
             offset = ofs[idx], weights = pw[idx])
    }
    f_stream <- fastglm_streaming(chunks, n_chunks = K, family = poisson())
    f_full   <- fastglm(X, yp, family = poisson(), offset = ofs, weights = pw,
                       method = 2)
    expect_equal(unname(coef(f_stream)), unname(coef(f_full)), tolerance = 1e-7)
})

test_that("streaming gaussian recovers exact OLS", {
    d <- make_glm_data(n = 800, p = 4, response = "gaussian")
    K <- 4; chunk_size <- d$n / K
    chunks <- function(k) {
        idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
        list(X = d$X[idx, , drop = FALSE], y = d$y[idx])
    }
    f_stream <- fastglm_streaming(chunks, n_chunks = K, family = gaussian())
    f_full   <- fastglm(d$X, d$y, family = gaussian(), method = 2)
    expect_equal(unname(coef(f_stream)), unname(coef(f_full)), tolerance = 1e-12)
})

test_that("streaming rejects bad chunk shapes", {
    chunks_bad <- function(k) {
        list(X = matrix(rnorm(20), 5, 4), y = rnorm(4))   # nrow != length(y)
    }
    expect_error(fastglm_streaming(chunks_bad, n_chunks = 1, family = gaussian()),
                 regexp = "rows")
})

test_that("sparse dgCMatrix path matches dense path on a binomial logit fit", {
    skip_if_not_installed("Matrix")
    d <- make_glm_data(n = 600, p = 5, response = "binomial")
    Xs <- methods::as(d$X, "CsparseMatrix")

    f_dense  <- fastglm(d$X, d$y, family = binomial(), method = 2)
    f_sparse <- fastglm(Xs,  d$y, family = binomial(), method = 2)

    expect_equal(unname(coef(f_sparse)), unname(coef(f_dense)), tolerance = 1e-10)
    expect_equal(f_sparse$deviance, f_dense$deviance, tolerance = 1e-10)
})

test_that("sparse rejects unsupported decomposition methods", {
    skip_if_not_installed("Matrix")
    d <- make_glm_data(n = 200, p = 4, response = "binomial")
    Xs <- methods::as(d$X, "CsparseMatrix")
    for (m in c(0, 1, 4, 5))
        expect_error(fastglm(Xs, d$y, family = binomial(), method = m),
                     regexp = "(LLT|LDLT|sparse)")
})

test_that("one-hot encoded categorical recovery: sparse matches dense", {
    skip_if_not_installed("Matrix")
    set.seed(2)
    n <- 500; k <- 8
    g <- sample.int(k, n, replace = TRUE)
    Xd <- model.matrix(~ factor(g))      # dense one-hot with intercept
    beta_true <- c(0.3, rep(c(0.6, -0.4), length.out = ncol(Xd) - 1))
    eta <- Xd %*% beta_true
    y <- rbinom(n, 1, plogis(eta))

    Xs <- methods::as(Xd, "CsparseMatrix")
    f_dense  <- fastglm(Xd, y, family = binomial(), method = 3)
    f_sparse <- fastglm(Xs, y, family = binomial(), method = 3)
    expect_equal(unname(coef(f_sparse)), unname(coef(f_dense)), tolerance = 1e-10)
})

test_that("vcovHC matches sandwich::vcovHC for binomial logit", {
    skip_if_not_installed("sandwich")
    d <- make_glm_data(n = 500, p = 5, response = "binomial")
    df <- data.frame(d$X, y = d$y)
    gl <- glm(y ~ . - 1, data = df, family = binomial(), control = list(epsilon = 1e-12))
    gf <- fastglm(d$X, d$y, family = binomial(), method = 2, tol = 1e-12)

    for (tp in c("HC0", "HC1", "HC2", "HC3")) {
        v_fast <- sandwich::vcovHC(gf, type = tp)
        v_sw   <- sandwich::vcovHC(gl, type = tp)
        expect_equal(unname(v_fast), unname(v_sw), tolerance = 1e-6,
                     info = paste0("binomial ", tp))
    }
})

test_that("vcovHC matches sandwich for gaussian and poisson", {
    skip_if_not_installed("sandwich")
    for (resp in c("gaussian", "poisson")) {
        d <- make_glm_data(n = 500, p = 4, response = resp)
        df <- data.frame(d$X, y = d$y)
        gl <- glm(y ~ . - 1, data = df, family = family_for(resp),
                  control = list(epsilon = 1e-12))
        gf <- fastglm(d$X, d$y, family = family_for(resp), method = 2, tol = 1e-12)
        for (tp in c("HC0", "HC2", "HC3")) {
            v_fast <- sandwich::vcovHC(gf, type = tp)
            v_sw   <- sandwich::vcovHC(gl, type = tp)
            expect_equal(unname(v_fast), unname(v_sw), tolerance = 1e-6,
                         info = paste0(resp, " ", tp))
        }
    }
})

test_that("vcovCL matches sandwich::vcovCL", {
    skip_if_not_installed("sandwich")
    d <- make_glm_data(n = 500, p = 4, response = "binomial")
    df <- data.frame(d$X, y = d$y)
    gl <- glm(y ~ . - 1, data = df, family = binomial(), control = list(epsilon = 1e-12))
    gf <- fastglm(d$X, d$y, family = binomial(), method = 2, tol = 1e-12)
    cluster <- rep(seq_len(25), length.out = d$n)

    v_fast <- sandwich::vcovCL(gf, cluster = cluster, type = "HC1")
    v_sw   <- sandwich::vcovCL(gl, cluster = cluster, type = "HC1")
    # Use absolute-error comparison: scale by overall matrix norm so a single
    # near-zero off-diagonal entry doesn't dominate the relative tolerance.
    expect_lt(max(abs(v_fast - v_sw)) / max(abs(v_sw)), 1e-5)
})

test_that("vcovHC works on sparse and big.matrix fits", {
    skip_if_not_installed("sandwich")
    skip_if_not_installed("Matrix")
    skip_if_not_installed("bigmemory")
    d <- make_glm_data(n = 400, p = 4, response = "binomial")
    df <- data.frame(d$X, y = d$y)
    gl <- glm(y ~ . - 1, data = df, family = binomial())

    f_sparse <- fastglm(methods::as(d$X, "CsparseMatrix"), d$y,
                        family = binomial(), method = 2)
    f_big    <- fastglm(bigmemory::as.big.matrix(d$X), d$y,
                        family = binomial(), method = 2)
    v_sw <- sandwich::vcovHC(gl, type = "HC0")
    expect_equal(unname(sandwich::vcovHC(f_sparse, type = "HC0")), unname(v_sw),
                 tolerance = 1e-6)
    expect_equal(unname(sandwich::vcovHC(f_big,    type = "HC0")), unname(v_sw),
                 tolerance = 1e-6)
})

# Force the R-callback fallback by wrapping a family so the (family, link)
# string no longer matches a native code.
disguise_family <- function(fam) {
    fam$family <- paste0(fam$family, "-disguised")
    fam
}

test_that("native and R-callback paths agree on tier-1 families", {
    cases <- list(
        list(name = "gaussian:identity", fam = gaussian(),                    resp = "gaussian"),
        list(name = "binomial:logit",    fam = binomial("logit"),             resp = "binomial"),
        list(name = "binomial:probit",   fam = binomial("probit"),            resp = "binomial"),
        list(name = "binomial:cloglog",  fam = binomial("cloglog"),           resp = "binomial"),
        list(name = "poisson:log",       fam = poisson("log"),                resp = "poisson"),
        list(name = "Gamma:log",         fam = Gamma("log"),                  resp = "gamma_log"),
        list(name = "Gamma:inverse",     fam = Gamma("inverse"),              resp = "gamma_log")
    )

    for (cc in cases) {
        d <- make_glm_data(n = 400, p = 4, response = cc$resp)
        f_native <- fastglm(d$X, d$y, family = cc$fam,                method = 2)
        f_callback <- fastglm(d$X, d$y, family = disguise_family(cc$fam), method = 2)
        expect_equal(unname(coef(f_native)), unname(coef(f_callback)),
                     tolerance = 1e-7,
                     info = paste0(cc$name, " coefficients"))
        expect_equal(f_native$deviance, f_callback$deviance, tolerance = 1e-8,
                     info = paste0(cc$name, " deviance"))
    }
})

test_that("big.matrix path matches dense path", {
    skip_if_not_installed("bigmemory")
    d <- make_glm_data(n = 800, p = 5, response = "binomial")
    Xb <- bigmemory::as.big.matrix(d$X)

    f_dense <- fastglm(d$X, d$y, family = binomial(), method = 2)
    f_big   <- fastglm(Xb,  d$y, family = binomial(), method = 2)

    expect_equal(unname(coef(f_big)), unname(coef(f_dense)), tolerance = 1e-10)
    expect_equal(f_big$deviance, f_dense$deviance, tolerance = 1e-10)
})

test_that("big.matrix rejects QR / SVD methods (would defeat streaming)", {
    skip_if_not_installed("bigmemory")
    d <- make_glm_data(n = 200, p = 4, response = "binomial")
    Xb <- bigmemory::as.big.matrix(d$X)
    for (m in c(0, 1, 4, 5))
        expect_error(fastglm(Xb, d$y, family = binomial(), method = m),
                     regexp = "LLT|LDLT|big.matrix")
})

test_that("FASTGLM_CHUNK_ROWS environment variable does not change results", {
    skip_if_not_installed("bigmemory")
    d <- make_glm_data(n = 1200, p = 4, response = "poisson")
    Xb <- bigmemory::as.big.matrix(d$X)
    f_default <- fastglm(Xb, d$y, family = poisson(), method = 2)

    old <- Sys.getenv("FASTGLM_CHUNK_ROWS", unset = NA)
    on.exit({
        if (is.na(old)) Sys.unsetenv("FASTGLM_CHUNK_ROWS")
        else Sys.setenv(FASTGLM_CHUNK_ROWS = old)
    })
    Sys.setenv(FASTGLM_CHUNK_ROWS = "100")
    f_small <- fastglm(Xb, d$y, family = poisson(), method = 2)
    expect_equal(unname(coef(f_default)), unname(coef(f_small)), tolerance = 1e-12)
})

.onLoad <- function(libname, pkgname)
{
    if (requireNamespace("sandwich", quietly = TRUE))
    {
        sw <- asNamespace("sandwich")
        registerS3method("vcovHC", "fastglm",    vcovHC.fastglm,    envir = sw)
        registerS3method("vcovHC", "fastglmFit", vcovHC.fastglmFit, envir = sw)
        registerS3method("estfun", "fastglm",    estfun.fastglm,    envir = sw)
        registerS3method("estfun", "fastglmFit", estfun.fastglmFit, envir = sw)
        registerS3method("bread",  "fastglm",    bread.fastglm,     envir = sw)
        registerS3method("bread",  "fastglmFit", bread.fastglmFit,  envir = sw)
    }
}

#' Negative binomial family with known dispersion
#'
#' A built-in negative-binomial family for use with [fastglm()] and
#' [fastglmPure()] when `theta` (the NB2 dispersion) is known. Equivalent to
#' `MASS::negative.binomial(theta, link)` but without taking a hard dependency
#' on the **MASS** package. The variance is `mu + mu^2 / theta`; as
#' `theta -> Inf` it reduces to Poisson.
#'
#' For joint estimation of `theta` together with `beta`, use the dedicated
#' (forthcoming) `fastglm_nb()` entry point. The `negbin()` family here is
#' intended for use cases where `theta` has been pre-specified or estimated
#' separately.
#'
#' @param theta the (known) dispersion parameter; must be positive.
#' @param link character; one of `"log"` (default), `"sqrt"`, or `"identity"`.
#'
#' @returns A `family` object with class `"family"` and
#'   `family$family == "Negative Binomial(theta)"`. The object also carries
#'   the slot `family$theta`, which `family_code()` and the dispatch layer
#'   use to detect the native NB fast path.
#'
#' @examples
#' set.seed(1)
#' n <- 500
#' x  <- cbind(1, matrix(rnorm(n * 2), n, 2))
#' mu <- exp(x %*% c(0.3, 0.4, -0.2))
#' y  <- rpois(n, lambda = mu * rgamma(n, shape = 2, rate = 2))   # NB(theta=2)
#'
#' fit <- fastglm(x, y, family = negbin(theta = 2, link = "log"))
#' coef(fit)
#'
#' @export
negbin <- function(theta, link = "log") {
    if (!is.numeric(theta) || length(theta) != 1L || !is.finite(theta) || theta <= 0)
        stop("'theta' must be a positive finite scalar.", call. = FALSE)
    link <- match.arg(link, c("log", "sqrt", "identity"))

    # Reuse MASS's negative.binomial machinery if available — its
    # initialize/aic/dev.resids closures are battle-tested.  Otherwise build
    # the family from scratch.  Either way we attach $theta so the dispatch
    # layer in family_code() / family_params() picks up the NB2 fast path.
    if (requireNamespace("MASS", quietly = TRUE)) {
        fam <- MASS::negative.binomial(theta = theta, link = link)
        fam$theta <- theta
        return(fam)
    }

    # ---- Fallback implementation (no MASS) ---------------------------------
    stats <- make.link(link)

    variance   <- function(mu)  mu + mu^2 / theta
    validmu    <- function(mu)  all(is.finite(mu)) && all(mu > 0)
    dev.resids <- function(y, mu, wt) {
        # 2 * wt * [ y * log(y/mu) - (y + theta) * log((y + theta)/(mu + theta)) ]
        a  <- ifelse(y > 0, y * log(y / mu), 0)
        yt <- y + theta
        mt <- mu + theta
        b  <- yt * log(yt / mt)
        2 * wt * (a - b)
    }
    aic <- function(y, n, mu, wt, dev) {
        # -2 * sum( wt * (lgamma(y + theta) - lgamma(theta) - lgamma(y + 1) +
        #                 theta*log(theta/(theta+mu)) + y*log(mu/(theta+mu))) )
        term1 <- lgamma(y + theta) - lgamma(theta) - lgamma(y + 1)
        term2 <- theta * log(theta / (theta + mu))
        term3 <- ifelse(y > 0, y * log(mu / (theta + mu)), 0)
        -2 * sum(wt * (term1 + term2 + term3))
    }
    initialize <- expression({
        if (any(y < 0)) stop("negative values not allowed for the negative binomial family")
        n       <- rep(1, nobs)
        mustart <- y + (y == 0) / 6
    })
    simfun <- function(object, nsim) {
        ftd <- fitted(object)
        rnbinom(nsim * length(ftd), mu = ftd, size = theta)
    }

    famname <- paste0("Negative Binomial(", format(theta, digits = 4), ")")

    structure(
        list(family     = famname,
             link       = link,
             linkfun    = stats$linkfun,
             linkinv    = stats$linkinv,
             variance   = variance,
             dev.resids = dev.resids,
             aic        = aic,
             mu.eta     = stats$mu.eta,
             initialize = initialize,
             validmu    = validmu,
             valideta   = stats$valideta,
             simulate   = simfun,
             theta      = theta),
        class = "family"
    )
}

#' Fast generalized linear model fitting
#'
#' `fastglm_fit()` is a fitting method for [glm()]. It works like `glm.fit()`, i.e., by being supplied to the `method` argument of `glm()`.
#' 
#' @param x a design matrix of dimension `n * p`. Can also be a `big.matrix` object from \pkg{bigmemory}.
#' @param y a vector of observations of length `n`.
#' @param weights an optional vector of 'prior weights' to be used in the fitting process. Should be `NULL` or a numeric vector.
#' @param start optional starting values for the parameters in the linear predictor.
#' @param etastart optional starting values for the linear predictor.
#' @param mustart optional starting values for the vector of means.
#' @param offset this can be used to specify an *a priori* known component to be included in the linear predictor during fitting. This should be `NULL` or a numeric vector of length equal to the number of cases.
#' @param family a description of the error distribution and link function to be used in the model. This must be a family function or the result of a call to a family function. (See [`family`] for details of family functions.)
#' @param control a list of parameters for controlling the fitting process. This is passed to `fastglm_control()`.
#' @param singular.ok,intercept See [glm.fit()].
#' @param fastmethod `integer`; the method used for fitting. Allowable values include 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition, 2 for the LLT Cholesky, 3 for the LDLT Cholesky, 4 for the full pivoted QR decomposition, and 5 for the Bidiagonal Divide and Conquer SVD. Default is 0. Can also be supplied as `method` when not supplied directly as an argument from `glm()` (see Examples).
#' @param tol `numeric`; threshold tolerance for convergence.
#' @param maxit `integer`; the maximum number of IRLS iterations.
#' @param firth `logical`; if `TRUE` apply Firth's (1993) bias-reducing
#'   penalty to the score function. Currently supported only for
#'   `family = binomial(link = "logit")` on dense `x`. See `logistf::logistf()`
#'   for the canonical reference implementation.
#' @param object a `fastglmFit` object; the output of a call to `glm()` with `method = fastglm_fit`.
#' @param \dots for `vcov()` and `summary()`, other arguments passed downstream.
#'
#' @details
#' The purpose of the functions documented on this page is to facilitate integration with existing [glm()] utilities in base R. `fastglm_fit()` is just a wrapper for [fastglmPure()] with some additional quality-of-life features. The `vcov()` and `summary()` methods use the unscaled coefficient covariance matrix returned directly from the C++ solver, so no refit is required.
#' 
#' @examples
#' set.seed(1234)
#' n <- 1e4
#' x <- matrix(rnorm(n * 25), ncol = 25)
#' eta <- 0.1 + 0.25 * x[,1] - 0.25 * x[,3] + 0.75 * x[,5] -0.35 * x[,6]
#' dat <- as.data.frame(x)
#' 
#' # binomial
#' dat$y <- rbinom(n, 1, pnorm(eta))
#' 
#' system.time({
#'     gl <- glm(y ~ ., data = dat,
#'               family = binomial)
#' })
#' 
#' system.time({
#'     gf0 <- glm(y ~ ., data = dat,
#'                family = binomial,
#'                method = fastglm_fit)
#' })
#' 
#' system.time({
#'     gf1 <- glm(y ~ ., data = dat,
#'                family = binomial,
#'                method = fastglm_fit,
#'                fastmethod = 1)
#' })
#' 
#' # poisson
#' dat$y <- rpois(n, eta^2)
#' 
#' system.time({
#'     gl <- glm(y ~ ., data = dat,
#'               family = poisson)
#' })
#' 
#' system.time({
#'     gf0 <- glm(y ~ ., data = dat,
#'                family = poisson,
#'                method = fastglm_fit)
#' })
#' 
#' system.time({
#'     gf1 <- glm(y ~ ., data = dat,
#'                family = poisson,
#'                method = fastglm_fit,
#'                fastmethod = 1)
#' })
#' 
#' # gamma
#' dat$y <- rgamma(n, exp(eta) * 1.75, 1.75)
#' 
#' system.time({
#'     gl <- glm(y ~ ., data = dat,
#'               family = Gamma(link = "log"))
#' })
#' 
#' system.time({
#'     gf0 <- glm(y ~ ., data = dat,
#'                family = Gamma(link = "log"),
#'                method = fastglm_fit)
#' })
#' 
#' system.time({
#'     gf1 <- glm(y ~ ., data = dat,
#'                family = Gamma(link = "log"),
#'                method = fastglm_fit,
#'                fastmethod = 1)
#' })
#' 
#' # Different (equivalent) ways of supplying
#' # control arguments:
#' gf1 <- glm(y ~ ., data = dat,
#'            family = Gamma(link = "log"),
#'            method = fastglm_fit,
#'            fastmethod = 1)
#' 
#' gf1 <- glm(y ~ ., data = dat,
#'            family = Gamma(link = "log"),
#'            method = fastglm_fit,
#'            control = list(fastmethod = 1))
#' 
#' gf1 <- glm(y ~ ., data = dat,
#'            family = Gamma(link = "log"),
#'            method = fastglm_fit,
#'            control = list(method = 1))

#' @export
fastglm_fit <- function(x, y,
                        weights  = rep(1, NROW(y)),
                        start    = NULL,
                        etastart = NULL,
                        mustart  = NULL,
                        offset   = rep(0, NROW(y)),
                        family   = gaussian(),
                        control  = list(),
                        intercept = TRUE,
                        singular.ok = TRUE,
                        firth = FALSE)
{
    control <- do.call("fastglm_control", control)
    if (!is.logical(firth) || length(firth) != 1L || is.na(firth))
        stop("'firth' must be TRUE or FALSE.", call. = FALSE)
    if (firth) {
        if (is.null(family$family) || family$family != "binomial" ||
            is.null(family$link) || family$link != "logit")
            stop("'firth = TRUE' currently requires family = binomial(link = \"logit\").",
                 call. = FALSE)
        if (inherits(x, "dgCMatrix") || (requireNamespace("bigmemory", quietly = TRUE) &&
                                          bigmemory::is.big.matrix(x)))
            stop("'firth = TRUE' is not supported for sparse or big.matrix designs.",
                 call. = FALSE)
        control$method <- 2L
    }
    
    is_sparse_matrix <- inherits(x, "dgCMatrix")
    if (is_sparse_matrix)
    {
        is_big_matrix <- FALSE
        if (!(control$method %in% c(2, 3)))
        {
            stop("for sparse (dgCMatrix) objects, 'method' must be 2 (LLT) or 3 (LDLT). ",
                 "QR / SVD on sparse matrices is not supported by this package.")
        }
    } else if (bigmemory::is.big.matrix(x))
    {
        is_big_matrix <- TRUE
        if (!(control$method %in% c(2, 3)))
        {
            stop("for big.matrix objects, 'method' must be 2 (LLT) or 3 (LDLT). ",
                 "QR / SVD methods would force the matrix to be fully read into RAM, ",
                 "defeating the purpose of bigmemory.")
        }
    } else if (is.matrix(x))
    {
        is_big_matrix <- FALSE
    } else
    {
        stop("x must be a matrix, a big.matrix object, or a Matrix::dgCMatrix")
    }
    
    xnames <- colnames(x)
    ynames <- if (is.matrix(y)) rownames(y) else names(y)
    
    nobs <- NROW(y)
    nvars <- ncol(x)
    
    if (is.null(weights)) 
        weights <- rep.int(1, nobs)
    else
        weights <- as.vector(weights)
    
    if (any(weights < 0))
        stop("negative weights not allowed")
    
    if (is.null(offset)) 
        offset <- rep.int(0, nobs)
    else
        offset  <- as.vector(offset)
    
    stopifnot(is.numeric(y), 
              is.numeric(weights),
              is.numeric(offset),
              NROW(y) == nrow(x),
              NROW(y) == NROW(weights),
              NROW(y) == NROW(offset)             
    )
    
    variance <- family$variance
    linkinv <- family$linkinv
    
    if (!is.function(variance) || !is.function(linkinv)) 
        stop("'family' argument seems not to be a valid family object", 
             call. = FALSE)
    
    if (is.null(family$family)) 
    {
        print(family)
        stop("'family' not recognized")
    }
    
    # from glm
    dev.resids  <- family$dev.resids
    aic         <- family$aic
    mu.eta      <- family$mu.eta 
    
    unless.null <- function(x, if.null) if (is.null(x)) if.null else x
    valideta    <- unless.null(family$valideta, function(eta) TRUE)
    validmu     <- unless.null(family$validmu,  function(mu)  TRUE)
    
    if (is.null(mustart)) 
    {
        ## calculates mustart and may change y and weights and set n (!)
        eval(family$initialize)
    } else 
    {
        mukeep <- mustart
        eval(family$initialize)
        mustart <- mukeep
    }
    
    y <- as.numeric(y)
    
    coefold <- NULL
    eta <-
        if (!is.null(etastart)) {
            etastart
        } else if (!is.null(start))
        {    
            if (length(start) != nvars)
            {
                stop(gettextf("length of 'start' should equal %d and correspond to initial coefs for %s", 
                              nvars, toString(xnames)), 
                     domain = NA)
            }
            
            coefold <- start
            offset + as.vector(x %*% start)
            
        } else family$linkfun(mustart)
    
    mu <- linkinv(eta)
    
    if (!(validmu(mu) && valideta(eta)))
        stop("cannot find valid starting values: please specify some", call. = FALSE)
    
    if (is.null(start)) start <- rep(0, nvars)
    
    fc <- family_code(family)
    fp <- family_params(family)

    if (is_sparse_matrix)
    {
        res <- fit_sparse_glm(x, drop(y), drop(weights), drop(offset),
                              drop(start), drop(mu), drop(eta),
                              variance, mu.eta, linkinv, dev.resids,
                              valideta, validmu,
                              as.integer(control$method), as.double(control$tol), as.integer(control$maxit),
                              as.integer(fc), fp)
        col_max <- apply(x, 2, max)
        col_min <- apply(x, 2, min)
        res$intercept <- any(is.int <- (col_max == col_min))
    } else if (!is_big_matrix && firth)
    {
        res <- fit_glm_firth(x, drop(y), drop(weights), drop(offset),
                             drop(start), drop(mu), drop(eta),
                             variance, mu.eta, linkinv, dev.resids,
                             valideta, validmu,
                             as.double(control$tol), as.integer(control$maxit),
                             as.integer(fc), fp)
        res$intercept <- any(is.int <- colMax_dense(x) == colMin_dense(x))
    } else if (!is_big_matrix)
    {
        res <- fit_glm(x, drop(y), drop(weights), drop(offset),
                       drop(start), drop(mu), drop(eta),
                       variance, mu.eta, linkinv, dev.resids,
                       valideta, validmu,
                       as.integer(control$method), as.double(control$tol), as.integer(control$maxit),
                       as.integer(fc), fp)

        res$intercept <- any(is.int <- colMax_dense(x) == colMin_dense(x))
    } else
    {
        res <- fit_big_glm(x@address, drop(y), drop(weights), drop(offset),
                           drop(start), drop(mu), drop(eta),
                           variance, mu.eta, linkinv, dev.resids,
                           valideta, validmu,
                           as.integer(control$method), as.double(control$tol), as.integer(control$maxit),
                           as.integer(fc), fp)

        res$intercept <- any(is.int <- big.colMax(x) == big.colMin(x))
    }
    
    if (!res$converged)
    {
        warning("fastglm_fit: algorithm did not converge", call. = FALSE)
    }
    
    eps <- 10 * .Machine$double.eps
    if (family$family == "binomial") 
    {
        if (any(res$fitted.values > 1 - eps) || any(res$fitted.values < eps))
            warning("fastglm_fit: fitted probabilities numerically 0 or 1 occurred", call. = FALSE)
    }
    if (family$family == "poisson") 
    {
        if (any(res$fitted.values < eps))
            warning("fastglm_fit: fitted rates numerically 0 occurred", call. = FALSE)
    }
    
    if (is.null(xnames))
    {
        ncx <- ncol(x)
        if (res$intercept)
        {
            which.int <- which(is.int)
            xnames    <- paste0("X", seq_len(ncx - 1L))
            names(res$coefficients) <- seq_len(ncx)
            names(res$coefficients)[-which.int] <- xnames
            names(res$coefficients)[which.int]  <- "(Intercept)"
        } else
        {
            names(res$coefficients) <- paste0("X", seq_len(ncx))
        }
    } else
    {
        names(res$coefficients) <- xnames
    }
    
    wtdmu         <- if (res$intercept) sum(weights * y) / sum(weights) else family$linkinv(offset)

    n.ok          <- nobs - sum(weights == 0)
    
    res$df.null   <- n.ok - as.integer(res$intercept)
    res$null.deviance <- sum(family$dev.resids(y, wtdmu, weights))
    res$family <- family
    res$prior.weights <- weights

    res$aic <- aic(y, nobs, res$fitted.values, res$prior.weights, res$deviance) + 2 * res$rank

    res$residuals <- (y - res$fitted.values) / family$mu.eta(res$linear.predictors)
    res$dispersion <-
        if (family$family %in% c("poisson", "binomial")) 1
        else if (res$df.residual > 0)
            sum((res$weights * res$residuals ^ 2)[weights > 0]) / res$df.residual
        else NaN

    res$y <- y
    res$n <- nobs
    res$x <- x        # reference to the model matrix; used by vcovHC / vcovCL
    res$class <- "fastglmFit"

    if (!is.null(res$cov.unscaled) && length(res$coefficients) > 0L) {
        nms <- names(res$coefficients)
        rownames(res$cov.unscaled) <- colnames(res$cov.unscaled) <- nms
    }

    res
}

#' @export
#' @rdname fastglm_fit
fastglm_control <- function(fastmethod = 0L, tol = 1e-7, maxit = 100L, ...) {
    if (...length() > 0L && "method" %in% ...names() && identical(fastmethod, 0L)) {
        fastmethod <- ...elt(which(...names() == "method")[1L])
        fastmethod_name <- "method"
    }
    else {
        fastmethod_name <- "fastmethod"
    }
    
    if (!is.numeric(fastmethod) || length(fastmethod) != 1L || !(fastmethod %in% 0:5)) 
        stop(sprintf("'%s' must be an interger between 0 and 5", fastmethod_name))
    
    if (!is.numeric(tol) || length(tol) != 1L || tol <= 0) 
        stop("value of 'tol' must be > 0")
    
    if (!is.numeric(maxit) || length(maxit) != 1L || maxit <= 0) 
        stop("maximum number of iterations must be > 0")
    
    list(method = fastmethod, tol = tol, maxit = maxit)
}

#' @exportS3Method stats::vcov fastglmFit
#' @rdname fastglm_fit
vcov.fastglmFit <- function(object, ...) {
    cov.unscaled <- object$cov.unscaled
    if (is.null(cov.unscaled)) {
        v <- diag(object[["se"]]^2, nrow = length(object$coefficients))
        rownames(v) <- colnames(v) <- names(coef(object))
        return(v)
    }
    disp <- if (is.null(object$dispersion) || is.nan(object$dispersion)) 1 else object$dispersion
    v <- disp * cov.unscaled
    rownames(v) <- colnames(v) <- names(coef(object))
    v
}

#' @exportS3Method summary fastglmFit
#' @rdname fastglm_fit
summary.fastglmFit <- function(object, ...) {
    summary.fastglm(object, ...)
}

skip_if_not_installed("logistf")

# Compare fastglm(firth = TRUE) against logistf::logistf, which is the
# canonical R implementation of Heinze & Schemper (2002) Firth's penalized
# logistic regression.  Coefficient agreement is to ~1e-7; SE agreement uses
# the looser 1e-2 because logistf reports a slightly different variance
# estimator -- our SE matches the standard (X'WX)^{-1} formula exactly.
.tight_logistf <- function() {
    logistf::logistf.control(lconv = 1e-12, gconv = 1e-12, xconv = 1e-12,
                              maxit = 200L)
}

test_that("Firth coefs match logistf on simulated logistic data", {
    set.seed(123)
    n <- 300
    x <- cbind(1, matrix(rnorm(n * 3), n, 3))
    eta <- x %*% c(0.2, 0.5, -0.4, 0.3)
    y  <- rbinom(n, 1, plogis(eta))

    m <- logistf::logistf(y ~ x[, -1], pl = FALSE, plconf = NULL,
                           control = .tight_logistf())
    f <- fastglm(x, y, family = binomial(), firth = TRUE,
                 tol = 1e-12, maxit = 200L)

    expect_equal(unname(coef(f)), unname(coef(m)), tolerance = 1e-7)
    # SE: logistf reports a different variance estimator than the standard
    # (X'WX)^{-1}; allow a looser tolerance here.
    expect_equal(unname(f$se), unname(sqrt(diag(vcov(m)))), tolerance = 1e-2)
    expect_true(f$converged)
    expect_s3_class(f, "fastglm_firth")
})

test_that("Firth converges under perfect separation (where unpenalized glm diverges)", {
    # Albert & Anderson (1984) classic separation case: y is perfectly
    # predicted by sign(x).  Standard glm() diverges; Firth must converge.
    set.seed(7)
    n <- 50
    x_var <- c(rnorm(n / 2, mean = -2), rnorm(n / 2, mean = 2))
    y     <- as.integer(x_var > 0)
    X     <- cbind(1, x_var)

    f <- fastglm(X, y, family = binomial(), firth = TRUE,
                 tol = 1e-12, maxit = 200L)
    expect_true(f$converged)
    expect_true(all(is.finite(coef(f))))
    expect_true(all(is.finite(f$se)))

    m <- logistf::logistf(y ~ x_var, pl = FALSE, plconf = NULL,
                           control = .tight_logistf())
    expect_equal(unname(coef(f)), unname(coef(m)), tolerance = 1e-6)
})

test_that("Firth matches logistf on a typical small-sample bias case", {
    # Heinze-Schemper-style: small n, modestly separated.
    set.seed(2024)
    n  <- 60
    z1 <- rnorm(n);  z2 <- rnorm(n)
    eta <- 0.5 + 1.2 * z1 - 0.8 * z2
    y   <- rbinom(n, 1, plogis(eta))
    X   <- cbind(1, z1, z2)

    m <- logistf::logistf(y ~ z1 + z2, pl = FALSE, plconf = NULL,
                           control = .tight_logistf())
    f <- fastglm(X, y, family = binomial(), firth = TRUE,
                 tol = 1e-12, maxit = 200L)

    expect_equal(unname(coef(f)), unname(coef(m)), tolerance = 1e-7)
})

test_that("firth = TRUE rejects unsupported families/links", {
    set.seed(1)
    n <- 50
    X <- cbind(1, rnorm(n))
    y <- rbinom(n, 1, 0.5)
    # probit not yet supported
    expect_error(fastglm(X, y, family = binomial("probit"), firth = TRUE),
                 "binomial\\(link = \"logit\"\\)")
    # poisson not allowed
    yp <- rpois(n, 2)
    expect_error(fastglm(X, yp, family = poisson(), firth = TRUE),
                 "binomial\\(link = \"logit\"\\)")
})

test_that("firth result reports unpenalized deviance and penalized.deviance", {
    set.seed(11)
    n <- 200
    X <- cbind(1, matrix(rnorm(n * 2), n, 2))
    y <- rbinom(n, 1, plogis(X %*% c(0.1, 0.5, -0.3)))
    f <- fastglm(X, y, family = binomial(), firth = TRUE)
    expect_true(is.finite(f$deviance))
    expect_true(is.finite(f$penalized.deviance))
    expect_true(is.finite(f$log.det.XtWX))
    # penalized = deviance - log|X'WX|
    expect_equal(f$penalized.deviance, f$deviance - f$log.det.XtWX,
                 tolerance = 1e-9)
    expect_true(isTRUE(f$firth))
})

test_that("fastglm_fit accepts firth = TRUE", {
    set.seed(99)
    n <- 100
    X <- cbind(1, matrix(rnorm(n * 2), n, 2))
    y <- rbinom(n, 1, plogis(X %*% c(0.1, 0.5, -0.3)))
    fit <- fastglm_fit(X, y, family = binomial(), firth = TRUE)
    expect_true(fit$converged)
    expect_true(isTRUE(fit$firth))
})

#' `summary()` method for `fastglm` fitted objects
#'
#' @param object `fastglm` fitted object
#' @param dispersion the dispersion parameter for the family used. Either a single numerical value or `NULL` (the default), when it is inferred from `object`.
#' @param ... not used
#' 
#' @returns A `summary.fastglm` object
#' 
#' @seealso [summary.glm()]
#' 
#' @method summary fastglm
#' 
#' @examples
#' x <- matrix(rnorm(10000 * 10), ncol = 10)
#' y <- 1 * (0.25 * x[,1] - 0.25 * x[,3] > rnorm(10000))
#' 
#' fit <- fastglm(x, y, family = binomial())
#' 
#' summary(fit)

#' @param correlation logical; if `TRUE`, the correlation matrix of the estimated parameters is returned.
#' @param symbolic.cor logical; if `TRUE`, print the correlations in a symbolic form (see `symnum`) rather than as numbers.
#' @exportS3Method summary fastglm
summary.fastglm <- function(object, dispersion = NULL,
                            correlation = FALSE, symbolic.cor = FALSE, ...)
{
    p <- object$rank

    est.disp <- FALSE
    df.r <- object$df.residual

    if (is.null(dispersion))
    {
        if (!(object$family$family %in% c("poisson", "binomial"))) est.disp <- TRUE
        dispersion <- object$dispersion
    }

    aliased <- is.na(coef(object))  # used in print method

    covmat.unscaled <- object$cov.unscaled
    if (is.null(covmat.unscaled) && p > 0)
    {
        # Fallback for older fitted objects without cov.unscaled
        covmat.unscaled <- diag(object$se ^ 2 / max(dispersion, .Machine$double.eps),
                                nrow = length(object$coefficients))
    }
    if (!is.null(covmat.unscaled) && length(object$coefficients) > 0L)
    {
        nms <- names(object$coefficients)
        rownames(covmat.unscaled) <- colnames(covmat.unscaled) <- nms
        # Match base R summary.glm: drop aliased rows/cols (rank x rank)
        if (any(aliased))
            covmat.unscaled <- covmat.unscaled[!aliased, !aliased, drop = FALSE]
    }
    covmat <- if (!is.null(covmat.unscaled)) dispersion * covmat.unscaled else NULL

    if (p > 0)
    {
        coef   <- object$coefficients
        se     <- object$se
        tvalue <- coef / se

        dn <- c("Estimate", "Std. Error")
        if (!est.disp)
        { # known dispersion
            pvalue <- 2 * pnorm(-abs(tvalue))
            coef.table <- cbind(coef, se, tvalue, pvalue)
            dimnames(coef.table) <- list(names(coef),
                                         c(dn, "z value","Pr(>|z|)"))
        } else if (df.r > 0)
        {
            pvalue <- 2 * pt(-abs(tvalue), df.r)
            coef.table <- cbind(coef, se, tvalue, pvalue)
            dimnames(coef.table) <- list(names(coef),
                                         c(dn, "t value","Pr(>|t|)"))
        } else
        { # df.r == 0
            coef.table <- cbind(coef, NaN, NaN, NaN)
            dimnames(coef.table) <- list(names(coef),
                                         c(dn, "t value","Pr(>|t|)"))
        }

        df.f <- length(aliased)
    } else
    {
        coef.table <- matrix(0, 0L, 4L)
        dimnames(coef.table) <-
            list(NULL, c("Estimate", "Std. Error", "t value", "Pr(>|t|)"))
        covmat.unscaled <- covmat <- matrix(0, 0L, 0L)
        df.f <- length(aliased)
    }
    df.int <- if (object$intercept) 1L else 0L

    ## these need not all exist, e.g. na.action.
    keep <- match(c("call","terms","family","deviance", "aic",
                    "contrasts", "df.residual","null.deviance","df.null",
                    "iter", "na.action"), names(object), 0L)
    ans <- c(object[keep],
             list(deviance.resid = residuals(object, type = "deviance"),
                  coefficients = coef.table,
                  aliased = aliased,
                  dispersion = dispersion,
                  df = c(object$rank, df.r, df.f),
                  cov.unscaled = covmat.unscaled,
                  cov.scaled = covmat))

    if (correlation && p > 0 && !is.null(covmat.unscaled))
    {
        dd <- sqrt(diag(covmat.unscaled))
        ans$correlation <- covmat.unscaled / outer(dd, dd)
        ans$symbolic.cor <- symbolic.cor
    }
    class(ans) <- "summary.glm"
    return(ans)
}

#' `vcov()` method for `fastglm` fitted objects
#'
#' @param object a fitted object of class inheriting from `"fastglm"`.
#' @param ... additional arguments (currently unused).
#'
#' @returns The estimated variance-covariance matrix of the fitted coefficients.
#' For rank-deficient fits, rows and columns corresponding to aliased
#' coefficients are filled with `NA`.
#'
#' @method vcov fastglm
#' @exportS3Method stats::vcov fastglm
vcov.fastglm <- function(object, ...)
{
    cov.unscaled <- object$cov.unscaled
    if (is.null(cov.unscaled))
    {
        v <- diag(object[["se"]]^2, nrow = length(object$coefficients))
        rownames(v) <- colnames(v) <- names(coef(object))
        return(v)
    }
    disp <- if (is.null(object$dispersion) || is.nan(object$dispersion)) 1 else object$dispersion
    v <- disp * cov.unscaled
    nms <- names(coef(object))
    rownames(v) <- colnames(v) <- nms
    v
}

#' Heteroskedasticity-consistent (HC) variance estimators for `fastglm` objects
#'
#' Methods for `sandwich::vcovHC()` on objects of class `"fastglm"` and
#' `"fastglmFit"`. Load `sandwich` (`library(sandwich)`) before calling
#' `vcovHC(fit)`; otherwise no `vcovHC` generic is in scope.
#'
#' @param object a fitted object of class `"fastglm"` or `"fastglmFit"`.
#' @param type one of `"HC0"`, `"HC1"`, `"HC2"`, `"HC3"`. Default `"HC3"` matches
#'   `sandwich::vcovHC.glm`.
#' @param ... not used.
#'
#' @returns A `p x p` heteroskedasticity-consistent variance-covariance matrix.
#'
#' @details
#' Computes the Eicker-Huber-White sandwich estimator
#' `bread %*% meat %*% bread`, where `bread = (X' W X)^{-1}` (already stored as
#' `cov.unscaled`) and `meat = X' diag(omega_i) X`. With `s_i = w_i^2 * r_i`
#' the score contribution from observation `i`, the omegas are:
#' \describe{
#'   \item{`HC0`}{`omega_i = s_i^2`}
#'   \item{`HC1`}{`HC0` rescaled by `n / (n - p)`}
#'   \item{`HC2`}{`omega_i = s_i^2 / (1 - h_i)`}
#'   \item{`HC3`}{`omega_i = s_i^2 / (1 - h_i)^2`}
#' }
#' where `r_i` is the working residual `(y - mu) / mu.eta(eta)`,
#' `w_i^2 = prior.weight * mu.eta(eta)^2 / variance(mu)` is the IRLS working
#' weight, and `h_i = w_i^2 * x_i' (X' W X)^(-1) x_i` is the IRLS leverage.
#' Equivalent to `sandwich::vcovHC.glm`.
#'
#' Requires the model matrix `x` stored on the fitted object (set automatically
#' by `fastglm()`, `fastglmPure()`, and `fastglm_fit()` since version 0.0.6).
#'
#' @examples
#' if (requireNamespace("sandwich", quietly = TRUE)) {
#'   x <- cbind(1, matrix(rnorm(500 * 4), ncol = 4))
#'   y <- rbinom(500, 1, plogis(x %*% c(0.2, 0.3, -0.4, 0.1, 0.2)))
#'   fit <- fastglm(x, y, family = binomial())
#'   sandwich::vcovHC(fit)
#'   sandwich::vcovHC(fit, type = "HC0")
#' }
#'
#' @name vcovHC.fastglm
NULL

#' @rdname vcovHC.fastglm
#' @exportS3Method sandwich::vcovHC fastglm
vcovHC.fastglm <- function(object, type = c("HC3", "HC2", "HC1", "HC0"), ...)
{
    type <- match.arg(type)
    .vcov_hc_fastglm(object, type)
}

#' @rdname vcovHC.fastglm
#' @exportS3Method sandwich::vcovHC fastglmFit
vcovHC.fastglmFit <- function(object, type = c("HC3", "HC2", "HC1", "HC0"), ...)
{
    type <- match.arg(type)
    .vcov_hc_fastglm(object, type)
}

#' Empirical estimating-function and bread methods for `fastglm` objects
#'
#' Methods for `sandwich::estfun()` and `sandwich::bread()`, which let
#' `sandwich::vcovCL()`, `sandwich::vcovBS()`, and the rest of the *sandwich*
#' machinery work directly on `fastglm` and `fastglmFit` objects.  Load
#' `sandwich` (`library(sandwich)`) before calling them.
#'
#' @param x a fitted object of class `"fastglm"` or `"fastglmFit"`.
#' @param ... not used.
#'
#' @details
#' `estfun(x)` returns the per-observation empirical estimating-function
#' contributions, an `n x p` matrix whose `i`-th row is
#' `(y_i - mu_i) * mu.eta_i / variance(mu_i) * x_i / dispersion`. `bread(x)`
#' returns `(X' W X)^(-1) * n * dispersion` (the *sandwich* convention).
#' For poisson, binomial, and negative-binomial families the dispersion is
#' fixed at 1, exactly matching `sandwich::estfun.glm()` /
#' `sandwich::bread.glm()`.
#'
#' @returns `estfun()` returns an `n x p` matrix; `bread()` returns a
#'   `p x p` matrix.
#'
#' @name fastglm-sandwich
NULL

# Internal: dispersion used by sandwich's GLM bread/estfun convention.
.sandwich_dispersion <- function(object, inp)
{
    fam <- object$family$family
    if (substr(fam, 1L, 17L) %in% c("poisson", "binomial", "Negative Binomial"))
        return(1)
    sum_w2 <- sum(inp$w2, na.rm = TRUE)
    if (sum_w2 <= 0) return(1)
    sum(inp$s^2, na.rm = TRUE) / sum_w2
}

#' @rdname fastglm-sandwich
#' @exportS3Method sandwich::estfun fastglm
estfun.fastglm <- function(x, ...)
{
    inp  <- .vcov_hc_meat_inputs(x)
    disp <- .sandwich_dispersion(x, inp)
    rval <- (inp$s / disp) * inp$x
    colnames(rval) <- names(x$coefficients)
    rval
}

#' @rdname fastglm-sandwich
#' @exportS3Method sandwich::estfun fastglmFit
estfun.fastglmFit <- estfun.fastglm

#' @rdname fastglm-sandwich
#' @exportS3Method sandwich::bread fastglm
bread.fastglm <- function(x, ...)
{
    inp  <- .vcov_hc_meat_inputs(x)
    disp <- .sandwich_dispersion(x, inp)
    nms  <- names(x$coefficients)
    out  <- inp$bread * inp$n * disp
    rownames(out) <- colnames(out) <- nms
    out
}

#' @rdname fastglm-sandwich
#' @exportS3Method sandwich::bread fastglmFit
bread.fastglmFit <- bread.fastglm

# Internal: pull (X, working residuals, working weights, bread) off a fitted
# fastglm/fastglmFit object.  Returns a list with elements x, r, w, bread, p.
.vcov_hc_meat_inputs <- function(object)
{
    x <- object$x
    if (is.null(x))
        stop("vcovHC/vcovCL require the design matrix to be stored on the fitted object. ",
             "Refit with fastglm/fastglmPure/fastglm_fit (>= 0.0.6).", call. = FALSE)
    if (inherits(x, "big.matrix"))
        x <- x[]
    if (inherits(x, "dgCMatrix"))
        x <- as.matrix(x)

    bread <- object$cov.unscaled
    if (is.null(bread))
        stop("'cov.unscaled' missing from fitted object; refit with current fastglm.", call. = FALSE)

    fam <- object$family
    eta <- object$linear.predictors
    mu  <- object$fitted.values
    pw  <- object$prior.weights
    if (is.null(pw)) pw <- rep(1, length(eta))

    mu_eta_vec <- fam$mu.eta(eta)
    var_mu_vec <- fam$variance(mu)
    y <- object$y
    r <- (y - mu) / mu_eta_vec                              # working residual
    w2 <- pw * mu_eta_vec^2 / var_mu_vec                    # working weight (W in (X'WX))

    # GLM score scalar: s_i = w^2_i * r_i = pw_i * mu_eta_i * (y_i - mu_i) / V(mu_i).
    # Note: do NOT divide by dispersion here. sandwich::estfun.glm divides by an
    # internal dispersion estimate, but sandwich::bread.glm multiplies by the
    # SAME dispersion, so the two cancel in the final sandwich formula.
    s <- w2 * r

    list(x = x, r = r, s = s, w2 = w2, bread = bread,
         p = length(object$coefficients), n = length(y))
}

.vcov_hc_fastglm <- function(object, type)
{
    inp <- .vcov_hc_meat_inputs(object)
    x <- inp$x; s <- inp$s; w2 <- inp$w2; bread <- inp$bread
    n <- inp$n; p <- inp$p

    omega <- s^2

    if (type %in% c("HC2", "HC3"))
    {
        # IRLS leverage h_i = w_i^2 * x_i' (X'WX)^{-1} x_i
        h <- w2 * rowSums((x %*% bread) * x)
        h <- pmin(h, 1 - 1e-12)
        if (type == "HC2") omega <- omega / (1 - h)
        else               omega <- omega / (1 - h)^2
    }

    XtMX <- crossprod(x, omega * x)
    V <- bread %*% XtMX %*% bread

    if (type == "HC1")
        V <- V * (n / max(n - p, 1))

    nms <- names(object$coefficients)
    rownames(V) <- colnames(V) <- nms
    V
}

#' @exportS3Method print fastglm
print.fastglm <- function(x, digits = max(3L, getOption("digits") - 3L), ...)
{
    cat("\nCall:  ", paste(deparse(x$call), sep = "\n", collapse = "\n"), 
        "\n\n", sep = "")
    
    if (length(coef(x)) > 0L) {
        cat("Coefficients")
        if (is.character(co <- x$contrasts)) {
            cat("  [contrasts: ", apply(cbind(names(co), co), 1L, paste, collapse = "="), "]")
        }
        
        cat(":\n")
        print.default(format(x$coefficients, digits = digits), 
                      print.gap = 2, quote = FALSE)
    }
    else {
        cat("No coefficients\n\n")
    }
}

#' @exportS3Method residuals fastglm
residuals.fastglm <- function(object, 
                              type = c("deviance", "pearson", "working", "response", "partial"), 
                              ...)
{
    class(object) <- "glm"
    
    residuals(object, type = type, ...)
}

#' @exportS3Method logLik fastglm
logLik.fastglm <- function(object, ...)
{
    class(object) <- "glm"
    
    logLik(object, ...)
}

#' @exportS3Method deviance fastglm
deviance.fastglm <- function(object, ...)
{
    class(object) <- "glm"
    
    deviance(object, ...)
}

#' @exportS3Method family fastglm
family.fastglm <- function(object, ...)
{
    class(object) <- "glm"
    
    family(object, ...)
}


#' Obtains predictions and optionally estimates standard errors of those predictions from a fitted generalized linear model object.
#' @param object a fitted object of class inheriting from `"fastglm"`.
#' @param newdata a matrix to be used for prediction.
#' @param type the type of prediction required. The default is on the scale of the linear predictors;
#' the alternative "\code{response}" is on the scale of the response variable. Thus for a default binomial
#' model the default predictions are of log-odds (probabilities on logit scale) and \code{type = "response"}
#'  gives the predicted probabilities. The "\code{terms}" option returns a matrix giving the fitted values of each
#'  term in the model formula on the linear predictor scale.
#'
#' The value of this argument can be abbreviated.
#' @param se.fit logical switch indicating if standard errors are required.
#' @param dispersion the dispersion of the GLM fit to be assumed in computing the standard errors.
#' If omitted, that returned by \code{summary} applied to the object is used.
#' @param ... further arguments passed to or from other methods.
#' @export
predict.fastglm <- function(object,
                            newdata = NULL,
                            type = c("link", "response"),
                            se.fit = FALSE,
                            dispersion = NULL, ...)
{
    type <- match.arg(type)

    if (is.null(dispersion))
        dispersion <- if (is.null(object$dispersion) || is.nan(object$dispersion)) 1 else object$dispersion

    pred <- predict_fastglm_lm(object, newdata, se.fit, dispersion = dispersion, ...)

    if (type == "response")
    {
        fam <- family(object)
        if (se.fit)
        {
            mu_eta <- fam$mu.eta(pred$fit)
            pred$fit    <- fam$linkinv(pred$fit)
            pred$se.fit <- pred$se.fit * abs(mu_eta)
        } else
        {
            pred <- fam$linkinv(pred)
        }
    }
    pred
}


predict_fastglm_lm <- function(object, newdata, se.fit = FALSE, dispersion = 1)
{
    dims <- dim(newdata)
    if (is.null(dims))
    {
        newdata <- as.matrix(newdata)
        dims <- dim(newdata)
    }
    beta <- object$coefficients

    if (dims[2L] != length(beta))
    {
        stop("newdata provided does not match fitted model 'object'")
    }
    eta <- drop(newdata %*% beta)

    if (!se.fit) return(eta)

    cov.unscaled <- object$cov.unscaled
    if (is.null(cov.unscaled))
        stop("standard errors of predictions require 'cov.unscaled' from a refit; reinstall fastglm and refit")

    cov.scaled <- dispersion * cov.unscaled
    se <- sqrt(rowSums((newdata %*% cov.scaled) * newdata))
    list(fit = eta, se.fit = se, residual.scale = sqrt(dispersion))
}

#' Hurdle (two-part) Poisson / Negative-Binomial regression
#'
#' Fits a hurdle model for count data with excess zeros: a binary
#' (zero / non-zero) regression on the full sample combined with a zero-
#' truncated count regression on the positive subset. The two parts have
#' independent likelihoods and are estimated jointly. With `dist = "negbin"`
#' the negative-binomial dispersion `theta` is estimated by an inner
#' Brent-1D MLE inside the C++ driver, alternating with the count-side
#' IRLS.
#'
#' This is the fastglm analogue of [pscl::hurdle()] -- coefficients agree
#' to 1e-6 on standard datasets, with all numerical loops in C++.
#'
#' @param formula a [Formula::Formula] of the form `y ~ x1 + x2 | z1 + z2`
#'   where the right-hand side after `|` specifies the zero-model design.
#'   If `|` is absent, the same RHS is used for both parts.
#' @param data optional data frame / environment from which to draw model
#'   matrix variables.
#' @param subset,na.action standard model-frame arguments.
#' @param weights optional non-negative prior weights.
#' @param offset optional vector or matrix.  If a 2-column matrix, columns
#'   are taken as `(count_offset, zero_offset)`.
#' @param dist count component distribution: `"poisson"` or `"negbin"`.
#' @param zero.dist zero/non-zero distribution; only `"binomial"` is
#'   currently supported.
#' @param link character, link for the zero/non-zero binomial.  One of
#'   `"logit"` (default), `"probit"`, `"cloglog"`, `"log"`.
#' @param init.theta optional positive scalar starting value for the NB
#'   dispersion.  If `NULL`, a method-of-moments initializer is used.
#' @param tol,maxit IRLS-loop convergence tolerance and iteration cap.
#' @param outer.tol,outer.maxit `(beta, theta)` outer-loop convergence
#'   tolerance and iteration cap (only used when `dist = "negbin"`).
#' @param theta.tol,theta.maxit Brent-loop tolerance / iteration cap for
#'   the inner theta MLE (only used when `dist = "negbin"`).
#' @param model logical; keep the model frame on the result.
#' @param y,x logical; keep the response / design matrices on the result.
#'
#' @returns A list of class `c("fastglm_hurdle", "fastglm")` with elements
#'   `coefficients` (a list with `count` and `zero`), `se` (likewise),
#'   `vcov` (block-diagonal), `loglik`, `loglik_count`, `loglik_zero`,
#'   `theta`, `SE.theta`, `iter_*`, and `converged`.
#'
#' @examples
#' set.seed(1)
#' n <- 300
#' x1 <- rnorm(n); x2 <- rnorm(n)
#' eta_count <- 1.0 + 0.4*x1 - 0.2*x2
#' lam <- exp(eta_count)
#' y_count <- rpois(n, lam)
#' eta_zero  <- -0.5 + 0.6*x1
#' p_pos     <- plogis(eta_zero)
#' is_pos    <- rbinom(n, 1, p_pos)
#' # zero-truncated: resample any rare zero from positive Poisson
#' for (i in seq_len(n)) {
#'   while (y_count[i] == 0) y_count[i] <- rpois(1, lam[i])
#' }
#' y <- ifelse(is_pos == 1, y_count, 0)
#' df <- data.frame(y = y, x1 = x1, x2 = x2)
#' fit <- fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson")
#' coef(fit, model = "count")
#' coef(fit, model = "zero")
#'
#' @export
fastglm_hurdle <- function(formula, data, subset, na.action,
                           weights, offset,
                           dist        = c("poisson", "negbin"),
                           zero.dist   = "binomial",
                           link        = c("logit", "probit", "cloglog", "log"),
                           init.theta  = NULL,
                           tol         = 1e-8,
                           maxit       = 100L,
                           outer.tol   = 1e-7,
                           outer.maxit = 50L,
                           theta.tol   = 1e-8,
                           theta.maxit = 100L,
                           model = TRUE, y = TRUE, x = FALSE)
{
    if (!requireNamespace("Formula", quietly = TRUE))
        stop("Package 'Formula' is required for fastglm_hurdle(); install it.",
             call. = FALSE)

    dist      <- match.arg(dist)
    link      <- match.arg(link)
    zero.dist <- match.arg(zero.dist, choices = "binomial")

    # ---- 1. Build the model frame using Formula's two-RHS support ------------
    cl <- match.call()
    if (missing(data)) data <- environment(formula)
    f  <- Formula::Formula(formula)
    has_two_rhs <- length(f)[2] >= 2

    mf <- match.call(expand.dots = FALSE)
    mf$formula <- f
    mf$dist <- mf$zero.dist <- mf$link <- mf$init.theta <- NULL
    mf$tol  <- mf$maxit  <- mf$outer.tol <- mf$outer.maxit <- NULL
    mf$theta.tol <- mf$theta.maxit <- NULL
    mf$model <- mf$y <- mf$x <- NULL
    mf[[1L]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())

    mt_count <- terms(f, data = data, rhs = 1L)
    mt_zero  <- if (has_two_rhs) terms(f, data = data, rhs = 2L)
                else             mt_count

    Y_resp <- model.response(mf, "numeric")
    if (!is.numeric(Y_resp) || any(Y_resp < 0))
        stop("'y' must be non-negative.", call. = FALSE)
    n <- length(Y_resp)

    X_count <- model.matrix(mt_count, mf)
    Z_zero  <- model.matrix(mt_zero,  mf)

    # weights / offset
    wt <- model.weights(mf)
    if (is.null(wt)) wt <- rep(1, n)
    if (!is.numeric(wt) || any(wt < 0))
        stop("'weights' must be non-negative.", call. = FALSE)

    off <- model.offset(mf)
    if (is.null(off)) {
        off_count <- rep(0, n); off_zero <- rep(0, n)
    } else if (is.matrix(off) && ncol(off) == 2L) {
        off_count <- as.numeric(off[, 1L]); off_zero <- as.numeric(off[, 2L])
    } else {
        off_count <- as.numeric(off);       off_zero <- as.numeric(off)
    }

    # ---- 2. Map link to fam_code & family-fns for the zero/non-zero binomial -
    zero_family <- switch(link,
                          "logit"   = binomial("logit"),
                          "probit"  = binomial("probit"),
                          "cloglog" = binomial("cloglog"),
                          "log"     = binomial("log"))
    zero_fam_code <- switch(link,
                            "logit"   = 3L,
                            "probit"  = 4L,
                            "cloglog" = 5L,
                            "log"     = 6L)

    dist_code <- switch(dist, "poisson" = 0L, "negbin" = 1L)

    init_th <- if (!is.null(init.theta)) {
        if (!is.numeric(init.theta) || length(init.theta) != 1L ||
            !is.finite(init.theta) || init.theta <= 0)
            stop("'init.theta' must be a positive finite scalar.", call. = FALSE)
        init.theta
    } else {
        -1
    }

    # ---- 3. Dispatch into C++ ------------------------------------------------
    res <- fit_glm_hurdle(
        x_count           = X_count,
        z_zero            = Z_zero,
        y                 = as.numeric(Y_resp),
        weights           = wt,
        offset_count      = off_count,
        offset_zero       = off_zero,
        dist_code         = dist_code,
        zero_fam_code     = zero_fam_code,
        init_theta        = as.double(init_th),
        tol               = as.double(tol),
        maxit             = as.integer(maxit),
        outer_tol         = as.double(outer.tol),
        outer_maxit       = as.integer(outer.maxit),
        theta_tol         = as.double(theta.tol),
        theta_maxit       = as.integer(theta.maxit),
        var_fun_zero      = zero_family$variance,
        mu_eta_fun_zero   = zero_family$mu.eta,
        linkinv_fun_zero  = zero_family$linkinv,
        dev_resids_fun_zero = zero_family$dev.resids,
        valideta_fun_zero = zero_family$valideta,
        validmu_fun_zero  = zero_family$validmu)

    # ---- 4. Cosmetic post-processing -----------------------------------------
    cn_count <- colnames(X_count); if (is.null(cn_count)) cn_count <- paste0("xc", seq_len(ncol(X_count)))
    cn_zero  <- colnames(Z_zero);  if (is.null(cn_zero))  cn_zero  <- paste0("xz", seq_len(ncol(Z_zero)))
    names(res$coefficients_count) <- cn_count
    names(res$coefficients_zero)  <- cn_zero
    names(res$se_count) <- cn_count
    names(res$se_zero)  <- cn_zero
    if (!is.null(res$vcov_count)) dimnames(res$vcov_count) <- list(cn_count, cn_count)
    if (!is.null(res$vcov_zero))  dimnames(res$vcov_zero)  <- list(cn_zero,  cn_zero)

    coefficients <- list(count = res$coefficients_count,
                         zero  = res$coefficients_zero)
    se           <- list(count = res$se_count, zero = res$se_zero)

    p_c <- length(coefficients$count); p_z <- length(coefficients$zero)
    pp  <- p_c + p_z
    vcov_full <- matrix(0, pp, pp)
    rownames(vcov_full) <- colnames(vcov_full) <-
        c(paste0("count_", cn_count), paste0("zero_", cn_zero))
    vcov_full[seq_len(p_c), seq_len(p_c)] <- res$vcov_count
    vcov_full[p_c + seq_len(p_z), p_c + seq_len(p_z)] <- res$vcov_zero

    out <- list(
        coefficients   = coefficients,
        se             = se,
        vcov           = vcov_full,
        loglik         = res$loglik,
        loglik_count   = res$loglik_count,
        loglik_zero    = res$loglik_zero,
        theta          = res$theta,
        SE.theta       = res[["SE.theta"]],
        iter_count     = res$iter_count,
        iter_zero      = res$iter_zero,
        outer_iter     = res$outer_iter,
        theta_iter     = res$theta_iter,
        converged      = res$converged,
        n              = res$n,
        n_positive     = res$n_positive,
        df.residual    = n - res$df,
        dist           = dist,
        link           = link,
        zero.dist      = zero.dist,
        formula        = f,
        terms          = list(count = mt_count, zero = mt_zero),
        levels         = .getXlevels(mt_count, mf),
        contrasts      = list(count = attr(X_count, "contrasts"),
                              zero  = attr(Z_zero,  "contrasts")),
        weights        = wt,
        offset         = list(count = off_count, zero = off_zero),
        eta_count      = res$eta_count,
        mu_count_truncated = res$mu_count_truncated,
        lambda         = res$lambda,
        eta_zero       = res$eta_zero,
        mu_zero        = res$mu_zero,
        call           = cl
    )
    if (model) out$model <- mf
    if (y)     out$y     <- Y_resp
    if (x)     out$x     <- list(count = X_count, zero = Z_zero)
    class(out) <- c("fastglm_hurdle", "fastglm")
    out
}

#' @export
coef.fastglm_hurdle <- function(object, model = c("full", "count", "zero"), ...)
{
    model <- match.arg(model)
    if (model == "full") {
        cn <- names(object$coefficients$count)
        zn <- names(object$coefficients$zero)
        out <- c(object$coefficients$count, object$coefficients$zero)
        names(out) <- c(paste0("count_", cn), paste0("zero_", zn))
        return(out)
    }
    object$coefficients[[model]]
}

#' @export
vcov.fastglm_hurdle <- function(object, model = c("full", "count", "zero"), ...)
{
    model <- match.arg(model)
    if (model == "full") return(object$vcov)
    p_c <- length(object$coefficients$count)
    p_z <- length(object$coefficients$zero)
    if (model == "count") return(object$vcov[seq_len(p_c), seq_len(p_c), drop = FALSE])
    object$vcov[p_c + seq_len(p_z), p_c + seq_len(p_z), drop = FALSE]
}

#' @export
logLik.fastglm_hurdle <- function(object, ...)
{
    df <- length(object$coefficients$count) + length(object$coefficients$zero) +
          (object$dist == "negbin")
    structure(object$loglik, df = df, nobs = object$n,
              class = "logLik")
}

#' @export
print.fastglm_hurdle <- function(x, digits = max(3, getOption("digits") - 3), ...)
{
    cat("\nfastglm_hurdle:  hurdle ", x$dist, " | ", x$zero.dist,
        " (link = ", x$link, ")\n", sep = "")
    cat("Call:  "); print(x$call)
    cat("\nCount model coefficients (zero-truncated", x$dist, "):\n")
    print(format(x$coefficients$count, digits = digits), quote = FALSE)
    cat("\nZero-hurdle model coefficients (binomial,", x$link, "):\n")
    print(format(x$coefficients$zero, digits = digits), quote = FALSE)
    if (x$dist == "negbin")
        cat("\nTheta:", format(x$theta, digits = digits),
            "  SE:", format(x[["SE.theta"]], digits = digits), "\n")
    cat("Log-likelihood:", format(x$loglik, digits = digits),
        "on", length(x$coefficients$count) + length(x$coefficients$zero) +
             (x$dist == "negbin"), "df\n")
    invisible(x)
}

skip_if_not_installed("pscl")

# fastglm_hurdle vs pscl::hurdle on simulated data and the canonical
# bioChemists data shipped with pscl.  Coefficient agreement to 1e-4
# (pscl uses optim/BFGS, fastglm uses Fisher-scoring IRLS — both reach
# the same MLE but with slightly different numerical paths).

.tight_pscl <- function() pscl::hurdle.control(method = "BFGS",
                                               reltol = 1e-12,
                                               maxit  = 5000)

test_that("Poisson hurdle: coefs match pscl::hurdle on simulated data", {
    set.seed(11)
    n  <- 400
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
    lam   <- exp(eta_c)
    eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
    p_pos <- plogis(eta_z)
    is_pos <- rbinom(n, 1, p_pos)
    # zero-truncated Poisson draws on the positives
    yt <- integer(n)
    for (i in seq_len(n)) {
        repeat {
            v <- rpois(1, lam[i])
            if (v > 0) { yt[i] <- v; break }
        }
    }
    y <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x1 = x1, x2 = x2)

    f  <- fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson",
                         tol = 1e-12, maxit = 200L)
    p  <- pscl::hurdle(y ~ x1 + x2, data = df, dist = "poisson",
                       control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    # Joint log-likelihood agreement
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-6)
    expect_true(f$converged)
})

test_that("Poisson hurdle: SEs agree with pscl::hurdle", {
    set.seed(12)
    n  <- 500
    x  <- rnorm(n)
    eta_c <- 0.5 + 0.3 * x
    lam   <- exp(eta_c)
    eta_z <- -0.2 + 0.4 * x
    is_pos <- rbinom(n, 1, plogis(eta_z))
    yt <- integer(n)
    for (i in seq_len(n)) {
        repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
    }
    y <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x = x)

    f <- fastglm_hurdle(y ~ x, data = df, dist = "poisson",
                        tol = 1e-12, maxit = 200L)
    p <- pscl::hurdle(y ~ x, data = df, dist = "poisson",
                      control = .tight_pscl())

    expect_equal(unname(f$se$count), unname(sqrt(diag(vcov(p)))[1:2]),
                 tolerance = 1e-3)
    expect_equal(unname(f$se$zero),  unname(sqrt(diag(vcov(p)))[3:4]),
                 tolerance = 1e-3)
})

test_that("Hurdle handles two-RHS formula y ~ x | z (different zero design)", {
    set.seed(13)
    n  <- 400
    x1 <- rnorm(n);  x2 <- rnorm(n);  z1 <- rnorm(n)
    eta_c <- 0.6 + 0.3 * x1 - 0.2 * x2
    lam   <- exp(eta_c)
    eta_z <- -0.3 + 0.7 * z1
    is_pos <- rbinom(n, 1, plogis(eta_z))
    yt <- integer(n)
    for (i in seq_len(n)) {
        repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
    }
    y <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x1 = x1, x2 = x2, z1 = z1)

    f <- fastglm_hurdle(y ~ x1 + x2 | z1, data = df, dist = "poisson",
                        tol = 1e-12, maxit = 200L)
    p <- pscl::hurdle(y ~ x1 + x2 | z1, data = df, dist = "poisson",
                      control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    expect_length(coef(f, "zero"), 2L)         # intercept + z1
    expect_length(coef(f, "count"), 3L)        # intercept + x1 + x2
})

test_that("NB hurdle: coefs and theta match pscl::hurdle", {
    skip_if_not_installed("MASS")
    set.seed(14)
    n  <- 600
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.8 + 0.4 * x1 - 0.3 * x2
    lam   <- exp(eta_c)
    eta_z <- -0.4 + 0.5 * x1
    is_pos <- rbinom(n, 1, plogis(eta_z))
    # zero-truncated NB draws (theta = 1.5)
    yt <- integer(n)
    for (i in seq_len(n)) {
        repeat {
            v <- MASS::rnegbin(1, mu = lam[i], theta = 1.5)
            if (v > 0) { yt[i] <- v; break }
        }
    }
    y <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x1 = x1, x2 = x2)

    f <- fastglm_hurdle(y ~ x1 + x2, data = df, dist = "negbin",
                        tol = 1e-10, maxit = 200L,
                        outer.tol = 1e-9, outer.maxit = 50L,
                        theta.tol = 1e-10)
    p <- pscl::hurdle(y ~ x1 + x2, data = df, dist = "negbin",
                      control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-3)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-3)
    expect_equal(unname(f$theta), unname(p$theta), tolerance = 1e-2)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-4)
})

test_that("Hurdle Poisson: log-link returns valid block-diagonal vcov", {
    set.seed(15)
    n  <- 300
    x  <- rnorm(n)
    eta_c <- 0.6 + 0.3 * x
    is_pos <- rbinom(n, 1, plogis(-0.2 + 0.5 * x))
    yt <- integer(n)
    for (i in seq_len(n)) {
        repeat { v <- rpois(1, exp(eta_c[i])); if (v > 0) { yt[i] <- v; break } }
    }
    y <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x = x)
    f <- fastglm_hurdle(y ~ x, data = df, dist = "poisson")
    V <- vcov(f, "full")
    p_c <- length(coef(f, "count"))
    # off-diagonal blocks must be zero -- the two parts factorize
    expect_true(all(V[seq_len(p_c), -seq_len(p_c)] == 0))
    expect_true(all(diag(V) > 0))
})

test_that("pscl::hurdle on bioChemists matches fastglm_hurdle", {
    skip_if_not_installed("pscl")
    data("bioChemists", package = "pscl")
    f <- fastglm_hurdle(art ~ ., data = bioChemists, dist = "poisson",
                        tol = 1e-12, maxit = 200L)
    p <- pscl::hurdle(art ~ ., data = bioChemists, dist = "poisson",
                      control = .tight_pscl())
    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-6)
})

test_that("Hurdle with zero.dist != binomial errors clearly", {
    set.seed(1)
    n <- 50
    df <- data.frame(y = rpois(n, 1) + 1, x = rnorm(n))
    expect_error(fastglm_hurdle(y ~ x, data = df,
                                zero.dist = "poisson"),
                 regexp = "binomial")
})

# Map a family object to the integer FamilyCode enum understood by the C++
# solver.  Returns -1L when the (family, link) pair has no native fast path,
# in which case the solver falls back to the per-iteration R callbacks.
family_code <- function(family) {
    if (is.null(family) || is.null(family$family) || is.null(family$link))
        return(-1L)

    famname <- family$family
    link    <- family$link

    # Quasi-binomial / quasi-poisson share C++ kernels with binomial / poisson.
    # Dispersion handling is done R-side based on family$family.
    if (famname == "quasibinomial") famname <- "binomial"
    if (famname == "quasipoisson")  famname <- "poisson"

    # MASS::negative.binomial(theta) sets family$family to "Negative Binomial(K)".
    if (grepl("^Negative Binomial", famname)) famname <- "negbin"

    # statmod::tweedie() uses a generic "power" link; map to the standard four.
    # var.power / link.power are not direct slots on the family object — they
    # live in the closure environment of family$variance / family$linkfun.
    if (famname == "Tweedie") {
        lp <- tweedie_link_power(family)
        if (is.null(lp) || !is.finite(lp)) lp <- 0
        link <- switch(as.character(lp),
                       "0"   = "log",
                       "1"   = "identity",
                       "-1"  = "inverse",
                       "0.5" = "sqrt",
                       link)
    }

    key <- paste0(famname, ":", link)
    code <- switch(key,
        "gaussian:identity"          = 0L,
        "gaussian:log"               = 1L,
        "gaussian:inverse"           = 2L,
        "binomial:logit"             = 3L,
        "binomial:probit"            = 4L,
        "binomial:cloglog"           = 5L,
        "binomial:log"               = 6L,
        "poisson:log"                = 7L,
        "poisson:identity"           = 8L,
        "poisson:sqrt"               = 9L,
        "Gamma:log"                  = 10L,
        "Gamma:inverse"              = 11L,
        "Gamma:identity"             = 12L,
        "inverse.gaussian:1/mu^2"    = 13L,
        "inverse.gaussian:log"       = 14L,
        "inverse.gaussian:identity"  = 15L,
        "inverse.gaussian:inverse"   = 16L,
        "negbin:log"                 = 17L,
        "negbin:sqrt"                = 18L,
        "negbin:identity"            = 19L,
        "Tweedie:log"                = 20L,
        "Tweedie:identity"           = 21L,
        "Tweedie:inverse"            = 22L,
        "Tweedie:sqrt"               = 23L,
        -1L)
    code
}

# Extract parameters needed by params-aware families.  Returns a length-3
# numeric vector (theta, var.power, link.power); inert defaults are used
# for params-free families.
family_params <- function(family) {
    out <- c(theta = 1.0, var_power = 0.0, link_power = 0.0)
    if (is.null(family) || is.null(family$family)) return(out)

    fam <- family$family

    # Negative binomial: theta is encoded in the family$family string,
    # e.g. "Negative Binomial(2.34)".  family$theta is also set by
    # MASS::negative.binomial() / fastglm's negbin().
    if (grepl("^Negative Binomial", fam)) {
        if (!is.null(family$theta) && is.finite(family$theta) && family$theta > 0) {
            out["theta"] <- family$theta
        } else {
            m <- regmatches(fam, regexec("\\(([^)]+)\\)", fam))[[1]]
            if (length(m) >= 2L) {
                v <- suppressWarnings(as.numeric(m[2]))
                if (is.finite(v) && v > 0) out["theta"] <- v
            }
        }
    }

    if (fam == "Tweedie") {
        vp <- tweedie_var_power(family)
        if (is.numeric(vp) && length(vp) == 1L && is.finite(vp))
            out["var_power"]  <- vp
        lp <- tweedie_link_power(family)
        if (is.numeric(lp) && length(lp) == 1L && is.finite(lp))
            out["link_power"] <- lp
    }

    out
}

# statmod::tweedie() hides var.power / link.power inside the closure
# environment of family$variance / family$linkfun.  Extract them defensively;
# fall back to parsing the link string ("mu^0" -> 0).
tweedie_var_power <- function(family) {
    if (!is.null(family$var.power))
        return(family$var.power)
    e <- tryCatch(environment(family$variance), error = function(e) NULL)
    if (!is.null(e) && exists("var.power", envir = e, inherits = FALSE))
        return(get("var.power", envir = e, inherits = FALSE))
    if (!is.null(e) && exists("p", envir = e, inherits = FALSE))
        return(get("p", envir = e, inherits = FALSE))
    NA_real_
}

tweedie_link_power <- function(family) {
    if (!is.null(family$link.power))
        return(family$link.power)
    e <- tryCatch(environment(family$linkfun), error = function(e) NULL)
    if (!is.null(e) && exists("link.power", envir = e, inherits = FALSE))
        return(get("link.power", envir = e, inherits = FALSE))
    # Fallback: parse the link string.  statmod sets it to e.g. "mu^0".
    if (!is.null(family$link) && grepl("^mu\\^", family$link)) {
        v <- suppressWarnings(as.numeric(sub("^mu\\^", "", family$link)))
        if (is.finite(v)) return(v)
    }
    NA_real_
}

#' Fast generalized linear model fitting
#'
#' @param x input model matrix. Must be a matrix object
#' @param y numeric response vector of length nobs.
#' @param family a description of the error distribution and link function to be used in the model.
#' For \code{fastglmPure} this can only be the result of a call to a family function.
#' (See \code{\link[stats]{family}} for details of family functions.)
#' @param weights an optional vector of 'prior weights' to be used in the fitting process. Should be a numeric vector.
#' @param offset this can be used to specify an a priori known component to be included in the linear predictor during fitting. 
#' This should be a numeric vector of length equal to the number of cases
#' @param start starting values for the parameters in the linear predictor.
#' @param etastart starting values for the linear predictor.
#' @param mustart values for the vector of means.
#' @param method an integer scalar with value 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition,
#' 2 for the LLT Cholesky, 3 for the LDLT Cholesky, 4 for the full pivoted QR decomposition, 5 for the Bidiagonal Divide and 
#' Conquer SVD
#' @param tol threshold tolerance for convergence. Should be a positive real number
#' @param maxit maximum number of IRLS iterations. Should be an integer
#' @param firth logical; if `TRUE` apply Firth's (1993) bias-reducing penalty
#'   to the score function. Currently supported only for
#'   `family = binomial(link = "logit")` on dense `x`. The penalty modifies
#'   the working response by `h_i (0.5 - mu_i) / (mu_i (1 - mu_i))` where
#'   `h_i` is the leverage; convergence is checked on the sup-norm of the
#'   coefficient update. See `logistf::logistf()` for the canonical
#'   reference implementation.
#' @return A list with the elements
#' \item{coefficients}{a vector of coefficients}
#' \item{se}{a vector of the standard errors of the coefficient estimates}
#' \item{rank}{a scalar denoting the computed rank of the model matrix}
#' \item{df.residual}{a scalar denoting the degrees of freedom in the model}
#' \item{residuals}{the vector of residuals}
#' \item{s}{a numeric scalar - the root mean square for residuals}
#' \item{fitted.values}{the vector of fitted values}
#' @seealso [fastglm_fit()]
#' @export
#' @examples
#'
#' set.seed(1)
#' x <- matrix(rnorm(1000 * 25), ncol = 25)
#' eta <- 0.1 + 0.25 * x[,1] - 0.25 * x[,3] + 0.75 * x[,5] -0.35 * x[,6] #0.25 * x[,1] - 0.25 * x[,3]
#' y <- 1 * (eta > rnorm(1000))
#' 
#' yp <- rpois(1000, eta ^ 2)
#' yg <- rgamma(1000, exp(eta) * 1.75, 1.75)
#' 
#' # binomial
#' system.time(gl1 <- glm.fit(x, y, family = binomial()))
#' 
#' system.time(gf1 <- fastglmPure(x, y, family = binomial(), tol = 1e-8))
#' 
#' system.time(gf2 <- fastglmPure(x, y, family = binomial(), method = 1, tol = 1e-8))
#' 
#' system.time(gf3 <- fastglmPure(x, y, family = binomial(), method = 2, tol = 1e-8))
#' 
#' system.time(gf4 <- fastglmPure(x, y, family = binomial(), method = 3, tol = 1e-8))
#' 
#' max(abs(coef(gl1) - gf1$coef))
#' max(abs(coef(gl1) - gf2$coef))
#' max(abs(coef(gl1) - gf3$coef))
#' max(abs(coef(gl1) - gf4$coef))
#' 
#' # poisson
#' system.time(gl1 <- glm.fit(x, yp, family = poisson(link = "log")))
#' 
#' system.time(gf1 <- fastglmPure(x, yp, family = poisson(link = "log"), tol = 1e-8))
#' 
#' system.time(gf2 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 1, tol = 1e-8))
#' 
#' system.time(gf3 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 2, tol = 1e-8))
#' 
#' system.time(gf4 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 3, tol = 1e-8))
#' 
#' max(abs(coef(gl1) - gf1$coef))
#' max(abs(coef(gl1) - gf2$coef))
#' max(abs(coef(gl1) - gf3$coef))
#' max(abs(coef(gl1) - gf4$coef))
#' 
#' # gamma
#' system.time(gl1 <- glm.fit(x, yg, family = Gamma(link = "log")))
#' 
#' system.time(gf1 <- fastglmPure(x, yg, family = Gamma(link = "log"), tol = 1e-8))
#' 
#' system.time(gf2 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 1, tol = 1e-8))
#' 
#' system.time(gf3 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 2, tol = 1e-8))
#' 
#' system.time(gf4 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 3, tol = 1e-8))
#' 
#' max(abs(coef(gl1) - gf1$coef))
#' max(abs(coef(gl1) - gf2$coef))
#' max(abs(coef(gl1) - gf3$coef))
#' max(abs(coef(gl1) - gf4$coef))
#' 
fastglmPure <- function(x, y,
                        family   = gaussian(),
                        weights  = rep(1, NROW(y)),
                        offset   = rep(0, NROW(y)),
                        start    = NULL,
                        etastart = NULL,
                        mustart  = NULL,
                        method   = 0L,
                        tol      = 1e-7,
                        maxit    = 100L,
                        firth    = FALSE)
{
    if (!is.logical(firth) || length(firth) != 1L || is.na(firth))
        stop("'firth' must be TRUE or FALSE.", call. = FALSE)
    if (firth) {
        if (is.null(family$family) || family$family != "binomial" ||
            is.null(family$link) || family$link != "logit")
            stop("'firth = TRUE' currently requires family = binomial(link = \"logit\").",
                 call. = FALSE)
        if (inherits(x, "dgCMatrix") || (requireNamespace("bigmemory", quietly = TRUE) &&
                                          bigmemory::is.big.matrix(x)))
            stop("'firth = TRUE' is not supported for sparse or big.matrix designs.",
                 call. = FALSE)
        # Firth uses the same Cholesky factor for the leverage and the WLS
        # update; force LLT internally regardless of caller's `method`.
        method <- 2L
    }
    weights <- as.vector(weights)
    offset  <- as.vector(offset)
    
    is_sparse_matrix <- inherits(x, "dgCMatrix")
    if (is_sparse_matrix)
    {
        is_big_matrix <- FALSE
        if (method != 2 & method != 3)
        {
            stop("for sparse (dgCMatrix) objects, 'method' must be 2 (LLT) or 3 (LDLT). ",
                 "QR / SVD on sparse matrices is not supported by this package.")
        }
    } else if (is.big.matrix(x))
    {
        is_big_matrix <- TRUE
        if (method != 2 & method != 3)
        {
            stop("for big.matrix objects, 'method' must be 2 (LLT) or 3 (LDLT). ",
                 "QR / SVD methods would force the matrix to be fully read into RAM, ",
                 "defeating the purpose of bigmemory.")
        }
    } else if (is.matrix(x))
    {
        is_big_matrix <- FALSE
    } else
    {
        stop("x must be a matrix, a big.matrix object, or a Matrix::dgCMatrix")
    }
    
    
    stopifnot(is.numeric(y), 
              is.numeric(weights),
              is.numeric(offset),
              NROW(y) == nrow(x),
              NROW(y) == NROW(weights),
              NROW(y) == NROW(offset),
              is.numeric(method),
              is.numeric(tol),
              is.numeric(maxit),
              tol[1] > 0,
              maxit[1] > 0              
              )
    
    nobs  <- n <- NROW(y)
    nvars <- NCOL(x)
    if(is.null(family$family)) 
    {
        print(family)
        stop("'family' not recognized")
    }
    
    if( any(weights < 0) ) stop("negative weights not allowed")
    
    if (method[1] > 5L || method[1] < 0)
    {
        stop("Invalid decomposition method specified. Choose from 0, 1, 2, 3, 4, or 5.")
    }
    
    cnames <- colnames(x)
    
    # from glm
    variance    <- family$variance
    dev.resids  <- family$dev.resids
    aic         <- family$aic
    linkinv     <- family$linkinv
    mu.eta      <- family$mu.eta 
    
    unless.null <- function(x, if.null) if(is.null(x)) if.null else x
    valideta    <- unless.null(family$valideta, function(eta) TRUE)
    validmu     <- unless.null(family$validmu,  function(mu)  TRUE)
    
    
    if(is.null(mustart)) 
    {
        ## calculates mustart and may change y and weights and set n (!)
        eval(family$initialize)
    } else 
    {
        mukeep <- mustart
        eval(family$initialize)
        mustart <- mukeep
    }
    
    y <- as.numeric(y)
    
    coefold <- NULL
    eta <-
        if(!is.null(etastart)) {
            etastart
        } else if(!is.null(start))
            {    
                if (length(start) != nvars)
                {
                    stop(gettextf("length of 'start' should equal %d", nvars),
                         domain = NA)
                } else 
                {
                    coefold <- start
                    offset + as.vector(x %*% start)
                }
            } else family$linkfun(mustart)
    mu <- linkinv(eta)
    
    if (!(validmu(mu) && valideta(eta)))
        stop("cannot find valid starting values: please specify some", call. = FALSE)
    
    if (is.null(start)) start <- rep(0, nvars)
    
    fc <- family_code(family)
    fp <- family_params(family)

    if (is_sparse_matrix)
    {
        res <- fit_sparse_glm(x, drop(y), drop(weights), drop(offset),
                              drop(start), drop(mu), drop(eta),
                              family$variance, family$mu.eta, family$linkinv, family$dev.resids,
                              family$valideta, family$validmu,
                              as.integer(method[1]), as.double(tol[1]), as.integer(maxit[1]),
                              as.integer(fc), fp)
        # Detect intercept-like columns (all entries identical) by max == min.
        col_max <- apply(x, 2, max)
        col_min <- apply(x, 2, min)
        res$intercept <- any(is.int <- (col_max == col_min))
    } else if (!is_big_matrix && firth)
    {
        res <- fit_glm_firth(x, drop(y), drop(weights), drop(offset),
                             drop(start), drop(mu), drop(eta),
                             family$variance, family$mu.eta, family$linkinv, family$dev.resids,
                             family$valideta, family$validmu,
                             as.double(tol[1]), as.integer(maxit[1]),
                             as.integer(fc), fp)
        res$intercept <- any(is.int <- colMax_dense(x) == colMin_dense(x))
    } else if (!is_big_matrix)
    {
        res <- fit_glm(x, drop(y), drop(weights), drop(offset),
                       drop(start), drop(mu), drop(eta),
                       family$variance, family$mu.eta, family$linkinv, family$dev.resids,
                       family$valideta, family$validmu,
                       as.integer(method[1]), as.double(tol[1]), as.integer(maxit[1]),
                       as.integer(fc), fp)

        res$intercept <- any(is.int <- colMax_dense(x) == colMin_dense(x))
    } else
    {
        res <- fit_big_glm(x@address, drop(y), drop(weights), drop(offset),
                           drop(start), drop(mu), drop(eta),
                           family$variance, family$mu.eta, family$linkinv, family$dev.resids,
                           family$valideta, family$validmu,
                           as.integer(method[1]), as.double(tol[1]), as.integer(maxit[1]),
                           as.integer(fc), fp)

        res$intercept <- any(is.int <- big.colMax(x) == big.colMin(x))
    }
    
    if (!res$converged)
    {
        warning("fit_glm: algorithm did not converge", call. = FALSE)
    }

    eps <- 10*.Machine$double.eps
    if (family$family == "binomial") 
    {
        if (any(res$fitted.values > 1 - eps) || any(res$fitted.values < eps))
            warning("fit_glm: fitted probabilities numerically 0 or 1 occurred", call. = FALSE)
    }
    if (family$family == "poisson") 
    {
        if (any(res$fitted.values < eps))
            warning("fit_glm: fitted rates numerically 0 occurred", call. = FALSE)
    }
    
    if (is.null(cnames))
    {
        ncx <- ncol(x)
        if (res$intercept)
        {
            which.int <- which(is.int)
            cnames    <- paste0("X", 1:(ncx - 1) )
            names(res$coefficients) <- 1:ncx
            names(res$coefficients)[-which.int] <- cnames
            names(res$coefficients)[which.int]  <- "(Intercept)"
        } else
        {
            names(res$coefficients) <- paste0("X", 1:ncx)
        }
    } else
    {
        names(res$coefficients) <- cnames
    }
    
    res$family <- family
    res$prior.weights <- weights
    res$y <- y
    res$n <- n
    res$x <- x        # reference to the model matrix; used by vcovHC / vcovCL
    res
}

#' Fast generalized linear model fitting
#'
#' @param x input model matrix. Must be a matrix object 
#' @param y numeric response vector of length nobs.
#' @param family a description of the error distribution and link function to be used in the model. 
#' For \code{fastglm} this can be a character string naming a family function, a family function or the 
#' result of a call to a family function. For \code{fastglmPure} only the third option is supported. 
#' (See \code{\link[stats]{family}} for details of family functions.)
#' @param weights an optional vector of 'prior weights' to be used in the fitting process. Should be a numeric vector.
#' @param offset this can be used to specify an a priori known component to be included in the linear predictor during fitting. 
#' This should be a numeric vector of length equal to the number of cases
#' @param start starting values for the parameters in the linear predictor.
#' @param etastart starting values for the linear predictor.
#' @param mustart values for the vector of means.
#' @param method an integer scalar with value 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition,   
#' 2 for the LLT Cholesky, or 3 for the LDLT Cholesky
#' @param tol threshold tolerance for convergence. Should be a positive real number
#' @param maxit maximum number of IRLS iterations. Should be an integer
#' @param firth logical; if `TRUE` apply Firth's (1993) bias-reducing penalty
#'   to the score function. Currently supported only for
#'   `family = binomial(link = "logit")` on dense `x`. See `logistf::logistf()`
#'   for the canonical reference implementation.
#' @return A list with the elements
#' \item{coefficients}{a vector of coefficients}
#' \item{se}{a vector of the standard errors of the coefficient estimates}
#' \item{rank}{a scalar denoting the computed rank of the model matrix}
#' \item{df.residual}{a scalar denoting the degrees of freedom in the model}
#' \item{residuals}{the vector of residuals}
#' \item{s}{a numeric scalar - the root mean square for residuals}
#' \item{fitted.values}{the vector of fitted values}
#' @seealso [fastglm_fit()]
#' @export
#' @examples
#'
#' x <- matrix(rnorm(10000 * 100), ncol = 100)
#' y <- 1 * (0.25 * x[,1] - 0.25 * x[,3] > rnorm(10000))
#' 
#' system.time(gl1 <- glm.fit(x, y, family = binomial()))
#' 
#' system.time(gf1 <- fastglm(x, y, family = binomial()))
#' 
#' system.time(gf2 <- fastglm(x, y, family = binomial(), method = 1))
#' 
#' system.time(gf3 <- fastglm(x, y, family = binomial(), method = 2))
#' 
#' system.time(gf4 <- fastglm(x, y, family = binomial(), method = 3))
#' 
#' max(abs(coef(gl1) - gf1$coef))
#' max(abs(coef(gl1) - gf2$coef))
#' max(abs(coef(gl1) - gf3$coef))
#' max(abs(coef(gl1) - gf4$coef))
#' 
#' 
#' \dontrun{
#' nrows <- 50000
#' ncols <- 50
#' bkFile <- "bigmat2.bk"
#' descFile <- "bigmatk2.desc"
#' bigmat <- filebacked.big.matrix(nrow=nrows, ncol=ncols, type="double",
#'                                 backingfile=bkFile, backingpath=".",
#'                                 descriptorfile=descFile,
#'                                 dimnames=c(NULL,NULL))
#' for (i in 1:ncols) bigmat[,i] = rnorm(nrows)*i
#' y <- 1*(rnorm(nrows) + bigmat[,1] > 0)
#' 
#' system.time(gfb1 <- fastglm(bigmat, y, family = binomial(), method = 3))
#' }
#'
fastglm <- function(x, ...)
{
    UseMethod("fastglm")
}


#' bigLm default
#'
#' @param ... not used
#' @rdname fastglm
#' @method fastglm default
#' @exportS3Method fastglm default
fastglm.default <- function(x, y,
                            family = gaussian(),
                            weights = NULL,
                            offset = NULL,
                            start    = NULL,
                            etastart = NULL,
                            mustart  = NULL,
                            method = 0L, tol = 1e-8, maxit = 100L,
                            firth = FALSE,
                            ...)
{
    ## family
    if(is.character(family))
    {
        family <- get(family, mode = "function", envir = parent.frame())
    }
    if(is.function(family)) family <- family()
    if(is.null(family$family)) 
    {
        print(family)
        stop("'family' not recognized")
    }
    
    #y             <- as.numeric(y)
    
    ## avoid problems with 1D arrays, but keep names
    if(length(dim(y)) == 1L) 
    {
        nm <- rownames(y)
        dim(y) <- NULL
        if(!is.null(nm)) names(y) <- nm
    }
    
    nobs <- NROW(y)
    
    aic <- family$aic
    
    if (is.null(weights)) weights <- rep(1, nobs)
    if (is.null(offset))  offset  <- rep(0, nobs)
    
    res     <- fastglmPure(x, y, family, weights, offset,
                           start, etastart, mustart,
                           method, tol, maxit, firth = firth)
    y <- res$y
    
    res$residuals <- (y - res$fitted.values) / family$mu.eta(res$linear.predictors)
    #res$y         <- y
    
    # from summary.glm()
    dispersion <-
        if(family$family %in% c("poisson", "binomial"))  1
        else if(res$df.residual > 0)
        {
            est.disp <- TRUE
            if(any(weights == 0))
                warning("observations with zero weight not used for calculating dispersion")
            sum((res$weights*res$residuals ^ 2)[weights > 0]) / res$df.residual
        } else
        {
            est.disp <- TRUE
            NaN
        }
    
    res$dispersion <- dispersion
    
    if (!is.nan(dispersion)) res$se <- res$se * sqrt(dispersion)
    
    wtdmu         <- if (res$intercept) sum(weights * y) / sum(weights) else family$linkinv(offset)
    nulldev       <- sum(family$dev.resids(y, wtdmu, weights))
    
    n.ok          <- nobs - sum(weights == 0)
    nulldf        <- n.ok - as.integer(res$intercept)
    res$df.null   <- nulldf
    
    res$null.deviance <- nulldev
    
    rank <- res$rank
    dev  <- res$deviance
    
    aic.model <- aic(y, res$n, res$fitted.values, res$prior.weights, dev) + 2 * rank
    
    res$aic <- aic.model
    
    # will change later
    boundary <- FALSE
    
    if (boundary)
    {
        warning("fit_glm: algorithm stopped at boundary value", call. = FALSE)
    }
    
    
    res$call      <- match.call()

    class(res)    <- if (isTRUE(res$firth)) c("fastglm_firth", "fastglm") else "fastglm"
    res
}

skip_if_not_installed("MASS")

# Helper: simulate NB(mu, theta) data with a known design
sim_nb_data <- function(n = 500, p = 3, theta = 2.5, seed = 7L) {
    set.seed(seed)
    X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    colnames(X) <- c("(Intercept)", paste0("x", seq_len(p - 1)))
    beta_true <- c(0.3, 0.4, -0.2, 0.25, 0.1)[seq_len(p)]
    eta <- as.vector(X %*% beta_true)
    mu  <- exp(eta)
    y   <- MASS::rnegbin(n, mu = mu, theta = theta)
    list(X = X, y = y, beta_true = beta_true)
}

test_that("native NB log link agrees with glm() on MASS::negative.binomial", {
    for (theta in c(0.5, 2.0, 50)) {
        d   <- sim_nb_data(n = 600, p = 3, theta = theta)
        fam <- MASS::negative.binomial(theta = theta, link = "log")
        gfit <- glm(d$y ~ d$X[, -1], family = fam)
        ffit <- fastglm(d$X, d$y, family = fam, method = 2)
        expect_equal(unname(coef(ffit)), unname(coef(gfit)),
                     tolerance = 1e-8,
                     info = paste0("NB log link, theta = ", theta))
        expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-8)
        # fastglm follows summary.glm()'s convention: NB uses the Pearson
        # dispersion estimate (theta is *known* on the family object, but
        # summary.glm doesn't bake that in).
        expect_equal(ffit$dispersion, summary(gfit)$dispersion,
                     tolerance = 1e-8)
    }
})

test_that("native NB sqrt link agrees with glm()", {
    d   <- sim_nb_data(n = 600, p = 3, theta = 2.5)
    fam <- MASS::negative.binomial(theta = 2.5, link = "sqrt")
    eta_pos <- pmax(d$X %*% c(0.3, 0.4, -0.2), 0.5)
    mu_pos  <- as.vector(eta_pos)^2
    set.seed(11)
    yp <- MASS::rnegbin(length(mu_pos), mu = mu_pos, theta = 2.5)
    gfit <- glm(yp ~ d$X[, -1], family = fam)
    ffit <- fastglm(d$X, yp,    family = fam, method = 2)
    expect_equal(unname(coef(ffit)), unname(coef(gfit)), tolerance = 1e-7)
    expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-7)
})

test_that("built-in negbin() matches MASS::negative.binomial()", {
    d <- sim_nb_data(n = 500, p = 3, theta = 1.7)
    f1 <- fastglm(d$X, d$y, family = negbin(theta = 1.7), method = 2)
    f2 <- fastglm(d$X, d$y, family = MASS::negative.binomial(theta = 1.7),
                  method = 2)
    expect_equal(unname(coef(f1)), unname(coef(f2)), tolerance = 1e-12)
    expect_equal(f1$deviance, f2$deviance, tolerance = 1e-12)
})

test_that("NB native path matches R-callback fallback", {
    # Disguise the family name so family_code() falls back to -1.
    disguise <- function(fam) { fam$family <- paste0(fam$family, "_disguised"); fam }
    d   <- sim_nb_data(n = 400, p = 3, theta = 2)
    fam <- MASS::negative.binomial(theta = 2)
    f_native <- fastglm(d$X, d$y, family = fam,            method = 2)
    f_cb     <- fastglm(d$X, d$y, family = disguise(fam),  method = 2)
    expect_equal(unname(coef(f_native)), unname(coef(f_cb)), tolerance = 1e-8)
    expect_equal(f_native$deviance, f_cb$deviance, tolerance = 1e-8)
})

test_that("NB SE / vcov are unaffected by Pearson-disp override (disp = 1)", {
    d   <- sim_nb_data(n = 500, p = 3, theta = 2)
    fam <- MASS::negative.binomial(theta = 2)
    ffit <- fastglm(d$X, d$y, family = fam, method = 2)
    gfit <- glm(d$y ~ d$X[, -1], family = fam)
    # Coefficient SEs should match those of glm() to high precision.
    expect_equal(unname(ffit$se), unname(sqrt(diag(vcov(gfit)))),
                 tolerance = 1e-7)
})

# quasibinomial / quasipoisson share C++ kernels with binomial / poisson.
# The only behavioural differences are (a) the dispersion is estimated from
# the Pearson statistic instead of being clamped to 1, and (b) summary()
# reports it as such.

test_that("native quasi-poisson matches glm(family = quasipoisson())", {
    set.seed(13)
    n <- 400; p <- 3
    X  <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    eta <- X %*% c(0.3, 0.4, -0.2)
    y   <- rpois(n, exp(eta))
    fam <- quasipoisson()
    gfit <- glm(y ~ X[, -1], family = fam)
    ffit <- fastglm(X, y,    family = fam, method = 2)
    expect_equal(unname(coef(ffit)), unname(coef(gfit)), tolerance = 1e-12)
    expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-10)
    # Pearson-based dispersion should match summary.glm exactly.
    expect_equal(ffit$dispersion, summary(gfit)$dispersion, tolerance = 1e-10)
    # Coefficient SEs are scaled by sqrt(dispersion); compare those too.
    expect_equal(unname(ffit$se), unname(sqrt(diag(vcov(gfit)))),
                 tolerance = 1e-8)
})

test_that("native quasi-binomial matches glm(family = quasibinomial())", {
    set.seed(17)
    n <- 600; p <- 3
    X  <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    eta <- X %*% c(0.2, 0.4, -0.1)
    y   <- rbinom(n, 1, plogis(eta))
    fam <- quasibinomial()
    gfit <- glm(y ~ X[, -1], family = fam)
    ffit <- fastglm(X, y,    family = fam, method = 2)
    expect_equal(unname(coef(ffit)), unname(coef(gfit)), tolerance = 1e-12)
    expect_equal(ffit$deviance, gfit$deviance, tolerance = 1e-10)
    expect_equal(ffit$dispersion, summary(gfit)$dispersion, tolerance = 1e-10)
    expect_equal(unname(ffit$se), unname(sqrt(diag(vcov(gfit)))),
                 tolerance = 1e-8)
})

test_that("over-dispersed quasi-poisson recovers >1 dispersion", {
    # Generate y ~ NB(mu, theta = 2)  so true dispersion ~ 1 + mu/theta > 1.
    skip_if_not_installed("MASS")
    set.seed(23)
    n <- 500; p <- 3
    X  <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    mu <- exp(X %*% c(1, 0.3, -0.2))
    y  <- MASS::rnegbin(n, mu = mu, theta = 2)
    ffit <- fastglm(X, y, family = quasipoisson(), method = 2)
    gfit <- glm(y ~ X[, -1], family = quasipoisson())
    expect_equal(ffit$dispersion, summary(gfit)$dispersion, tolerance = 1e-10)
    expect_gt(ffit$dispersion, 1.05)
})

#' Fit a GLM by streaming row-blocks of the design matrix
#'
#' `fastglm_streaming()` runs an IRLS GLM fit where the design matrix is
#' produced one chunk at a time by a user-supplied closure. It never holds
#' more than a single chunk in memory at once, plus a `p x p` accumulator.
#' This is the front-end for fitting on data sources too large to load
#' completely into RAM (Arrow datasets, Parquet files, DuckDB query results,
#' on-disk CSV streams, etc.).
#'
#' @param chunk_callback a function. Called as `chunk_callback(k)` for
#'   `k = 1, ..., n_chunks`. Must return a list with elements
#'   \describe{
#'     \item{`X`}{an `n_k x p` numeric matrix (chunk of the design matrix).}
#'     \item{`y`}{numeric vector of length `n_k` (response).}
#'     \item{`weights`}{optional; numeric vector of length `n_k` of prior weights.}
#'     \item{`offset`}{optional; numeric vector of length `n_k` of offsets.}
#'   }
#'   Every chunk must have the same number of columns, in the same order. The
#'   closure is called multiple times per IRLS iteration, so it should be
#'   reasonably cheap (e.g. an Arrow scanner that reads from columnar files).
#' @param n_chunks integer; the number of chunks to iterate over.
#' @param family a `family` object describing the error distribution and link.
#'   See [stats::family()].
#' @param start optional length-`p` numeric vector of starting coefficients.
#' @param method integer; `2` for LLT Cholesky (default) or `3` for LDLT
#'   Cholesky. QR / SVD methods are not supported in streaming mode.
#' @param tol convergence tolerance on the relative change in deviance.
#' @param maxit maximum number of IRLS iterations.
#'
#' @returns A list with class `"fastglm"` containing the same elements as
#'   [fastglm()], including `coefficients`, `cov.unscaled`, `deviance`,
#'   `iter`, `converged`, etc. The design matrix is *not* attached, so
#'   `sandwich::vcovHC()` / `sandwich::vcovCL()` will require re-streaming.
#'
#' @details
#' The IRLS loop and step-halving (Marschner 2011) run entirely in C++; the
#' R closure is called only to *deliver* one chunk at a time. For tier-1
#' families (gaussian / binomial / poisson / Gamma / inverse.gaussian on
#' their common links) the family functions are evaluated inline in C++, so
#' the only R round-trip per iteration is the chunk fetch.
#'
#' Standard errors and `cov.unscaled` come from the final `(X' W X)`
#' Cholesky factor, exactly as in the in-memory path.
#'
#' @section Arrow / Parquet recipe:
#'
#' Wrap an Arrow scanner in a closure -- pull each `RecordBatch` as a chunk,
#' build the model matrix, and return it together with the response:
#' `chunks(k)` opens the dataset, scans the `k`-th batch, calls
#' `as.data.frame()`, then returns
#' `list(X = model.matrix(~ x1 + x2, data = tbl), y = tbl$y)`. Pass that
#' closure plus `n_chunks = length(batches)` to `fastglm_streaming()`.
#'
#' The same recipe works for DuckDB (one `dbReadTable` per chunk), CSV
#' streamers, and custom binary formats.
#'
#' @examples
#' # Simulate a "data source" that yields the design matrix in 4 row-blocks.
#' set.seed(1)
#' n <- 1000; p <- 5
#' X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
#' y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.5, -0.3, 0.4, -0.2)))
#' chunk_size <- 250
#' chunks <- function(k) {
#'   idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
#'   list(X = X[idx, , drop = FALSE], y = y[idx])
#' }
#'
#' fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
#' fit_full   <- fastglm(X, y, family = binomial(), method = 2)
#' max(abs(coef(fit_stream) - coef(fit_full)))
#'
#' @export
fastglm_streaming <- function(chunk_callback,
                              n_chunks,
                              family   = gaussian(),
                              start    = NULL,
                              method   = 2L,
                              tol      = 1e-7,
                              maxit    = 100L)
{
    if (!is.function(chunk_callback))
        stop("'chunk_callback' must be a function", call. = FALSE)
    if (!is.numeric(n_chunks) || length(n_chunks) != 1L || n_chunks < 1)
        stop("'n_chunks' must be a positive integer", call. = FALSE)
    n_chunks <- as.integer(n_chunks)
    method <- as.integer(method)
    if (!(method %in% c(2L, 3L)))
        stop("for streaming fits, 'method' must be 2 (LLT) or 3 (LDLT).", call. = FALSE)

    if (is.character(family))   family <- get(family, mode = "function", envir = parent.frame())
    if (is.function(family))    family <- family()
    if (is.null(family$family)) stop("'family' not recognized", call. = FALSE)

    fam_code <- family_code(family)
    fam_par  <- family_params(family)

    # Pull one chunk to discover p and column names, and to do up-front
    # shape validation.  The C++ solver re-pulls all chunks (including this
    # one) on each pass.
    c1 <- chunk_callback(1L)
    if (!is.list(c1) || is.null(c1$X) || is.null(c1$y))
        stop("chunk_callback(1) did not return a list with 'X' and 'y'.",
             call. = FALSE)
    if (!is.matrix(c1$X) && !inherits(c1$X, "Matrix"))
        stop("chunk$X must be a matrix", call. = FALSE)
    if (nrow(c1$X) != length(c1$y))
        stop(sprintf("chunk$X has %d rows but length(y) = %d",
                     nrow(c1$X), length(c1$y)), call. = FALSE)
    p <- ncol(c1$X)
    cnames <- colnames(c1$X)
    rm(c1)

    # Default start = 0_p.
    if (is.null(start)) {
        start_vec <- rep(0, p)
    } else {
        if (length(start) != p)
            stop(sprintf("length(start) = %d does not match number of columns (%d)",
                         length(start), p), call. = FALSE)
        start_vec <- as.numeric(start)
    }

    # R callbacks are only used when fam_code == -1.  For tier-1 families the
    # C++ path never invokes them, but Rcpp still requires Function values.
    valideta <- if (is.null(family$valideta)) function(eta) TRUE else family$valideta
    validmu  <- if (is.null(family$validmu))  function(mu)  TRUE else family$validmu

    res <- fit_streaming_glm(
        chunk_callback = chunk_callback,
        n_chunks       = n_chunks,
        p              = as.integer(p),
        type           = method,
        tol            = tol,
        maxit          = as.integer(maxit),
        fam_code       = as.integer(fam_code),
        var            = family$variance,
        mu_eta         = family$mu.eta,
        linkinv        = family$linkinv,
        dev_resids     = family$dev.resids,
        valideta       = valideta,
        validmu        = validmu,
        start          = start_vec,
        fam_params     = fam_par
    )

    # C++ always reports Pearson-based dispersion; override to 1 for
    # poisson / binomial.  Quasi-binomial / quasi-poisson share C++ family
    # codes with binomial / poisson but keep the estimated dispersion.
    # NB families also use the Pearson estimate (same convention as
    # glm() + MASS::negative.binomial in summary()).
    dispersion <- res$dispersion
    is_fixed_disp <- family$family %in% c("poisson", "binomial")
    if (is_fixed_disp) dispersion <- 1
    if (!identical(dispersion, res$dispersion)) {
        if (is.finite(res$dispersion) && res$dispersion > 0)
            res$se <- res$se / sqrt(res$dispersion)
        if (is.finite(dispersion) && dispersion > 0)
            res$se <- res$se * sqrt(dispersion)
    }

    has_intercept <- isTRUE(res$has_intercept)
    intercept_mask <- as.logical(res$intercept_mask)

    # Names
    if (is.null(cnames)) {
        if (has_intercept) {
            ix <- which(intercept_mask)
            cnames <- character(p)
            cnames[ix]  <- "(Intercept)"
            cnames[-ix] <- paste0("X", seq_len(p - length(ix)))
        } else {
            cnames <- paste0("X", seq_len(p))
        }
    }
    beta <- as.numeric(res$coefficients); names(beta) <- cnames
    se   <- as.numeric(res$se);           names(se)   <- cnames
    cov_unscaled <- res$cov.unscaled
    rownames(cov_unscaled) <- colnames(cov_unscaled) <- cnames

    out <- list(
        coefficients      = beta,
        se                = se,
        cov.unscaled      = cov_unscaled,
        fitted.values     = NULL,
        linear.predictors = NULL,
        deviance          = res$deviance,
        null.deviance     = res$null.deviance,
        df.null           = res$df.null,
        df.residual       = res$df.residual,
        rank              = res$rank,
        iter              = res$iter,
        converged         = res$converged,
        family            = family,
        dispersion        = dispersion,
        n                 = res$n,
        intercept         = has_intercept,
        prior.weights     = NULL,
        y                 = NULL,
        x                 = NULL,
        call              = match.call()
    )
    class(out) <- "fastglm"
    out
}

#' Zero-inflated Poisson / Negative-Binomial regression
#'
#' Fits a zero-inflated count model: a binary inflation component combined
#' with a Poisson or NB count component, where the count component is the
#' *original* (non-truncated) distribution.  Excess zeros above what the
#' count model implies are absorbed by the inflation component.
#'
#' Estimation uses an EM algorithm with all numerical loops in C++.  The
#' E-step computes the posterior `tau_i = P(Z_i = 1 | y_i)` analytically;
#' M-steps fit the inflation logit/probit/cloglog/log and the Poisson/NB
#' count regression via the existing native fastglm IRLS (with
#' weights `w_i (1 - tau_i)` on the count side).  For NB, an inner Brent
#' MLE re-estimates `theta` after the count-side beta step.  Final vcov
#' comes from the numerical Jacobian of the analytical observed score at
#' the EM fixed point (block-structured for `(gamma, beta, theta)`).
#'
#' This is the fastglm analogue of [pscl::zeroinfl()] -- coefficients
#' agree to 1e-5 on standard datasets (slightly looser than hurdle to
#' allow for the EM iteration), with all numerical loops in C++.
#'
#' @param formula a [Formula::Formula] of the form `y ~ x1 + x2 | z1 + z2`
#'   where the right-hand side after `|` specifies the inflation design.
#'   If `|` is absent, the same RHS is used for both parts.
#' @param data optional data frame / environment.
#' @param subset,na.action standard model-frame arguments.
#' @param weights optional non-negative prior weights.
#' @param offset optional vector or matrix.  If a 2-column matrix, columns
#'   are taken as `(count_offset, zero_offset)`.
#' @param dist count component distribution: `"poisson"` or `"negbin"`.
#' @param link character, link for the inflation binomial.  One of
#'   `"logit"` (default), `"probit"`, `"cloglog"`, `"log"`.
#' @param init.theta optional positive scalar starting value for the NB
#'   dispersion.  If `NULL`, a method-of-moments pilot is used.
#' @param em.tol,em.maxit EM convergence tolerance and iteration cap.
#' @param tol,maxit IRLS-loop convergence tolerance and iteration cap
#'   (used inside each M-step).
#' @param theta.tol,theta.maxit Brent-loop tolerance / iteration cap for
#'   the inner theta MLE (only used when `dist = "negbin"`).
#' @param model logical; keep the model frame on the result.
#' @param y,x logical; keep the response / design matrices on the result.
#'
#' @returns A list of class `c("fastglm_zi", "fastglm")` with elements
#'   `coefficients` (a list with `count` and `zero`), `se` (likewise),
#'   `vcov` (full, including theta if NB), `loglik`, `theta`, `SE.theta`,
#'   `tau` (posterior P(Z=1|y)), `em_iter`, `converged`.
#'
#' @examples
#' set.seed(1)
#' n <- 400
#' x  <- rnorm(n)
#' eta_count <- 0.8 + 0.4*x
#' lam <- exp(eta_count)
#' eta_zero <- -0.6 + 0.5*x
#' p_inflate <- plogis(eta_zero)
#' z <- rbinom(n, 1, p_inflate)
#' y <- ifelse(z == 1, 0, rpois(n, lam))
#' df <- data.frame(y = y, x = x)
#' fit <- fastglm_zi(y ~ x, data = df, dist = "poisson")
#' coef(fit, model = "count")
#' coef(fit, model = "zero")
#'
#' @export
fastglm_zi <- function(formula, data, subset, na.action,
                       weights, offset,
                       dist        = c("poisson", "negbin"),
                       link        = c("logit", "probit", "cloglog", "log"),
                       init.theta  = NULL,
                       em.tol      = 1e-8,
                       em.maxit    = 100L,
                       tol         = 1e-9,
                       maxit       = 100L,
                       theta.tol   = 1e-8,
                       theta.maxit = 100L,
                       model = TRUE, y = TRUE, x = FALSE)
{
    if (!requireNamespace("Formula", quietly = TRUE))
        stop("Package 'Formula' is required for fastglm_zi(); install it.",
             call. = FALSE)

    dist <- match.arg(dist)
    link <- match.arg(link)

    cl <- match.call()
    if (missing(data)) data <- environment(formula)
    f  <- Formula::Formula(formula)
    has_two_rhs <- length(f)[2] >= 2

    mf <- match.call(expand.dots = FALSE)
    mf$formula <- f
    mf$dist <- mf$link <- mf$init.theta <- NULL
    mf$em.tol <- mf$em.maxit <- NULL
    mf$tol <- mf$maxit <- mf$theta.tol <- mf$theta.maxit <- NULL
    mf$model <- mf$y <- mf$x <- NULL
    mf[[1L]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())

    mt_count <- terms(f, data = data, rhs = 1L)
    mt_zero  <- if (has_two_rhs) terms(f, data = data, rhs = 2L) else mt_count

    Y_resp <- model.response(mf, "numeric")
    if (!is.numeric(Y_resp) || any(Y_resp < 0))
        stop("'y' must be non-negative.", call. = FALSE)
    n <- length(Y_resp)

    X_count <- model.matrix(mt_count, mf)
    Z_zero  <- model.matrix(mt_zero,  mf)

    wt <- model.weights(mf)
    if (is.null(wt)) wt <- rep(1, n)
    if (!is.numeric(wt) || any(wt < 0))
        stop("'weights' must be non-negative.", call. = FALSE)

    off <- model.offset(mf)
    if (is.null(off)) {
        off_count <- rep(0, n); off_zero <- rep(0, n)
    } else if (is.matrix(off) && ncol(off) == 2L) {
        off_count <- as.numeric(off[, 1L]); off_zero <- as.numeric(off[, 2L])
    } else {
        off_count <- as.numeric(off);       off_zero <- as.numeric(off)
    }

    zero_family <- switch(link,
                          "logit"   = binomial("logit"),
                          "probit"  = binomial("probit"),
                          "cloglog" = binomial("cloglog"),
                          "log"     = binomial("log"))
    zero_fam_code <- switch(link,
                            "logit"   = 3L,
                            "probit"  = 4L,
                            "cloglog" = 5L,
                            "log"     = 6L)
    dist_code <- switch(dist, "poisson" = 0L, "negbin" = 1L)

    count_family <- if (dist == "poisson") poisson("log") else
                    # MASS::negative.binomial(theta) provides the family
                    # callbacks we need; use a dummy theta -- the C++ side
                    # honours fam_code/fam_params directly so the value here
                    # only matters as a fallback.
                    {
                        if (!requireNamespace("MASS", quietly = TRUE))
                            stop("Package 'MASS' is required for dist='negbin' fastglm_zi().",
                                 call. = FALSE)
                        MASS::negative.binomial(theta = 1)
                    }

    init_th <- if (!is.null(init.theta)) {
        if (!is.numeric(init.theta) || length(init.theta) != 1L ||
            !is.finite(init.theta) || init.theta <= 0)
            stop("'init.theta' must be a positive finite scalar.", call. = FALSE)
        init.theta
    } else {
        -1
    }

    res <- fit_glm_zi(
        x_count            = X_count,
        z_zero             = Z_zero,
        y                  = as.numeric(Y_resp),
        weights            = wt,
        offset_count       = off_count,
        offset_zero        = off_zero,
        dist_code          = dist_code,
        zero_fam_code      = zero_fam_code,
        init_theta         = as.double(init_th),
        tol                = as.double(tol),
        maxit              = as.integer(maxit),
        em_tol             = as.double(em.tol),
        em_maxit           = as.integer(em.maxit),
        theta_tol          = as.double(theta.tol),
        theta_maxit        = as.integer(theta.maxit),
        var_fun_zero       = zero_family$variance,
        mu_eta_fun_zero    = zero_family$mu.eta,
        linkinv_fun_zero   = zero_family$linkinv,
        dev_resids_fun_zero= zero_family$dev.resids,
        valideta_fun_zero  = zero_family$valideta,
        validmu_fun_zero   = zero_family$validmu,
        var_fun_count      = count_family$variance,
        mu_eta_fun_count   = count_family$mu.eta,
        linkinv_fun_count  = count_family$linkinv,
        dev_resids_fun_count = count_family$dev.resids,
        valideta_fun_count = count_family$valideta,
        validmu_fun_count  = count_family$validmu)

    cn_count <- colnames(X_count); if (is.null(cn_count)) cn_count <- paste0("xc", seq_len(ncol(X_count)))
    cn_zero  <- colnames(Z_zero);  if (is.null(cn_zero))  cn_zero  <- paste0("xz", seq_len(ncol(Z_zero)))
    names(res$coefficients_count) <- cn_count
    names(res$coefficients_zero)  <- cn_zero
    names(res$se_count) <- cn_count
    names(res$se_zero)  <- cn_zero
    if (!is.null(res$vcov_count)) dimnames(res$vcov_count) <- list(cn_count, cn_count)
    if (!is.null(res$vcov_zero))  dimnames(res$vcov_zero)  <- list(cn_zero,  cn_zero)

    coefficients <- list(count = res$coefficients_count,
                         zero  = res$coefficients_zero)
    se           <- list(count = res$se_count, zero = res$se_zero)

    p_c <- length(coefficients$count); p_z <- length(coefficients$zero)
    pp  <- p_c + p_z + (dist == "negbin")
    rn <- c(paste0("count_", cn_count), paste0("zero_", cn_zero))
    if (dist == "negbin") rn <- c(rn, "theta")
    if (NROW(res$vcov_full) == pp) {
        rownames(res$vcov_full) <- colnames(res$vcov_full) <- rn
    }

    out <- list(
        coefficients   = coefficients,
        se             = se,
        vcov           = res$vcov_full,
        loglik         = res$loglik,
        theta          = res$theta,
        SE.theta       = res[["SE.theta"]],
        tau            = res$tau,
        em_iter        = res$em_iter,
        converged      = res$converged,
        n              = res$n,
        df.residual    = n - res$df,
        dist           = dist,
        link           = link,
        formula        = f,
        terms          = list(count = mt_count, zero = mt_zero),
        levels         = .getXlevels(mt_count, mf),
        contrasts      = list(count = attr(X_count, "contrasts"),
                              zero  = attr(Z_zero,  "contrasts")),
        weights        = wt,
        offset         = list(count = off_count, zero = off_zero),
        eta_count      = res$eta_count,
        mu_count       = res$mu_count,
        eta_zero       = res$eta_zero,
        call           = cl
    )
    if (model) out$model <- mf
    if (y)     out$y     <- Y_resp
    if (x)     out$x     <- list(count = X_count, zero = Z_zero)
    class(out) <- c("fastglm_zi", "fastglm")
    out
}

#' @export
coef.fastglm_zi <- function(object, model = c("full", "count", "zero"), ...)
{
    model <- match.arg(model)
    if (model == "full") {
        cn <- names(object$coefficients$count)
        zn <- names(object$coefficients$zero)
        out <- c(object$coefficients$count, object$coefficients$zero)
        names(out) <- c(paste0("count_", cn), paste0("zero_", zn))
        return(out)
    }
    object$coefficients[[model]]
}

#' @export
vcov.fastglm_zi <- function(object, model = c("full", "count", "zero"), ...)
{
    model <- match.arg(model)
    if (model == "full") return(object$vcov)
    p_c <- length(object$coefficients$count)
    p_z <- length(object$coefficients$zero)
    if (model == "count") return(object$vcov[seq_len(p_c), seq_len(p_c), drop = FALSE])
    object$vcov[p_c + seq_len(p_z), p_c + seq_len(p_z), drop = FALSE]
}

#' @export
logLik.fastglm_zi <- function(object, ...)
{
    df <- length(object$coefficients$count) + length(object$coefficients$zero) +
          (object$dist == "negbin")
    structure(object$loglik, df = df, nobs = object$n,
              class = "logLik")
}

#' @export
print.fastglm_zi <- function(x, digits = max(3, getOption("digits") - 3), ...)
{
    cat("\nfastglm_zi:  zero-inflated ", x$dist, " (zero link = ", x$link,
        ")\n", sep = "")
    cat("Call:  "); print(x$call)
    cat("\nCount model coefficients (", x$dist, "):\n", sep = "")
    print(format(x$coefficients$count, digits = digits), quote = FALSE)
    cat("\nZero-inflation coefficients (binomial,", x$link, "):\n")
    print(format(x$coefficients$zero, digits = digits), quote = FALSE)
    if (x$dist == "negbin")
        cat("\nTheta:", format(x$theta, digits = digits),
            "  SE:", format(x[["SE.theta"]], digits = digits), "\n")
    cat("Log-likelihood:", format(x$loglik, digits = digits),
        "on", length(x$coefficients$count) + length(x$coefficients$zero) +
             (x$dist == "negbin"), "df  (EM iter:", x$em_iter, ")\n")
    invisible(x)
}

skip_if_not_installed("pscl")

# fastglm_zi vs pscl::zeroinfl on simulated data and the canonical
# bioChemists data shipped with pscl.  Tolerances are slightly looser
# than hurdle (1e-4 vs 1e-5) because both fastglm_zi and pscl::zeroinfl
# end up at the same MLE but via different optimization paths -- pscl
# uses optim/BFGS on the full joint likelihood, fastglm_zi uses EM with
# an inner Brent for theta.

.tight_pscl <- function() pscl::zeroinfl.control(method = "BFGS",
                                                 reltol = 1e-12,
                                                 maxit  = 5000)

test_that("Poisson ZI: coefs match pscl::zeroinfl on simulated data", {
    set.seed(21)
    n  <- 500
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
    lam   <- exp(eta_c)
    eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
    p_inf <- plogis(eta_z)
    z     <- rbinom(n, 1, p_inf)
    y     <- ifelse(z == 1, 0L, rpois(n, lam))
    df    <- data.frame(y = y, x1 = x1, x2 = x2)

    f  <- fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson",
                     em.tol = 1e-10, em.maxit = 200L,
                     tol = 1e-10, maxit = 200L)
    p  <- pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson",
                         control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-5)
    expect_true(f$converged)
})

test_that("Poisson ZI: SEs agree with pscl::zeroinfl", {
    set.seed(22)
    n  <- 600
    x  <- rnorm(n)
    eta_c <- 0.5 + 0.3 * x
    eta_z <- -0.2 + 0.4 * x
    z     <- rbinom(n, 1, plogis(eta_z))
    y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
    df    <- data.frame(y = y, x = x)

    f <- fastglm_zi(y ~ x, data = df, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 200L,
                    tol = 1e-10, maxit = 200L)
    p <- pscl::zeroinfl(y ~ x, data = df, dist = "poisson",
                        control = .tight_pscl())

    expect_equal(unname(f$se$count), unname(sqrt(diag(vcov(p)))[1:2]),
                 tolerance = 1e-2)
    expect_equal(unname(f$se$zero),  unname(sqrt(diag(vcov(p)))[3:4]),
                 tolerance = 1e-2)
})

test_that("ZI handles two-RHS formula y ~ x | z (different zero design)", {
    set.seed(23)
    n  <- 500
    x1 <- rnorm(n);  x2 <- rnorm(n);  z1 <- rnorm(n)
    eta_c <- 0.6 + 0.3 * x1 - 0.2 * x2
    eta_z <- -0.3 + 0.7 * z1
    z     <- rbinom(n, 1, plogis(eta_z))
    y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
    df    <- data.frame(y = y, x1 = x1, x2 = x2, z1 = z1)

    f <- fastglm_zi(y ~ x1 + x2 | z1, data = df, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 200L,
                    tol = 1e-10, maxit = 200L)
    p <- pscl::zeroinfl(y ~ x1 + x2 | z1, data = df, dist = "poisson",
                        control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    expect_length(coef(f, "zero"), 2L)         # intercept + z1
    expect_length(coef(f, "count"), 3L)        # intercept + x1 + x2
})

test_that("NB ZI: coefs and theta match pscl::zeroinfl", {
    skip_if_not_installed("MASS")
    set.seed(24)
    n  <- 800
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.8 + 0.4 * x1 - 0.3 * x2
    lam   <- exp(eta_c)
    eta_z <- -0.4 + 0.5 * x1
    z     <- rbinom(n, 1, plogis(eta_z))
    y     <- ifelse(z == 1, 0L,
                    MASS::rnegbin(n, mu = lam, theta = 2.0))
    df    <- data.frame(y = y, x1 = x1, x2 = x2)

    f <- fastglm_zi(y ~ x1 + x2, data = df, dist = "negbin",
                    em.tol = 1e-9, em.maxit = 300L,
                    tol = 1e-10, maxit = 200L,
                    theta.tol = 1e-9)
    p <- pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "negbin",
                        control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 5e-3)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 5e-3)
    expect_equal(unname(f$theta), unname(p$theta), tolerance = 5e-2)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-4)
})

test_that("ZI Poisson: probit zero link works", {
    set.seed(25)
    n  <- 400
    x  <- rnorm(n)
    eta_c <- 0.5 + 0.4 * x
    eta_z <- -0.3 + 0.6 * x
    z     <- rbinom(n, 1, pnorm(eta_z))
    y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
    df    <- data.frame(y = y, x = x)

    f <- fastglm_zi(y ~ x, data = df, dist = "poisson", link = "probit",
                    em.tol = 1e-10, em.maxit = 200L,
                    tol = 1e-10, maxit = 200L)
    p <- pscl::zeroinfl(y ~ x, data = df, dist = "poisson", link = "probit",
                        control = .tight_pscl())

    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-3)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-3)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-4)
})

test_that("pscl::zeroinfl on bioChemists matches fastglm_zi", {
    skip_if_not_installed("pscl")
    data("bioChemists", package = "pscl")
    f <- fastglm_zi(art ~ ., data = bioChemists, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 300L,
                    tol = 1e-10, maxit = 200L)
    p <- pscl::zeroinfl(art ~ ., data = bioChemists, dist = "poisson",
                        control = .tight_pscl())
    expect_equal(unname(coef(f, "count")), unname(p$coefficients$count),
                 tolerance = 1e-4)
    expect_equal(unname(coef(f, "zero")),  unname(p$coefficients$zero),
                 tolerance = 1e-4)
    expect_equal(as.numeric(logLik(f)), as.numeric(logLik(p)),
                 tolerance = 1e-5)
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

colMax_dense <- function(A) {
    .Call(`_fastglm_colMax_dense`, A)
}

colMin_dense <- function(A) {
    .Call(`_fastglm_colMin_dense`, A)
}

fit_glm <- function(x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code = -1L, fam_params = NULL) {
    .Call(`_fastglm_fit_glm`, x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code, fam_params)
}

fit_big_glm <- function(x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code = -1L, fam_params = NULL) {
    .Call(`_fastglm_fit_big_glm`, x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code, fam_params)
}

fit_sparse_glm <- function(x_sparse, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code = -1L, fam_params = NULL) {
    .Call(`_fastglm_fit_sparse_glm`, x_sparse, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, type, tol, maxit, fam_code, fam_params)
}

fit_glm_firth <- function(x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, tol, maxit, fam_code, fam_params = NULL) {
    .Call(`_fastglm_fit_glm_firth`, x, y, weights, offset, start, mu, eta, var, mu_eta, linkinv, dev_resids, valideta, validmu, tol, maxit, fam_code, fam_params)
}

fit_glm_hurdle <- function(x_count, z_zero, y, weights, offset_count, offset_zero, dist_code, zero_fam_code, init_theta, tol, maxit, outer_tol, outer_maxit, theta_tol, theta_maxit, var_fun_zero, mu_eta_fun_zero, linkinv_fun_zero, dev_resids_fun_zero, valideta_fun_zero, validmu_fun_zero) {
    .Call(`_fastglm_fit_glm_hurdle`, x_count, z_zero, y, weights, offset_count, offset_zero, dist_code, zero_fam_code, init_theta, tol, maxit, outer_tol, outer_maxit, theta_tol, theta_maxit, var_fun_zero, mu_eta_fun_zero, linkinv_fun_zero, dev_resids_fun_zero, valideta_fun_zero, validmu_fun_zero)
}

fit_glm_nb <- function(x, y, weights, offset, start, mu_init, eta_init, var_fun, mu_eta_fun, linkinv_fun, dev_resids_fun, valideta_fun, validmu_fun, type, tol, maxit, fam_code, init_theta, theta_tol, theta_maxit, outer_maxit, outer_tol) {
    .Call(`_fastglm_fit_glm_nb`, x, y, weights, offset, start, mu_init, eta_init, var_fun, mu_eta_fun, linkinv_fun, dev_resids_fun, valideta_fun, validmu_fun, type, tol, maxit, fam_code, init_theta, theta_tol, theta_maxit, outer_maxit, outer_tol)
}

fit_glm_zi <- function(x_count, z_zero, y, weights, offset_count, offset_zero, dist_code, zero_fam_code, init_theta, tol, maxit, em_tol, em_maxit, theta_tol, theta_maxit, var_fun_zero, mu_eta_fun_zero, linkinv_fun_zero, dev_resids_fun_zero, valideta_fun_zero, validmu_fun_zero, var_fun_count, mu_eta_fun_count, linkinv_fun_count, dev_resids_fun_count, valideta_fun_count, validmu_fun_count) {
    .Call(`_fastglm_fit_glm_zi`, x_count, z_zero, y, weights, offset_count, offset_zero, dist_code, zero_fam_code, init_theta, tol, maxit, em_tol, em_maxit, theta_tol, theta_maxit, var_fun_zero, mu_eta_fun_zero, linkinv_fun_zero, dev_resids_fun_zero, valideta_fun_zero, validmu_fun_zero, var_fun_count, mu_eta_fun_count, linkinv_fun_count, dev_resids_fun_count, valideta_fun_count, validmu_fun_count)
}

fit_streaming_glm <- function(chunk_callback, n_chunks, p, type, tol, maxit, fam_code, var, mu_eta, linkinv, dev_resids, valideta, validmu, start = NULL, fam_params = NULL) {
    .Call(`_fastglm_fit_streaming_glm`, chunk_callback, n_chunks, p, type, tol, maxit, fam_code, var, mu_eta, linkinv, dev_resids, valideta, validmu, start, fam_params)
}

big.crossprod <- function(x)
{
    if (!is.big.matrix(x))
    {
        stop("object must be a big.matrix object")
    }
    .Call("crossprod_big", x@address)
}

big.colSums <- function(x)
{
    if (!is.big.matrix(x))
    {
        stop("object must be a big.matrix object")
    }
    .Call("colsums_big", x@address)
}

big.colMax <- function(x)
{
    if (!is.big.matrix(x))
    {
        stop("object must be a big.matrix object")
    }
    .Call("colmax_big", x@address)
}

big.colMin <- function(x)
{
    if (!is.big.matrix(x))
    {
        stop("object must be a big.matrix object")
    }
    .Call("colmin_big", x@address)
}


#' big.matrix prod
#'
#' @param x big.matrix
#' @param y numeric vector
#' @docType methods
#' @import methods
#' @importFrom methods new
#' @importFrom bigmemory as.big.matrix is.big.matrix
#' @rdname extract-methods
setMethod("%*%",signature(x="big.matrix", y="vector"),
          function(x, y)
          {
              if(dim(x)[2] != length(y)) stop("non-conformant matrices")
              return( .Call("prod_vec_big", x@address, y) )
          },
          valueClass="vector"
)

#' big.matrix prod
#'
#' @docType methods
#' @rdname extract-methods
setMethod("%*%",signature(x="vector", y="big.matrix"),
          function(x, y)
          {
              if(dim(y)[2] != length(x)) stop("non-conformant matrices")
              return( .Call("prod_vec_big_right", x, y@address) )
          },
          valueClass="vector"
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)

## -----------------------------------------------------------------------------
library(fastglm)

## -----------------------------------------------------------------------------
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

fit <- fastglm(x, y, family = binomial(link = "cloglog"))
summary(fit)

## -----------------------------------------------------------------------------
fit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) + unclass(alcgp),
            data = esoph, family = binomial(link = "cloglog"),
            method = fastglm_fit)

## -----------------------------------------------------------------------------
set.seed(123)
n <- 5000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rbinom(n, 1, plogis(x %*% rnorm(p) * 0.05))

system.time(f0 <- fastglm(x, y, family = binomial()))                 # default
system.time(f2 <- fastglm(x, y, family = binomial(), method = 2))     # LLT
system.time(f3 <- fastglm(x, y, family = binomial(), method = 3))     # LDLT

## -----------------------------------------------------------------------------
fit <- fastglm(x, y, family = binomial(), method = 2)

V    <- vcov(fit)
dim(V)

## standard errors agree with the diagonal of vcov()
all.equal(sqrt(diag(V)), unname(coef(summary(fit))[, "Std. Error"]))

## -----------------------------------------------------------------------------
library(sandwich)

V_hc <- vcovHC(fit, type = "HC0")
V_hc[1:3, 1:3]

cluster <- rep(seq_len(20), length.out = n)
V_cl    <- vcovCL(fit, cluster = cluster, type = "HC1")
V_cl[1:3, 1:3]

## -----------------------------------------------------------------------------
xnew <- x[1:5, , drop = FALSE]
predict(fit, newdata = xnew, type = "response", se.fit = TRUE)

## -----------------------------------------------------------------------------
n <- 4000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.4, -0.2, 0.3)))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))

## ----eval = TRUE--------------------------------------------------------------
library(microbenchmark)

set.seed(7)
n <- 50000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rpois(n, exp(x %*% rnorm(p) * 0.05 + 1))

## force the R-callback path by giving the family object an unrecognised name
disguise <- function(fam) { fam$family <- paste0(fam$family, "*"); fam }

mb <- microbenchmark(
    native   = fastglm(x, y, family = poisson(),            method = 2),
    callback = fastglm(x, y, family = disguise(poisson()),  method = 2),
    times = 5L
)
print(mb)

---
title: "Negative Binomial Convergence: fastglm vs MASS"
author: "Jared Huling"
date: "2026-05-27"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Negative Binomial Convergence: fastglm vs MASS}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



Standard negative-binomial GLM fitting alternates between IRLS for the
regression coefficients $\beta$ (at fixed dispersion $\theta$) and a
one-dimensional MLE update for $\theta$.
Both `MASS::glm.nb()` and `fastglm_nb()` follow this strategy, but their
implementations differ in initialization and numerical handling.  This
vignette compares the two on a deliberately challenging data-generating
process where MASS frequently fails to converge.

## Data-generating process

We use a moderately high-dimensional negative-binomial model
($n = 200$, $p = 9$) with strong alternating effects and true
$\theta = 1$.
The large coefficients produce fitted means spanning many orders of
magnitude, which stresses the IRLS solver when the initial $\theta$
estimate is far from the truth.


``` r
library(fastglm)
library(MASS)

n <- 200
p <- 9
beta_true  <- c(4, 1.5, -1.2, 1.5, -1.2, 1.5, -1.2, 1.5, -1.2)
theta_true <- 1
n_reps     <- 200
```

## Convergence comparison

For each of 200 replications we draw a fresh design matrix and
response, then fit with both `fastglm_nb()` and `MASS::glm.nb()`.  We
record whether each method converges without error or warning, and when
both converge, we compare the maximized log-likelihoods.


``` r
results <- data.frame(
    seed         = integer(n_reps),
    mass_conv    = logical(n_reps),
    fg_conv      = logical(n_reps),
    mass_loglik  = numeric(n_reps),
    fg_loglik    = numeric(n_reps),
    mass_theta   = numeric(n_reps),
    fg_theta     = numeric(n_reps)
)

for (i in seq_len(n_reps)) {
    set.seed(i)
    X  <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
    mu <- exp(X %*% beta_true)
    y  <- MASS::rnegbin(n, mu = mu, theta = theta_true)

    ## MASS::glm.nb
    mass_ok   <- FALSE
    mass_ll   <- NA_real_
    mass_th   <- NA_real_
    tryCatch(
        withCallingHandlers(
            {
                m <- MASS::glm.nb(y ~ X[, -1])
                if (m$converged) {
                    mass_ok <- TRUE
                    mass_ll <- as.numeric(logLik(m))
                    mass_th <- m$theta
                }
            },
            warning = function(w) invokeRestart("muffleWarning")
        ),
        error = function(e) NULL
    )

    ## fastglm_nb
    fg_ok  <- FALSE
    fg_ll  <- NA_real_
    fg_th  <- NA_real_
    tryCatch({
        f <- fastglm_nb(X, y, link = "log")
        if (f$converged && !any(is.nan(coef(f)))) {
            fg_ok <- TRUE
            fg_ll <- as.numeric(logLik(f))
            fg_th <- f$theta
        }
    }, error = function(e) NULL)

    results$seed[i]        <- i
    results$mass_conv[i]   <- mass_ok
    results$fg_conv[i]     <- fg_ok
    results$mass_loglik[i] <- mass_ll
    results$fg_loglik[i]   <- fg_ll
    results$mass_theta[i]  <- mass_th
    results$fg_theta[i]    <- fg_th
}
```

### Summary


``` r
tab <- with(results, data.frame(
    Method     = c("MASS::glm.nb", "fastglm_nb"),
    Converged  = c(sum(mass_conv), sum(fg_conv)),
    Failed     = c(sum(!mass_conv), sum(!fg_conv)),
    Rate       = sprintf("%.1f%%", 100 * c(mean(mass_conv), mean(fg_conv)))
))
knitr::kable(tab, align = "lrrl",
             caption = sprintf("Convergence over %d replications", n_reps))
```



Table: Convergence over 200 replications

|Method       | Converged| Failed|Rate   |
|:------------|---------:|------:|:------|
|MASS::glm.nb |       165|     35|82.5%  |
|fastglm_nb   |       200|      0|100.0% |



`fastglm_nb()` converges on every replication.  `MASS::glm.nb()` fails
on roughly one in six.

### Log-likelihood comparison

When both methods converge they should find the same MLE, so their
log-likelihoods should agree.  In a small number of replications
`MASS::glm.nb()` nominally converges but lands at a degenerate solution
with $\hat\theta > 10^6$ and very poor log-likelihood.  We exclude
these (identified by `fastglm` achieving a strictly higher
log-likelihood by more than 1 unit) to focus on genuine agreement:


``` r
both <- results$mass_conv & results$fg_conv
ll_diff <- results$fg_loglik[both] - results$mass_loglik[both]

n_degenerate <- sum(ll_diff > 1)
agree <- both
agree[both] <- ll_diff <= 1

cat(sprintf("Replications where both converge:         %d / %d\n",
            sum(both), n_reps))
#> Replications where both converge:         165 / 200
cat(sprintf("  of which MASS lands at degenerate MLE:  %d\n", n_degenerate))
#>   of which MASS lands at degenerate MLE:  3
cat(sprintf("  genuine agreement (|diff| <= 1):        %d\n", sum(agree)))
#>   genuine agreement (|diff| <= 1):        162

ll_good <- results$fg_loglik[agree] - results$mass_loglik[agree]
cat(sprintf("\nAmong agreeing fits:\n"))
#> 
#> Among agreeing fits:
cat(sprintf("  Max |logLik difference|:  %.2e\n", max(abs(ll_good))))
#>   Max |logLik difference|:  1.74e-06
cat(sprintf("  Mean |logLik difference|: %.2e\n", mean(abs(ll_good))))
#>   Mean |logLik difference|: 2.20e-08
```


``` r
plot(results$mass_loglik[agree], results$fg_loglik[agree],
     xlab = "MASS log-likelihood",
     ylab = "fastglm log-likelihood",
     main = "Log-likelihood agreement (both converge, excluding degenerate MASS fits)",
     pch = 16, cex = 0.6, col = "#2166ac80")
abline(0, 1, col = "grey40", lty = 2)
```

![plot of chunk loglik-plot](figure/loglik-plot-1.png)

### Theta estimates


``` r
plot(results$mass_theta[agree], results$fg_theta[agree],
     xlab = expression(hat(theta) ~ "(MASS)"),
     ylab = expression(hat(theta) ~ "(fastglm)"),
     main = expression("Dispersion" ~ hat(theta) ~ "agreement"),
     pch = 16, cex = 0.6, col = "#2166ac80")
abline(0, 1, col = "grey40", lty = 2)
```

![plot of chunk theta-plot](figure/theta-plot-1.png)

``` r

cat(sprintf("Max |theta difference|: %.2e\n",
            max(abs(results$fg_theta[agree] - results$mass_theta[agree]))))
#> Max |theta difference|: 4.27e-07
```

### Cases where only fastglm converges


``` r
fg_only <- results$fg_conv & !results$mass_conv
cat(sprintf("fastglm converges, MASS fails: %d replications\n", sum(fg_only)))
#> fastglm converges, MASS fails: 35 replications
if (any(fg_only)) {
    cat(sprintf("  theta range:  [%.3f, %.3f]\n",
                min(results$fg_theta[fg_only]),
                max(results$fg_theta[fg_only])))
    cat(sprintf("  logLik range: [%.1f, %.1f]\n",
                min(results$fg_loglik[fg_only]),
                max(results$fg_loglik[fg_only])))
}
#>   theta range:  [0.861, 1.242]
#>   logLik range: [-1123.1, -935.4]
```

---
title: "Firth Bias-Reduced GLMs with 'fastglm'"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Firth Bias-Reduced GLMs with 'fastglm'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

Firth's (1993) penalized likelihood removes the leading $O(1/n)$ bias from maximum likelihood estimates and produces finite estimates even under separation. *fastglm* implements the general mean-bias reduction of Kosmidis and Firth (2009, 2021), which extends Firth's original logistic penalty to arbitrary GLM families. The method is activated with a single flag:

```{r, eval = FALSE}
fastglm(x, y, family = binomial(), firth = TRUE)
```

The penalty adds $\frac{1}{2} \partial \log |X^\top W X| / \partial \beta$ to the score, which is equivalent to modifying the IRLS working response by

$$
\xi_i = \frac{\phi \, h_i \, \mu''(\eta_i) \, V(\mu_i)}{2 \, w_i \, [\mu'(\eta_i)]^3}
$$

where $h_i$ is the $i$-th diagonal of the hat matrix $H = W^{1/2} X (X^\top W X)^{-1} X^\top W^{1/2}$, $\mu''(\eta)$ is the second derivative of the inverse link, $V(\mu)$ is the variance function, $w_i$ is the prior weight, and $\phi$ is the dispersion (1 for binomial and Poisson families; estimated iteratively for Gaussian, Gamma, and inverse Gaussian). This is the AS_mean adjustment of Kosmidis and Firth (2009), the same method used by `brglm2::brglmFit(type = "AS_mean")`.

The adjustment is computed once per IRLS iteration using the leverages from the weighted least-squares decomposition that *fastglm* already performs, so the overhead relative to an unpenalized fit is modest.

```{r}
library(fastglm)
```

# Logistic regression under separation

The canonical application is logistic regression with (quasi-) separation, where the standard MLE diverges. The `sex2` dataset from *logistf* (Heinze and Schemper, 2002) demonstrates:

```{r}
data(sex2, package = "logistf")
X <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y <- sex2$case

fit_f <- fastglm(X, y, family = binomial(), firth = TRUE)
fit_l <- logistf::logistf(case ~ age + oc + vic + vicl + vis + dia,
                          data = sex2, pl = FALSE, plconf = NULL,
                          control = logistf::logistf.control(
                              lconv = 1e-12, gconv = 1e-12, xconv = 1e-12,
                              maxit = 200L))

cbind(fastglm = unname(coef(fit_f)), logistf = unname(coef(fit_l)))
max(abs(unname(coef(fit_f)) - unname(coef(fit_l))))
```

Coefficients agree to roughly 1e-7. The runtime advantage is substantial because *fastglm*'s Firth solver reuses the C++ IRLS infrastructure:

```{r}
mb <- microbenchmark::microbenchmark(
    fastglm = fastglm(X, y, family = binomial(), firth = TRUE),
    logistf = logistf::logistf(case ~ age + oc + vic + vicl + vis + dia,
                               data = sex2, pl = FALSE, plconf = NULL),
    times = 10L
)
print(mb, unit = "ms")
```

# General GLM families

`firth = TRUE` works with all standard R families and link functions. For each family below, we verify that *fastglm* reproduces the bias-reduced coefficients from `brglm2::brglmFit(type = "AS_mean")` --- both implement the same AS_mean adjustment, so the coefficients should agree to near machine precision.

```{r}
library(brglm2)
set.seed(123)
n <- 500
x1 <- rnorm(n);  x2 <- rnorm(n)
X <- cbind(1, x1, x2)
eta_true <- 0.5 + 0.3 * x1 - 0.2 * x2
```

## Binomial (logit, probit, cloglog)

```{r}
y_bin <- rbinom(n, 1, plogis(eta_true))
df <- data.frame(y = y_bin, x1 = x1, x2 = x2)

for (lnk in c("logit", "probit", "cloglog")) {
    fam <- binomial(lnk)
    f <- fastglm(X, y_bin, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = df, family = fam, method = "brglmFit",
             type = "AS_mean", control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("binomial %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}
```

## Poisson (log, sqrt)

```{r}
y_pois <- rpois(n, exp(eta_true))

for (lnk in c("log", "sqrt")) {
    fam <- poisson(lnk)
    f <- fastglm(X, y_pois, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y_pois, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("poisson  %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}
```

## Gamma (log, inverse)

```{r}
y_gam <- rgamma(n, shape = 5, rate = 5 / exp(eta_true))

for (lnk in c("log", "inverse")) {
    fam <- Gamma(lnk)
    f <- fastglm(X, y_gam, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y_gam, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("Gamma    %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}
```

## Gaussian (identity, log)

```{r}
y_gauss <- eta_true + rnorm(n, 0, 0.5)

for (lnk in c("identity", "log")) {
    fam <- gaussian(lnk)
    y0 <- if (lnk == "log") pmax(y_gauss, 0.01) else y_gauss
    f <- fastglm(X, y0, family = fam, firth = TRUE, tol = 1e-10)
    b <- glm(y ~ x1 + x2, data = data.frame(y = y0, x1 = x1, x2 = x2),
             family = fam, method = "brglmFit", type = "AS_mean",
             control = list(epsilon = 1e-10, maxit = 200))
    cat(sprintf("gaussian %-8s max|diff| = %.2e\n", lnk,
                max(abs(unname(coef(f)) - unname(coef(b))))))
}
```

## Inverse Gaussian (log)

```{r}
mu_ig <- exp(eta_true)
y_ig <- statmod::rinvgauss(n, mean = mu_ig, shape = 5)
y_ig[y_ig <= 0] <- 0.01

f <- fastglm(X, y_ig, family = inverse.gaussian("log"), firth = TRUE, tol = 1e-10)
b <- glm(y ~ x1 + x2, data = data.frame(y = y_ig, x1 = x1, x2 = x2),
         family = inverse.gaussian("log"), method = "brglmFit", type = "AS_mean",
         control = list(epsilon = 1e-10, maxit = 200))
cat(sprintf("inv.gauss log      max|diff| = %.2e\n",
            max(abs(unname(coef(f)) - unname(coef(b))))))
```

# Standard errors

The standard errors reported by `fastglm` with `firth = TRUE` come from the unscaled $(X^\top W X)^{-1}$ at convergence, the same quantity that `brglm2` uses. For unit-dispersion families (binomial, Poisson), these are the Firth-penalized standard errors directly; for other families, the dispersion is estimated as $\hat\phi = D / (n - p)$ and applied to the variance-covariance matrix.

```{r}
f <- fastglm(X, y_bin, family = binomial(), firth = TRUE, tol = 1e-10)
b <- glm(y ~ x1 + x2, data = df, family = binomial(), method = "brglmFit",
         type = "AS_mean", control = list(epsilon = 1e-10, maxit = 200))

cbind(fastglm_SE = f$se, brglm2_SE = summary(b)$coefficients[, "Std. Error"])
max(abs(f$se - summary(b)$coefficients[, "Std. Error"]))
```

# Penalized deviance

The Firth-penalized deviance is $D^* = D - \log |X^\top W X|$, where $D$ is the standard deviance and the log-determinant comes from the information matrix at the penalized MLE. Both the standard and penalized deviances are reported:

```{r}
f <- fastglm(X, y_bin, family = binomial(), firth = TRUE)
c(deviance = f$deviance,
  log.det.XtWX = f$log.det.XtWX,
  penalized.deviance = f$penalized.deviance)
```

# Speed comparison across families

A timing comparison of `fastglm(firth = TRUE)` against `brglm2::brglmFit(type = "AS_mean")` across families with `n = 10000` observations and `p = 5` covariates:

```{r}
set.seed(123)
n <- 10000;  p <- 5
x1 <- rnorm(n);  x2 <- rnorm(n); x3 <- rnorm(n)
x4 <- rnorm(n);  x5 <- rnorm(n)
X <- cbind(1, x1, x2, x3, x4, x5)
eta <- 0.5 + 0.3 * x1 - 0.2 * x2 + 0.1 * x3

families <- list(
    `binomial logit`   = list(fam = binomial("logit"),
                              y = rbinom(n, 1, plogis(eta))),
    `binomial probit`  = list(fam = binomial("probit"),
                              y = rbinom(n, 1, plogis(eta))),
    `binomial cloglog` = list(fam = binomial("cloglog"),
                              y = rbinom(n, 1, plogis(eta))),
    `poisson log`      = list(fam = poisson("log"),
                              y = rpois(n, exp(eta))),
    `Gamma log`        = list(fam = Gamma("log"),
                              y = rgamma(n, shape = 5, rate = 5 / exp(eta))),
    `gaussian identity`= list(fam = gaussian("identity"),
                              y = eta + rnorm(n, 0, 0.5))
)

df <- data.frame(y = 0, x1 = x1, x2 = x2, x3 = x3, x4 = x4, x5 = x5)
results <- list()
for (nm in names(families)) {
    fi <- families[[nm]]
    df$y <- fi$y
    mb <- microbenchmark::microbenchmark(
        fastglm = fastglm(X, fi$y, family = fi$fam, firth = TRUE),
        brglm2  = glm(y ~ x1 + x2 + x3 + x4 + x5, data = df,
                      family = fi$fam, method = "brglmFit", type = "AS_mean"),
        times = 10L
    )
    s <- summary(mb, unit = "ms")
    cat(sprintf("%-20s fastglm=%7.2f ms  brglm2=%7.2f ms  speedup=%.1fx\n",
                nm, s$median[1], s$median[2], s$median[2] / s$median[1]))
}
```

# References

- Firth, D. (1993). Bias reduction of maximum likelihood estimates. *Biometrika*, 80(1), 27--38. <doi:10.1093/biomet/80.1.27>

- Heinze, G. and Schemper, M. (2002). A solution to the problem of separation in logistic regression. *Statistics in Medicine*, 21(16), 2409--2419. <doi:10.1002/sim.1047>

- Kosmidis, I. and Firth, D. (2009). Bias reduction in exponential family nonlinear models. *Biometrika*, 96(4), 793--804. <doi:10.1093/biomet/asp055>

- Kosmidis, I. and Firth, D. (2021). Jeffreys-prior penalty, finiteness and shrinkage in binomial-response generalized linear models. *Biometrika*, 108(1), 71--82. <doi:10.1093/biomet/asaa052>

- Kosmidis, I. (2014). Bias in parametric estimation: reduction and useful side-effects. *WIREs Computational Statistics*, 6(3), 185--196. <doi:10.1002/wics.1296>

---
title: "Quick Usage Guide to the 'fastglm' Package"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Usage Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The `fastglm` package is intended to be a **fast** *and* **stable** alternative to the `glm()` and `glm2()` functions for fitting generalized lienar models. The The `fastglm` package is compatible with `R`'s `family` objects (see `?family`). The package can be installed via

```{r, eval = FALSE, echo=TRUE}
devtools::install_github("jaredhuling/fastglm")
```

and loaded via:

```{r, echo=TRUE}
library(fastglm)
```


## Example

Currently, the `fastglm` package does not allow for formula-based data input and is restricted to matrices. We use the following example to demonstrate the usage of `fastglm`:

```{r, echo = TRUE}
data(esoph)
x <- model.matrix(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp)
                                         + unclass(alcgp), data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

gfit1 <- fastglm(x = x, y = y, family = binomial(link = "cloglog"))

summary(gfit1)
```


## Computational stability 

The `fastglm` package does not compromise computational stability for speed. In fact, for many situations where `glm()` and even `glm2()` do not converge, `fastglm()` does converge.

As an example, consider the following data scenario, where the response distribution is (mildly) misspecified, but the link function is quite badly misspecified. In such scenarios, the standard IRLS algorithm tends to have convergence issues. The `glm2()` package was designed to handle such cases, however, it still can have convergence issues. The `fastglm()` package uses a similar step-halving technique as `glm2()`, but it starts at better initialized values and thus tends to have better convergence properties in practice. 

```{r, fig.show='hold'}
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- (exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000)) ) + 0.1


system.time(gfit1 <- fastglm(cbind(1, x), y, family = Gamma(link = "sqrt")))

system.time(gfit2 <- glm(y~x, family = Gamma(link = "sqrt")) )

system.time(gfit3 <- glm2::glm2(y~x, family = Gamma(link = "sqrt")) )

## Note that fastglm() returns estimates with the
## largest likelihood
logLik(gfit1)
logLik(gfit2)
logLik(gfit3)

coef(gfit1)[1:5]
coef(gfit2)[1:5]
coef(gfit3)[1:5]

## check convergence of fastglm
gfit1$converged
## number of IRLS iterations
gfit1$iter

## now check convergence for glm()
gfit2$converged
gfit2$iter

## check convergence for glm2()
gfit3$converged
gfit3$iter


## increasing number of IRLS iterations for glm() does not help that much
system.time(gfit2 <- glm(y~x, family = Gamma(link = "sqrt"), control = list(maxit = 100)) )

gfit2$converged
gfit2$iter

logLik(gfit1)
logLik(gfit2)

```

---
title: "Benchmark Study for 'fastglm'"
author: "Jared Huling"
date: "2026-05-01"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Benchmark Study for 'fastglm'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



This vignette is a more comprehensive benchmarking study than the small inline timings in the other vignettes. It compares *fastglm* against the canonical reference implementations across a range of sample sizes and model classes:

- standard GLMs against `stats::glm()`, `glm2::glm.fit2()`, and `speedglm::speedglm.wfit()`,
- the sparse and `big.matrix` solver paths against the dense path,
- negative-binomial regression against `MASS::glm.nb()`,
- Firth-penalized logistic against `logistf::logistf()`,
- hurdle and zero-inflated count regressions against `pscl::hurdle()` and `pscl::zeroinfl()`.

To re-run the benchmarks (after a code change, on a different machine, etc.), execute `vignettes/_precompile.R` from the package root.


``` r
library(fastglm)
suppressPackageStartupMessages({
    library(MASS)
    library(pscl)
    library(logistf)
    library(speedglm)
    library(Matrix)
    library(bigmemory)
    library(microbenchmark)
})

bench_unit <- "milliseconds"
.fmt <- function(mb) {
    s <- summary(mb, unit = bench_unit)
    s <- s[, c("expr", "min", "median", "mean", "max")]
    names(s) <- c("expr", "min (ms)", "median (ms)", "mean (ms)", "max (ms)")
    s
}
.collect <- function(mb, n) {
    s <- summary(mb, unit = bench_unit)
    data.frame(n = n, expr = as.character(s$expr), median = s$median,
               stringsAsFactors = FALSE)
}
.plot_scaling <- function(df, title) {
    methods <- unique(df$expr)
    cols    <- seq_along(methods)
    op <- par(mar = c(4.2, 4.2, 3, 1), las = 1)
    on.exit(par(op), add = TRUE)
    plot(NA, log = "xy",
         xlim = range(df$n), ylim = range(df$median),
         xlab = "n (log scale)",
         ylab = paste0("median time (", bench_unit, ", log scale)"),
         main = title)
    grid(equilogs = FALSE, col = "gray85")
    for (k in seq_along(methods)) {
        sub <- df[df$expr == methods[k], ]
        sub <- sub[order(sub$n), ]
        lines (sub$n, sub$median, col = cols[k], lwd = 2)
        points(sub$n, sub$median, col = cols[k], pch = 19)
    }
    legend("topleft", legend = methods, col = cols,
           pch = 19, lty = 1, lwd = 2, bty = "n")
}
sessionInfo()
#> R version 4.5.1 (2025-06-13)
#> Platform: aarch64-apple-darwin20
#> Running under: macOS Sequoia 15.7.3
#> 
#> Matrix products: default
#> BLAS:   /Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/lib/libRblas.0.dylib 
#> LAPACK: /Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/lib/libRlapack.dylib;  LAPACK version 3.12.1
#> 
#> locale:
#> [1] en_US.UTF-8/en_US.UTF-8/en_US.UTF-8/C/en_US.UTF-8/en_US.UTF-8
#> 
#> time zone: America/Chicago
#> tzcode source: internal
#> 
#> attached base packages:
#> [1] stats     graphics  grDevices utils     datasets  methods   base     
#> 
#> other attached packages:
#>  [1] microbenchmark_1.5.0 bigmemory_4.6.4      speedglm_0.3-5      
#>  [4] biglm_0.9-3          DBI_1.2.3            Matrix_1.7-5        
#>  [7] logistf_1.26.1       pscl_1.5.9           MASS_7.3-65         
#> [10] fastglm_0.0.5       
#> 
#> loaded via a namespace (and not attached):
#>  [1] sandwich_3.1-1         generics_0.1.4         tidyr_1.3.1           
#>  [4] shape_1.4.6.1          lattice_0.22-7         lme4_1.1-37           
#>  [7] magrittr_2.0.4         mitml_0.4-5            estimability_1.5.1    
#> [10] evaluate_1.0.4         grid_4.5.1             iterators_1.0.14      
#> [13] mice_3.18.0            mvtnorm_1.3-3          foreach_1.5.2         
#> [16] jomo_2.7-6             operator.tools_1.6.3.1 glmnet_4.1-10         
#> [19] nnet_7.3-20            backports_1.5.0        survival_3.8-3        
#> [22] multcomp_1.4-28        mgcv_1.9-3             purrr_1.2.1           
#> [25] bigmemory.sri_0.1.8    TH.data_1.1-3          codetools_0.2-20      
#> [28] reformulas_0.4.1       Rdpack_2.6.4           cli_3.6.5             
#> [31] rlang_1.1.7            rbibutils_2.3          splines_4.5.1         
#> [34] pan_1.9                tools_4.5.1            uuid_1.2-1            
#> [37] nloptr_2.2.1           minqa_1.2.8            dplyr_1.1.4           
#> [40] boot_1.3-31            broom_1.0.12           vctrs_0.7.2           
#> [43] R6_2.6.1               rpart_4.1.24           zoo_1.8-14            
#> [46] emmeans_1.11.2         lifecycle_1.0.4        pkgconfig_2.0.3       
#> [49] pillar_1.11.1          glue_1.8.0             Rcpp_1.1.0            
#> [52] xfun_0.52              tibble_3.3.0           tidyselect_1.2.1      
#> [55] knitr_1.50             xtable_1.8-4           nlme_3.1-168          
#> [58] formula.tools_1.7.1    compiler_4.5.1
```

# Standard GLMs

Logistic regression with an increasing number of observations, holding `p = 30` fixed. Comparing the default *fastglm* path (`method = 0`, pivoted QR), the LLT-Cholesky path (`method = 2`), `stats::glm.fit()`, `glm2::glm.fit2()`, and `speedglm::speedglm.wfit()`.


``` r
set.seed(1)
p <- 30

run_one_logit <- function(n) {
    X <- matrix(rnorm(n * p), n, p)
    beta <- rnorm(p) * 0.05
    y <- rbinom(n, 1, plogis(X %*% beta))
    microbenchmark(
        fastglm_qr   = fastglm(X, y, family = binomial(),            method = 0),
        fastglm_chol = fastglm(X, y, family = binomial(),            method = 2),
        glm.fit      = glm.fit(X, y, family = binomial()),
        glm2_fit2    = glm2::glm.fit2(X, y, family = binomial()),
        speedglm     = speedglm::speedglm.wfit(y = y, X = X,
                                               family = binomial(),
                                               intercept = FALSE),
        times = 5L
    )
}

logit_tim <- list()
for (n in c(1e3, 1e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_logit(n)
    print(.fmt(mb))
    logit_tim[[length(logit_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>           expr min (ms) median (ms) mean (ms) max (ms)
#> 1   fastglm_qr 0.782649    0.820000 1.0789150 2.105022
#> 2 fastglm_chol 0.396552    0.405285 0.4147232 0.455674
#> 3      glm.fit 2.432202    2.562992 2.9429800 4.450960
#> 4    glm2_fit2 2.499114    2.670330 2.9793962 4.506105
#> 5     speedglm 1.928189    1.949714 2.4329318 4.355225
#> 
#> --- n = 10000 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  8.585072    8.746571  8.860674  9.280924
#> 2 fastglm_chol  3.426247    4.690400 14.163278 53.598890
#> 3      glm.fit 27.636501   28.415173 28.887140 30.403837
#> 4    glm2_fit2 26.937287   28.327351 28.159472 28.958751
#> 5     speedglm 19.654744   20.303077 21.238689 24.291106
#> 
#> --- n = 1e+05 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  89.25753    90.25970  91.00553  95.08548
#> 2 fastglm_chol  37.25432    40.74473  40.27186  41.78884
#> 3      glm.fit 284.38035   325.13812 335.31097 419.44287
#> 4    glm2_fit2 326.96004   330.35184 341.28560 382.65243
#> 5     speedglm 192.51038   196.15589 215.62233 249.94162
logit_tim <- do.call(rbind, logit_tim)
.plot_scaling(logit_tim, "Logistic regression: median time vs n (p = 30)")
```

![plot of chunk unnamed-chunk-2](benchmarks-figs/unnamed-chunk-2-1.png)

The same comparison for Poisson regression with a log link:


``` r
run_one_poisson <- function(n) {
    X <- matrix(rnorm(n * p), n, p)
    beta <- rnorm(p) * 0.05
    y <- rpois(n, exp(X %*% beta + 1))
    microbenchmark(
        fastglm_qr   = fastglm(X, y, family = poisson(),            method = 0),
        fastglm_chol = fastglm(X, y, family = poisson(),            method = 2),
        glm.fit      = glm.fit(X, y, family = poisson()),
        glm2_fit2    = glm2::glm.fit2(X, y, family = poisson()),
        speedglm     = speedglm::speedglm.wfit(y = y, X = X,
                                               family = poisson(),
                                               intercept = FALSE),
        times = 5L
    )
}

pois_tim <- list()
for (n in c(1e3, 1e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_poisson(n)
    print(.fmt(mb))
    pois_tim[[length(pois_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>           expr min (ms) median (ms) mean (ms) max (ms)
#> 1   fastglm_qr 0.868954    0.969281 1.0339708 1.424545
#> 2 fastglm_chol 0.477650    0.491836 0.5633072 0.865428
#> 3      glm.fit 2.641138    2.913706 3.0120322 3.630304
#> 4    glm2_fit2 2.721662    2.880168 2.8681304 3.035230
#> 5     speedglm 2.068327    2.149589 2.2176162 2.455203
#> 
#> --- n = 10000 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  8.404467    8.701348  9.718296 13.499168
#> 2 fastglm_chol  3.310258    3.514110  3.615282  3.915787
#> 3      glm.fit 26.118312   26.669311 28.270172 31.484064
#> 4    glm2_fit2 26.971112   27.684020 28.347859 31.823749
#> 5     speedglm 19.046714   19.233592 19.286884 19.661427
#> 
#> --- n = 1e+05 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  89.36057    89.80078 100.39996 141.35496
#> 2 fastglm_chol  36.22313    38.34779  38.18333  39.22138
#> 3      glm.fit 325.01807   327.65740 330.18867 344.74932
#> 4    glm2_fit2 278.52657   354.66779 363.70979 433.98287
#> 5     speedglm 199.96290   204.68049 227.75442 284.12988
pois_tim <- do.call(rbind, pois_tim)
.plot_scaling(pois_tim, "Poisson regression: median time vs n (p = 30)")
```

![plot of chunk unnamed-chunk-3](benchmarks-figs/unnamed-chunk-3-1.png)

The Cholesky paths are 3&ndash;5x faster than `glm.fit()` at moderate `n` and the gap widens with sample size &mdash; the IRLS structure is the same, but each iteration's WLS solve is materially cheaper in *RcppEigen* than in compiled R + LAPACK. *speedglm* is competitive at small `n` but its single-threaded cross-product accumulation overtakes the savings as `n` grows, since *fastglm* lets Eigen parallelise the WLS solve.

# Sparse and big.matrix paths

A sparse design matrix from one-hot encoding a 100-level factor against a continuous covariate. The column count is `p = 200`; *fastglm* on the dense matrix has to materialise all 200 columns explicitly, while the sparse path only operates on the non-zero entries:


``` r
set.seed(2)
n   <- 5e4
g   <- factor(sample.int(100, n, replace = TRUE))
xn  <- rnorm(n)
Xd  <- model.matrix(~ g * xn)
Xs  <- as(Xd, "CsparseMatrix")
betas <- rnorm(ncol(Xd)) / 4
y   <- rbinom(n, 1, plogis(Xd %*% betas))

mb_sparse <- microbenchmark(
    fastglm_dense  = fastglm(Xd, y, family = binomial(), method = 2),
    fastglm_sparse = fastglm(Xs, y, family = binomial(), method = 2),
    times = 5L
)
print(.fmt(mb_sparse))
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1  fastglm_dense 218.7811    220.4451  222.2950 227.3766
#> 2 fastglm_sparse 247.9938    262.4799  268.1184 307.9439

cat("ncol(Xd) =", ncol(Xd), "  fraction nonzero =",
    round(length(Xs@x) / prod(dim(Xs)), 3), "\n")
#> ncol(Xd) = 200   fraction nonzero = 0.02
```

A `big.matrix` comparison on a moderately large dense design &mdash; `n = 5e5`, `p = 20`. The on-disc `big.matrix` runs in row-blocks but produces identical estimates:


``` r
set.seed(3)
n  <- 5e5
p  <- 20
X  <- matrix(rnorm(n * p), n, p)
Xb <- as.big.matrix(X)
y  <- rbinom(n, 1, plogis(X %*% rnorm(p) * 0.05))

mb_big <- microbenchmark(
    dense       = fastglm(X,  y, family = binomial(), method = 2),
    big.matrix  = fastglm(Xb, y, family = binomial(), method = 2),
    times = 3L
)
print(.fmt(mb_big))
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1      dense 129.0053    136.7083  136.9545 145.1500
#> 2 big.matrix 141.8254    142.1299  142.3005 142.9463

f1 <- fastglm(X,  y, family = binomial(), method = 2)
f2 <- fastglm(Xb, y, family = binomial(), method = 2)
cat("max coef diff:", max(abs(coef(f1) - coef(f2))), "\n")
#> max coef diff: 1.19349e-15
```

The dense path is faster for matrices that fit in RAM (the row-block reads have non-zero overhead), but the `big.matrix` path is the only viable option once the design exceeds available memory.

# Negative-binomial regression

`fastglm_nb()` against `MASS::glm.nb()` across a range of sample sizes. The data-generating process is NB(`mu`, `theta = 2`) with three covariates and an intercept:


``` r
set.seed(4)
run_one_nb <- function(n) {
    X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
    mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
    y  <- MASS::rnegbin(n, mu = mu, theta = 2)
    df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])
    microbenchmark(
        fastglm_nb = fastglm_nb(X, y),
        glm.nb     = MASS::glm.nb(y ~ x1 + x2 + x3, data = df),
        times = 5L
    )
}

nb_tim <- list()
for (n in c(1e3, 1e4, 1e5, 5e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_nb(n)
    print(.fmt(mb))
    nb_tim[[length(nb_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_nb 0.611515    0.686873 0.9694204 2.219494
#> 2     glm.nb 7.588854    7.803079 7.7879828 8.017263
#> 
#> --- n = 10000 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_nb  6.882834    7.264339   7.26912  7.707754
#> 2     glm.nb 59.773449   60.203990  62.97967 69.503487
#> 
#> --- n = 1e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_nb  58.78539    59.63278   62.1734  72.74929
#> 2     glm.nb 511.89140   513.57281  525.5684 568.34331
#> 
#> --- n = 5e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_nb  292.6352    294.9895  296.1172  300.6103
#> 2     glm.nb 2049.1221   2113.7185 2114.2318 2191.1647
nb_tim <- do.call(rbind, nb_tim)
.plot_scaling(nb_tim, "Negative-binomial joint MLE: median time vs n")
```

![plot of chunk unnamed-chunk-6](benchmarks-figs/unnamed-chunk-6-1.png)

Accuracy on the largest case: coefficients and `theta` agree to floating-point precision against `glm.nb`, since both maximise the same likelihood. The runtime difference is structural &mdash; `glm.nb` runs its IRLS + theta-MLE alternation in R, *fastglm_nb* runs both loops in C++.


``` r
set.seed(99)
n  <- 5e5
X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y  <- MASS::rnegbin(n, mu = mu, theta = 2)
df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])
f_nb <- fastglm_nb(X, y)
m_nb <- MASS::glm.nb(y ~ x1 + x2 + x3, data = df)
cat("coef max abs diff:", max(abs(unname(coef(f_nb)) - unname(coef(m_nb)))), "\n")
#> coef max abs diff: 2.930989e-14
cat("theta diff       :", abs(f_nb$theta - m_nb$theta), "\n")
#> theta diff       : 3.095302e-13
cat("logLik diff      :",
    abs(as.numeric(logLik(f_nb)) - as.numeric(logLik(m_nb))), "\n")
#> logLik diff      : 1.164153e-10
```

# Firth bias-reduced logistic

Firth-penalized logistic against `logistf::logistf()` on the canonical *sex2* example (Heinze and Schemper, 2002), and on a larger simulated example with mild quasi-separation:


``` r
data(sex2, package = "logistf")
X_sex2 <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y_sex2 <- sex2$case

mb_firth1 <- microbenchmark(
    fastglm = fastglm(X_sex2, y_sex2, family = binomial(), firth = TRUE),
    logistf = logistf(case ~ age + oc + vic + vicl + vis + dia, data = sex2),
    times = 10L
)
cat("\n--- Heinze-Schemper sex2 (n =", nrow(sex2), ") ---\n")
#> 
#> --- Heinze-Schemper sex2 (n = 239 ) ---
print(.fmt(mb_firth1))
#>      expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm 0.190978   0.2052255 0.2634332 0.699255
#> 2 logistf 6.731749   6.9578640 7.1186045 8.638290

set.seed(5)
n <- 2000;  p <- 10
X_big <- cbind(1, matrix(rnorm(n * p), n, p))
beta  <- rnorm(p + 1) * 0.4
y_big <- rbinom(n, 1, plogis(X_big %*% beta))
df_big <- data.frame(y = y_big, X_big[, -1])

mb_firth2 <- microbenchmark(
    fastglm = fastglm(X_big, y_big, family = binomial(), firth = TRUE),
    logistf = logistf(y ~ ., data = df_big, control = logistf.control(maxit = 50)),
    times = 5L
)
cat("\n--- simulated (n =", n, ", p =", p, ") ---\n")
#> 
#> --- simulated (n = 2000 , p = 10 ) ---
print(.fmt(mb_firth2))
#>      expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm  1.174322     1.25583  1.605494  2.688985
#> 2 logistf 48.478318    49.67937 52.509405 65.116241

# accuracy
f1 <- fastglm(X_sex2, y_sex2, family = binomial(), firth = TRUE)
l1 <- logistf(case ~ age + oc + vic + vicl + vis + dia, data = sex2)
cat("\nsex2 max coef diff:",
    max(abs(unname(coef(f1)) - unname(coef(l1)))), "\n")
#> 
#> sex2 max coef diff: 1.67938e-07

# small grouped barplot of medians
firth_df <- rbind(
    transform(.collect(mb_firth1, nrow(sex2)), case = "sex2"),
    transform(.collect(mb_firth2, n),          case = paste0("sim n=", n))
)
op <- par(mar = c(4.2, 4.2, 3, 1), las = 1)
m <- with(firth_df, tapply(median, list(expr, case), identity))
barplot(m, beside = TRUE, log = "y",
        col = c("#1f77b4", "#d62728"),
        ylab = paste0("median time (", bench_unit, ", log scale)"),
        main = "Firth-penalised logistic: median time per case",
        legend.text = rownames(m),
        args.legend = list(x = "topleft", bty = "n"))
```

![plot of chunk unnamed-chunk-8](benchmarks-figs/unnamed-chunk-8-1.png)

``` r
par(op)
```

# Hurdle models

`fastglm_hurdle` against `pscl::hurdle` across a range of sample sizes, Poisson count component:


``` r
set.seed(6)
run_one_hurdle <- function(n) {
    x1 <- rnorm(n);  x2 <- rnorm(n)
    lam    <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
    is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
    yt     <- integer(n)
    for (i in seq_len(n)) {
        repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
    }
    y  <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x1 = x1, x2 = x2)
    microbenchmark(
        fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson"),
        pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "poisson"),
        times = 5L
    )
}

hurdle_tim <- list()
for (n in c(1e3, 1e4, 5e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_hurdle(n)
    print(.fmt(mb))
    hurdle_tim[[length(hurdle_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_hurdle 0.502045    0.561741  1.274378 4.197539
#> 2    pscl_hurdle 4.674779    5.173462  5.296864 5.806051
#> 
#> --- n = 10000 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  2.586034    2.796159   3.00334  4.107052
#> 2    pscl_hurdle 41.390566   42.730569  45.19963 56.807796
#> 
#> --- n = 50000 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  13.52869    13.95025   13.9395  14.28608
#> 2    pscl_hurdle 267.35223   275.48417  274.6044 280.94914
#> 
#> --- n = 1e+05 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  28.58811    29.49532  31.32037  39.03048
#> 2    pscl_hurdle 472.64718   488.43985 487.85413 501.12820
hurdle_tim <- do.call(rbind, hurdle_tim)
.plot_scaling(hurdle_tim, "Hurdle Poisson: median time vs n")
```

![plot of chunk unnamed-chunk-9](benchmarks-figs/unnamed-chunk-9-1.png)

NB hurdle (with simultaneous `theta` MLE) on a single moderately large example:


``` r
set.seed(7)
n <- 1e4
x1 <- rnorm(n);  x2 <- rnorm(n)
lam <- exp(0.8 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1))
yt <- integer(n)
for (i in seq_len(n)) {
    repeat {
        v <- MASS::rnegbin(1, mu = lam[i], theta = 1.5)
        if (v > 0) { yt[i] <- v; break }
    }
}
y  <- ifelse(is_pos == 1, yt, 0L)
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_hurdle_nb <- microbenchmark(
    fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "negbin"),
    pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "negbin"),
    times = 3L
)
print(.fmt(mb_hurdle_nb))
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_hurdle 22.92876    25.15067  25.08678 27.18091
#> 2    pscl_hurdle 50.02750    51.05701  50.98267 51.86348
```

# Zero-inflated models

Zero-inflated Poisson against `pscl::zeroinfl()` across sample sizes:


``` r
set.seed(8)
run_one_zi <- function(n) {
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
    eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
    z     <- rbinom(n, 1, plogis(eta_z))
    y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
    df    <- data.frame(y = y, x1 = x1, x2 = x2)
    microbenchmark(
        fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson"),
        pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson"),
        times = 5L
    )
}

zi_tim <- list()
for (n in c(1e3, 1e4, 5e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_zi(n)
    print(.fmt(mb))
    zi_tim[[length(zi_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_zi 3.913245    4.128946  4.357406 5.386539
#> 2    pscl_zi 8.937139    9.242794  9.267902 9.536928
#> 
#> --- n = 10000 ---
#>         expr min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi 41.80028    43.05660  44.98070  49.08741
#> 2    pscl_zi 93.26643    93.93662  99.71534 111.59405
#> 
#> --- n = 50000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_zi 216.0857    217.6707  220.9450 230.0795
#> 2    pscl_zi 523.3382    536.0326  547.1348 595.4259
#> 
#> --- n = 1e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi  423.3069     428.439  429.8353  440.3532
#> 2    pscl_zi 1036.3304    1050.165 1048.7925 1058.5976
zi_tim <- do.call(rbind, zi_tim)
.plot_scaling(zi_tim, "Zero-inflated Poisson: median time vs n")
```

![plot of chunk unnamed-chunk-11](benchmarks-figs/unnamed-chunk-11-1.png)

Zero-inflated NB on a single moderately large example:


``` r
set.seed(9)
n  <- 1e4
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.8 + 0.4 * x1 - 0.3 * x2; lam <- exp(eta_c)
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z <- rbinom(n, 1, plogis(eta_z))
y <- ifelse(z == 1, 0L, MASS::rnegbin(n, mu = lam, theta = 2.0))
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_zi_nb <- microbenchmark(
    fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "negbin",
                            em.tol = 1e-8, em.maxit = 200L),
    pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "negbin"),
    times = 3L
)
print(.fmt(mb_zi_nb))
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi  62.05588    67.03602  69.34702  78.94915
#> 2    pscl_zi 142.46971   158.26016 154.09418 161.55267
```

# Summary

A compact summary of the speed advantage *fastglm* delivers, expressed as the ratio of reference-implementation median time to *fastglm* median time. Larger is better. The reference for each model class is the canonical R implementation; for the standard GLMs we report the ratio against the fastest among `glm.fit`, `glm2`, and `speedglm` so the comparison is conservative.


``` r
.speedup <- function(df, fast_label, ref_labels) {
    out <- lapply(split(df, df$n), function(sub) {
        fast <- sub$median[sub$expr == fast_label]
        if (length(fast) == 0) return(NULL)
        ref  <- min(sub$median[sub$expr %in% ref_labels])
        data.frame(n = sub$n[1], speedup = ref / fast)
    })
    do.call(rbind, out)
}

su_logit  <- .speedup(logit_tim,  "fastglm_chol",
                      c("glm.fit", "glm2_fit2", "speedglm"))
su_pois   <- .speedup(pois_tim,   "fastglm_chol",
                      c("glm.fit", "glm2_fit2", "speedglm"))
su_nb     <- .speedup(nb_tim,     "fastglm_nb",     "glm.nb")
su_hurdle <- .speedup(hurdle_tim, "fastglm_hurdle", "pscl_hurdle")
su_zi     <- .speedup(zi_tim,     "fastglm_zi",     "pscl_zi")

su_all <- rbind(
    transform(su_logit,  model = "logit"),
    transform(su_pois,   model = "poisson"),
    transform(su_nb,     model = "neg-binomial"),
    transform(su_hurdle, model = "hurdle"),
    transform(su_zi,     model = "zero-inflated")
)

op <- par(mar = c(4.2, 4.5, 3, 1), las = 1)
models <- unique(su_all$model)
cols   <- seq_along(models)
plot(NA, log = "xy",
     xlim = range(su_all$n), ylim = range(su_all$speedup),
     xlab = "n (log scale)",
     ylab = "speedup over reference (x, log scale)",
     main = "fastglm speedup vs canonical reference, by model class")
abline(h = 1, col = "gray70", lty = 2)
grid(equilogs = FALSE, col = "gray85")
for (k in seq_along(models)) {
    sub <- su_all[su_all$model == models[k], ]
    sub <- sub[order(sub$n), ]
    lines (sub$n, sub$speedup, col = cols[k], lwd = 2)
    points(sub$n, sub$speedup, col = cols[k], pch = 19)
}
legend("topleft", legend = models, col = cols, pch = 19, lty = 1, lwd = 2,
       bty = "n")
```

![plot of chunk unnamed-chunk-13](benchmarks-figs/unnamed-chunk-13-1.png)

``` r
par(op)
```

Across all six model classes the same picture holds:

- *fastglm* matches the canonical reference implementation to floating-point precision (or to within the EM tolerance for zero-inflation), so the numerical answer is the same.
- The runtime gap grows with sample size, since R-side overhead in the reference implementations is fixed-per-iteration. By `n = 1e5` the speedup is generally an order of magnitude or more.
- For models with an outer iteration (NB joint MLE, hurdle/ZI with NB), the gap is widest, since the entire outer loop is in C++ in *fastglm* and entirely in R for `MASS::glm.nb`, `pscl::hurdle`, `pscl::zeroinfl`.

# References

- Enea, M. (2009). Fitting linear models and generalized linear models with large data sets in R. In *Statistical Methods for the Analysis of Large Data-sets*, Italian Statistical Society. *speedglm* package documentation.

- Firth, D. (1993). Bias reduction of maximum likelihood estimates. *Biometrika*, 80(1), 27&ndash;38. <doi:10.1093/biomet/80.1.27>

- Heinze, G. and Schemper, M. (2002). A solution to the problem of separation in logistic regression. *Statistics in Medicine*, 21(16), 2409&ndash;2419. <doi:10.1002/sim.1047>

- Marschner, I. C. (2011). glm2: Fitting generalized linear models with convergence problems. *The R Journal*, 3(2), 12&ndash;15. <doi:10.32614/RJ-2011-012>

- Venables, W. N. and Ripley, B. D. (2002). *Modern Applied Statistics with S* (4th ed.). Springer.

- Zeileis, A., Kleiber, C., and Jackman, S. (2008). Regression models for count data in R. *Journal of Statistical Software*, 27(8), 1&ndash;25. <doi:10.18637/jss.v027.i08>

---
title: "Benchmark Study for 'fastglm'"
author: "Jared Huling"
date: "2026-05-27"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Benchmark Study for 'fastglm'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



This vignette is a more comprehensive benchmarking study than the small inline timings in the other vignettes. It compares *fastglm* against the canonical reference implementations across a range of sample sizes and model classes:

- standard GLMs against `stats::glm()`, `glm2::glm.fit2()`, and `speedglm::speedglm.wfit()`,
- the sparse and `big.matrix` solver paths against the dense path,
- negative-binomial regression against `MASS::glm.nb()`,
- Firth bias-reduced GLMs against `brglm2::brglmFit()` and `logistf::logistf()`, including sparse, streaming, and `big.matrix` Firth paths,
- hurdle and zero-inflated count regressions against `pscl::hurdle()` and `pscl::zeroinfl()`.

The vignette is **pre-compiled** &mdash; the timings below were produced once on the maintainer's machine and are baked into the shipped `.Rmd`. R CMD check / build does *not* re-run the benchmarks, which keeps the package well within CRAN's runtime budget. To re-run the benchmarks (after a code change, on a different machine, etc.), execute `vignettes/_precompile.R` from the package root.


``` r
library(fastglm)
suppressPackageStartupMessages({
    library(MASS)
    library(pscl)
    library(logistf)
    library(brglm2)
    library(speedglm)
    library(Matrix)
    library(bigmemory)
    library(microbenchmark)
})

bench_unit <- "milliseconds"
.fmt <- function(mb) {
    s <- summary(mb, unit = bench_unit)
    s <- s[, c("expr", "min", "median", "mean", "max")]
    names(s) <- c("expr", "min (ms)", "median (ms)", "mean (ms)", "max (ms)")
    s
}
.collect <- function(mb, n) {
    s <- summary(mb, unit = bench_unit)
    data.frame(n = n, expr = as.character(s$expr), median = s$median,
               stringsAsFactors = FALSE)
}
.plot_scaling <- function(df, title) {
    methods <- unique(df$expr)
    cols    <- seq_along(methods)
    op <- par(mar = c(4.2, 4.2, 3, 1), las = 1)
    on.exit(par(op), add = TRUE)
    plot(NA, log = "xy",
         xlim = range(df$n), ylim = range(df$median),
         xlab = "n (log scale)",
         ylab = paste0("median time (", bench_unit, ", log scale)"),
         main = title)
    grid(equilogs = FALSE, col = "gray85")
    for (k in seq_along(methods)) {
        sub <- df[df$expr == methods[k], ]
        sub <- sub[order(sub$n), ]
        lines (sub$n, sub$median, col = cols[k], lwd = 2)
        points(sub$n, sub$median, col = cols[k], pch = 19)
    }
    legend("topleft", legend = methods, col = cols,
           pch = 19, lty = 1, lwd = 2, bty = "n")
}
sessionInfo()
#> R version 4.5.1 (2025-06-13)
#> Platform: aarch64-apple-darwin20
#> Running under: macOS Sequoia 15.7.3
#> 
#> Matrix products: default
#> BLAS:   /Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/lib/libRblas.0.dylib 
#> LAPACK: /Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/lib/libRlapack.dylib;  LAPACK version 3.12.1
#> 
#> locale:
#> [1] en_US.UTF-8/en_US.UTF-8/en_US.UTF-8/C/en_US.UTF-8/en_US.UTF-8
#> 
#> time zone: America/Chicago
#> tzcode source: internal
#> 
#> attached base packages:
#> [1] stats     graphics  grDevices utils     datasets  methods   base     
#> 
#> other attached packages:
#>  [1] microbenchmark_1.5.0 bigmemory_4.6.4      speedglm_0.3-5      
#>  [4] biglm_0.9-3          DBI_1.2.3            Matrix_1.7-5        
#>  [7] brglm2_1.1.0         logistf_1.26.1       pscl_1.5.9          
#> [10] MASS_7.3-65          fastglm_0.1.1       
#> 
#> loaded via a namespace (and not attached):
#>  [1] shape_1.4.6.1          xfun_0.57              formula.tools_1.7.1   
#>  [4] lattice_0.22-7         numDeriv_2016.8-1.1    vctrs_0.7.2           
#>  [7] tools_4.5.1            Rdpack_2.6.4           generics_0.1.4        
#> [10] enrichwith_0.5.0       sandwich_3.1-1         tibble_3.3.0          
#> [13] pan_1.9                pkgconfig_2.0.3        jomo_2.7-6            
#> [16] uuid_1.2-1             lifecycle_1.0.4        compiler_4.5.1        
#> [19] statmod_1.5.1          codetools_0.2-20       glmnet_4.1-10         
#> [22] mice_3.18.0            pillar_1.11.1          nloptr_2.2.1          
#> [25] tidyr_1.3.1            reformulas_0.4.1       iterators_1.0.14      
#> [28] rpart_4.1.24           boot_1.3-31            multcomp_1.4-28       
#> [31] foreach_1.5.2          mitml_0.4-5            nlme_3.1-168          
#> [34] tidyselect_1.2.1       mvtnorm_1.3-7          dplyr_1.1.4           
#> [37] purrr_1.2.1            splines_4.5.1          operator.tools_1.6.3.1
#> [40] grid_4.5.1             cli_3.6.5              magrittr_2.0.4        
#> [43] survival_3.8-3         broom_1.0.12           TH.data_1.1-3         
#> [46] bigmemory.sri_0.1.8    backports_1.5.0        estimability_1.5.1    
#> [49] emmeans_1.11.2         nnet_7.3-20            lme4_1.1-37           
#> [52] zoo_1.8-14             evaluate_1.0.5         knitr_1.51            
#> [55] rbibutils_2.3          mgcv_1.9-3             nleqslv_3.3.5         
#> [58] rlang_1.1.7            Rcpp_1.1.1-1.1         xtable_1.8-4          
#> [61] glue_1.8.0             minqa_1.2.8            R6_2.6.1
```

# Standard GLMs

Logistic regression with an increasing number of observations, holding `p = 30` fixed. Comparing the default *fastglm* path (`method = 0`, pivoted QR), the LLT-Cholesky path (`method = 2`), `stats::glm.fit()`, `glm2::glm.fit2()`, and `speedglm::speedglm.wfit()`.


``` r
set.seed(1)
p <- 30

run_one_logit <- function(n) {
    X <- matrix(rnorm(n * p), n, p)
    beta <- rnorm(p) * 0.05
    y <- rbinom(n, 1, plogis(X %*% beta))
    microbenchmark(
        fastglm_qr   = fastglm(X, y, family = binomial(),            method = 0),
        fastglm_chol = fastglm(X, y, family = binomial(),            method = 2),
        glm.fit      = glm.fit(X, y, family = binomial()),
        glm2_fit2    = glm2::glm.fit2(X, y, family = binomial()),
        speedglm     = speedglm::speedglm.wfit(y = y, X = X,
                                               family = binomial(),
                                               intercept = FALSE),
        times = 5L
    )
}

logit_tim <- list()
for (n in c(1e3, 1e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_logit(n)
    print(.fmt(mb))
    logit_tim[[length(logit_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>           expr min (ms) median (ms) mean (ms) max (ms)
#> 1   fastglm_qr 0.814178    0.840377 1.0302726 1.835201
#> 2 fastglm_chol 0.407499    0.437060 0.4445794 0.490442
#> 3      glm.fit 2.542164    2.741219 3.2153102 5.206713
#> 4    glm2_fit2 2.649010    2.706902 3.1439702 4.932997
#> 5     speedglm 1.901129    1.948853 2.6562588 5.489572
#> 
#> --- n = 10000 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  9.035785    9.244516  9.219235  9.399209
#> 2 fastglm_chol  3.817182    4.067159  5.147083  7.060364
#> 3      glm.fit 26.703710   28.254740 28.411204 29.695931
#> 4    glm2_fit2 28.278438   29.838447 29.758587 30.884726
#> 5     speedglm 18.876605   20.634644 20.516703 21.866735
#> 
#> --- n = 1e+05 ---
#>           expr  min (ms) median (ms) mean (ms) max (ms)
#> 1   fastglm_qr  94.45797    96.77181  97.43920 102.0181
#> 2 fastglm_chol  41.94234    43.42736  43.64336  46.0380
#> 3      glm.fit 284.13230   326.98791 338.71854 430.3533
#> 4    glm2_fit2 282.30566   330.86590 322.29955 338.4800
#> 5     speedglm 194.95307   200.74326 218.98994 253.5205
logit_tim <- do.call(rbind, logit_tim)
.plot_scaling(logit_tim, "Logistic regression: median time vs n (p = 30)")
```

![plot of chunk unnamed-chunk-2](benchmarks-figs/unnamed-chunk-2-1.png)

The same comparison for Poisson regression with a log link:


``` r
run_one_poisson <- function(n) {
    X <- matrix(rnorm(n * p), n, p)
    beta <- rnorm(p) * 0.05
    y <- rpois(n, exp(X %*% beta + 1))
    microbenchmark(
        fastglm_qr   = fastglm(X, y, family = poisson(),            method = 0),
        fastglm_chol = fastglm(X, y, family = poisson(),            method = 2),
        glm.fit      = glm.fit(X, y, family = poisson()),
        glm2_fit2    = glm2::glm.fit2(X, y, family = poisson()),
        speedglm     = speedglm::speedglm.wfit(y = y, X = X,
                                               family = poisson(),
                                               intercept = FALSE),
        times = 5L
    )
}

pois_tim <- list()
for (n in c(1e3, 1e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_poisson(n)
    print(.fmt(mb))
    pois_tim[[length(pois_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>           expr min (ms) median (ms) mean (ms) max (ms)
#> 1   fastglm_qr 0.821558    0.862271 0.8638290 0.905116
#> 2 fastglm_chol 0.461701    0.483759 0.4833244 0.512828
#> 3      glm.fit 2.610962    2.733757 2.7812432 2.957945
#> 4    glm2_fit2 2.708214    2.762580 2.7819976 2.930106
#> 5     speedglm 2.001743    2.176731 2.1828892 2.322568
#> 
#> --- n = 10000 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  8.970759    9.997973 10.727215 14.682264
#> 2 fastglm_chol  3.632887    3.813738  4.198236  4.997244
#> 3      glm.fit 27.088823   28.157611 29.683188 32.856662
#> 4    glm2_fit2 27.685045   28.174954 28.769019 31.296653
#> 5     speedglm 20.020464   20.423576 20.401247 20.846614
#> 
#> --- n = 1e+05 ---
#>           expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1   fastglm_qr  87.14702    88.71338  88.92693  90.54793
#> 2 fastglm_chol  37.15887    41.49143  51.18298  95.70085
#> 3      glm.fit 273.61846   278.11046 299.31023 382.50733
#> 4    glm2_fit2 281.06599   341.42389 340.61508 437.43101
#> 5     speedglm 201.79015   210.07244 226.30256 263.78400
pois_tim <- do.call(rbind, pois_tim)
.plot_scaling(pois_tim, "Poisson regression: median time vs n (p = 30)")
```

![plot of chunk unnamed-chunk-3](benchmarks-figs/unnamed-chunk-3-1.png)

The Cholesky paths are 3&ndash;5x faster than `glm.fit()` at moderate `n` and the gap widens with sample size &mdash; the IRLS structure is the same, but each iteration's WLS solve is materially cheaper in *RcppEigen* than in compiled R + LAPACK. *speedglm* is competitive at small `n` but its single-threaded cross-product accumulation overtakes the savings as `n` grows, since *fastglm* lets Eigen parallelize the WLS solve.

# Sparse and big.matrix paths

A sparse design matrix from one-hot encoding a 100-level factor against a continuous covariate. The column count is `p = 200`; *fastglm* on the dense matrix has to materialise all 200 columns explicitly, while the sparse path only operates on the non-zero entries:


``` r
set.seed(2)
n   <- 5e4
g   <- factor(sample.int(100, n, replace = TRUE))
xn  <- rnorm(n)
Xd  <- model.matrix(~ g * xn)
Xs  <- as(Xd, "CsparseMatrix")
betas <- rnorm(ncol(Xd)) / 4
y   <- rbinom(n, 1, plogis(Xd %*% betas))

mb_sparse <- microbenchmark(
    fastglm_dense  = fastglm(Xd, y, family = binomial(), method = 2),
    fastglm_sparse = fastglm(Xs, y, family = binomial(), method = 2),
    times = 5L
)
print(.fmt(mb_sparse))
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1  fastglm_dense 218.4417    219.5026  223.2900 231.7417
#> 2 fastglm_sparse 255.6885    257.9629  277.3992 339.6492

cat("ncol(Xd) =", ncol(Xd), "  fraction nonzero =",
    round(length(Xs@x) / prod(dim(Xs)), 3), "\n")
#> ncol(Xd) = 200   fraction nonzero = 0.02
```

A `big.matrix` comparison on a moderately large dense design &mdash; `n = 5e5`, `p = 20`. The on-disc `big.matrix` runs in row-blocks but produces identical estimates:


``` r
set.seed(3)
n  <- 5e5
p  <- 20
X  <- matrix(rnorm(n * p), n, p)
Xb <- as.big.matrix(X)
y  <- rbinom(n, 1, plogis(X %*% rnorm(p) * 0.05))

mb_big <- microbenchmark(
    dense       = fastglm(X,  y, family = binomial(), method = 2),
    big.matrix  = fastglm(Xb, y, family = binomial(), method = 2),
    times = 3L
)
print(.fmt(mb_big))
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1      dense 142.7995    144.3081  147.2925 154.7700
#> 2 big.matrix 149.8009    150.0625  151.0845 153.3899

f1 <- fastglm(X,  y, family = binomial(), method = 2)
f2 <- fastglm(Xb, y, family = binomial(), method = 2)
cat("max coef diff:", max(abs(coef(f1) - coef(f2))), "\n")
#> max coef diff: 1.235123e-15
```

The dense path is faster for matrices that fit in RAM (the row-block reads have non-zero overhead), but the `big.matrix` path is the only viable option once the design exceeds available memory.

# Negative-binomial regression

`fastglm_nb()` against `MASS::glm.nb()` across a range of sample sizes. The data-generating process is NB(`mu`, `theta = 2`) with three covariates and an intercept:


``` r
set.seed(4)
run_one_nb <- function(n) {
    X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
    mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
    y  <- MASS::rnegbin(n, mu = mu, theta = 2)
    df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])
    microbenchmark(
        fastglm_nb = fastglm_nb(X, y),
        glm.nb     = MASS::glm.nb(y ~ x1 + x2 + x3, data = df),
        times = 5L
    )
}

nb_tim <- list()
for (n in c(1e3, 1e4, 1e5, 5e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_nb(n)
    print(.fmt(mb))
    nb_tim[[length(nb_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_nb 0.682240    0.765306 0.8300368 1.100358
#> 2     glm.nb 7.274835    7.453636 7.4239028 7.490864
#> 
#> --- n = 10000 ---
#>         expr  min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_nb  7.160486    11.09415  10.32162 11.87975
#> 2     glm.nb 57.768180    59.11031  62.14795 68.69858
#> 
#> --- n = 1e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_nb  64.22326    66.32877  68.20897  78.70184
#> 2     glm.nb 499.97466   509.85091 534.50479 588.59686
#> 
#> --- n = 5e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_nb  338.1368    345.7159  347.7577  360.6254
#> 2     glm.nb 2014.3564   2091.0658 2082.6364 2177.4795
nb_tim <- do.call(rbind, nb_tim)
.plot_scaling(nb_tim, "Negative-binomial joint MLE: median time vs n")
```

![plot of chunk unnamed-chunk-6](benchmarks-figs/unnamed-chunk-6-1.png)

Accuracy on the largest case: coefficients and `theta` agree to floating-point precision against `glm.nb`, since both maximize the same likelihood. The runtime difference is structural &mdash; `glm.nb` runs its IRLS + theta-MLE alternation in R, *fastglm_nb* runs both loops in C++.


``` r
set.seed(99)
n  <- 5e5
X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y  <- MASS::rnegbin(n, mu = mu, theta = 2)
df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])
f_nb <- fastglm_nb(X, y)
m_nb <- MASS::glm.nb(y ~ x1 + x2 + x3, data = df)
cat("coef max abs diff:", max(abs(unname(coef(f_nb)) - unname(coef(m_nb)))), "\n")
#> coef max abs diff: 2.198242e-14
cat("theta diff       :", abs(f_nb$theta - m_nb$theta), "\n")
#> theta diff       : 9.67848e-12
cat("logLik diff      :",
    abs(as.numeric(logLik(f_nb)) - as.numeric(logLik(m_nb))), "\n")
#> logLik diff      : 2.095476e-09
```

# Firth bias-reduced GLMs

Firth bias reduction (Firth, 1993; Kosmidis and Firth, 2009) is now supported for all standard GLM families via `firth = TRUE`. We benchmark against `brglm2::brglmFit(type = "AS_mean")`, which implements the same AS_mean adjustment, and against `logistf::logistf()` for the binomial-logit special case.

First, the canonical logistic benchmark on the *sex2* dataset and a larger simulated example:


``` r
data(sex2, package = "logistf")
X_sex2 <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y_sex2 <- sex2$case

mb_firth_logistf <- microbenchmark(
    fastglm = fastglm(X_sex2, y_sex2, family = binomial(), firth = TRUE),
    logistf = logistf(case ~ age + oc + vic + vicl + vis + dia, data = sex2),
    brglm2  = glm(case ~ age + oc + vic + vicl + vis + dia, data = sex2,
                  family = binomial(), method = "brglmFit", type = "AS_mean"),
    times = 10L
)
cat("\n--- Heinze-Schemper sex2 (n =", nrow(sex2), ") ---\n")
#> 
#> --- Heinze-Schemper sex2 (n = 239 ) ---
print(.fmt(mb_firth_logistf))
#>      expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm 0.255840   0.2977215 0.3056796 0.410082
#> 2 logistf 6.808460   7.2490050 7.3370607 8.907332
#> 3  brglm2 2.737611   2.9573505 3.1318014 5.136767
```

Next, a comprehensive comparison of `fastglm(firth = TRUE)` against `brglm2` across all supported families and links. We fix `n = 10000` and `p = 5` to see the per-family speedup:


``` r
set.seed(123)
n <- 10000;  p_firth <- 5
X_firth <- cbind(1, matrix(rnorm(n * p_firth), n, p_firth))
eta_firth <- X_firth %*% c(0.5, 0.3, -0.2, 0.1, 0.05, -0.1)

firth_families <- list(
    list(fam = binomial("logit"),       label = "binomial logit",
         y = rbinom(n, 1, plogis(eta_firth))),
    list(fam = binomial("probit"),      label = "binomial probit",
         y = rbinom(n, 1, plogis(eta_firth))),
    list(fam = binomial("cloglog"),     label = "binomial cloglog",
         y = rbinom(n, 1, plogis(eta_firth))),
    list(fam = poisson("log"),          label = "poisson log",
         y = rpois(n, exp(eta_firth))),
    list(fam = poisson("sqrt"),         label = "poisson sqrt",
         y = rpois(n, exp(eta_firth))),
    list(fam = Gamma("log"),            label = "Gamma log",
         y = rgamma(n, shape = 5, rate = 5 / exp(eta_firth))),
    list(fam = gaussian("identity"),    label = "gaussian identity",
         y = as.numeric(eta_firth) + rnorm(n, 0, 0.5)),
    list(fam = gaussian("log"),         label = "gaussian log",
         y = pmax(as.numeric(eta_firth) + rnorm(n, 0, 0.5), 0.01))
)

xnames <- paste0("x", seq_len(p_firth))
df_firth <- data.frame(X_firth[, -1])
names(df_firth) <- xnames
fml_firth <- as.formula(paste("y ~", paste(xnames, collapse = " + ")))

firth_all_tim <- list()
for (fi in firth_families) {
    df_firth$y <- fi$y
    mb <- microbenchmark(
        fastglm = fastglm(X_firth, fi$y, family = fi$fam, firth = TRUE),
        brglm2  = glm(fml_firth, data = df_firth, family = fi$fam,
                      method = "brglmFit", type = "AS_mean"),
        times = 10L
    )
    s <- summary(mb, unit = bench_unit)
    cat(sprintf("%-20s fastglm=%7.2f ms  brglm2=%7.2f ms  speedup=%5.1fx\n",
                fi$label, s$median[1], s$median[2], s$median[2] / s$median[1]))
    firth_all_tim[[length(firth_all_tim) + 1]] <-
        transform(.collect(mb, n), family = fi$label)
}
#> binomial logit       fastglm=   3.20 ms  brglm2=  13.82 ms  speedup=  4.3x
#> binomial probit      fastglm=   4.38 ms  brglm2=  16.65 ms  speedup=  3.8x
#> binomial cloglog     fastglm=   4.61 ms  brglm2=  17.95 ms  speedup=  3.9x
#> poisson log          fastglm=   3.39 ms  brglm2= 120.02 ms  speedup= 35.4x
#> poisson sqrt         fastglm=   4.93 ms  brglm2= 117.49 ms  speedup= 23.9x
#> Gamma log            fastglm=   4.42 ms  brglm2=  99.60 ms  speedup= 22.5x
#> gaussian identity    fastglm=   1.17 ms  brglm2=  19.82 ms  speedup= 17.0x
#> gaussian log         fastglm=   4.49 ms  brglm2=  28.43 ms  speedup=  6.3x
firth_all_tim <- do.call(rbind, firth_all_tim)
```

A grouped barplot showing the median time for each family:


``` r
op <- par(mar = c(7, 4.2, 3, 1), las = 2)
m <- with(firth_all_tim, tapply(median, list(expr, family), identity))
barplot(m, beside = TRUE, log = "y",
        col = c("#1f77b4", "#d62728"),
        ylab = paste0("median time (", bench_unit, ", log scale)"),
        main = paste0("Firth bias-reduced GLMs: fastglm vs brglm2 (n=", n, ")"),
        legend.text = rownames(m),
        args.legend = list(x = "topleft", bty = "n"))
```

![plot of chunk unnamed-chunk-10](benchmarks-figs/unnamed-chunk-10-1.png)

``` r
par(op)
```

Scaling with sample size for a representative subset of families:


``` r
set.seed(123)
firth_scaling_fams <- list(
    list(fam = binomial("logit"),    label = "binomial logit"),
    list(fam = poisson("log"),       label = "poisson log"),
    list(fam = Gamma("log"),         label = "Gamma log"),
    list(fam = gaussian("identity"), label = "gaussian identity")
)

firth_scale_tim <- list()
for (n_val in c(1e3, 1e4, 5e4)) {
    X_s <- cbind(1, matrix(rnorm(n_val * p_firth), n_val, p_firth))
    eta_s <- X_s %*% c(0.5, 0.3, -0.2, 0.1, 0.05, -0.1)
    df_s <- data.frame(X_s[, -1])
    names(df_s) <- xnames

    cat(sprintf("\n--- n = %d ---\n", n_val))
    for (fi in firth_scaling_fams) {
        y_s <- switch(fi$fam$family,
            "binomial" = rbinom(n_val, 1, plogis(eta_s)),
            "poisson"  = rpois(n_val, exp(eta_s)),
            "Gamma"    = rgamma(n_val, shape = 5, rate = 5 / exp(eta_s)),
            "gaussian" = as.numeric(eta_s) + rnorm(n_val, 0, 0.5)
        )
        df_s$y <- y_s
        mb <- microbenchmark(
            fastglm = fastglm(X_s, y_s, family = fi$fam, firth = TRUE),
            brglm2  = glm(fml_firth, data = df_s, family = fi$fam,
                          method = "brglmFit", type = "AS_mean"),
            times = 5L
        )
        s <- summary(mb, unit = bench_unit)
        cat(sprintf("  %-20s fastglm=%7.2f ms  brglm2=%7.2f ms\n",
                    fi$label, s$median[1], s$median[2]))
        firth_scale_tim[[length(firth_scale_tim) + 1]] <-
            transform(.collect(mb, n_val), family = fi$label)
    }
}
#> 
#> --- n = 1000 ---
#>   binomial logit       fastglm=   0.37 ms  brglm2=   2.37 ms
#>   poisson log          fastglm=   0.42 ms  brglm2=  13.57 ms
#>   Gamma log            fastglm=   0.44 ms  brglm2=  11.62 ms
#>   gaussian identity    fastglm=   0.52 ms  brglm2=   3.47 ms
#> 
#> --- n = 10000 ---
#>   binomial logit       fastglm=   3.81 ms  brglm2=  13.07 ms
#>   poisson log          fastglm=   3.54 ms  brglm2= 119.40 ms
#>   Gamma log            fastglm=   4.71 ms  brglm2=  95.07 ms
#>   gaussian identity    fastglm=   1.15 ms  brglm2=  20.10 ms
#> 
#> --- n = 50000 ---
#>   binomial logit       fastglm=  17.11 ms  brglm2=  71.58 ms
#>   poisson log          fastglm=  19.66 ms  brglm2= 598.03 ms
#>   Gamma log            fastglm=  21.66 ms  brglm2= 506.24 ms
#>   gaussian identity    fastglm=   5.74 ms  brglm2= 103.94 ms
firth_scale_tim <- do.call(rbind, firth_scale_tim)

# one panel per family
op <- par(mfrow = c(2, 2), mar = c(4, 4.2, 2.5, 1), las = 1)
for (flab in unique(firth_scale_tim$family)) {
    sub <- firth_scale_tim[firth_scale_tim$family == flab, ]
    .plot_scaling(sub, paste0("Firth ", flab))
}
```

![plot of chunk unnamed-chunk-11](benchmarks-figs/unnamed-chunk-11-1.png)

``` r
par(op)
```

## Firth across all decomposition methods

Firth bias reduction now works with all six decomposition methods (0&ndash;5), not just the LLT Cholesky. Here we verify that all methods produce identical coefficients and compare their timings:


``` r
set.seed(123)
n_meth <- 10000;  p_meth <- 10
X_meth <- cbind(1, matrix(rnorm(n_meth * (p_meth - 1)), n_meth, p_meth - 1))
y_meth <- rbinom(n_meth, 1, plogis(X_meth %*% rnorm(p_meth) * 0.1))

ref_coef <- coef(fastglm(X_meth, y_meth, family = binomial(), method = 2, firth = TRUE))
method_names <- c("ColPivQR", "QR", "LLT", "LDLT", "FullPivQR", "SVD")
for (m in c(0L, 1L, 3L, 4L, 5L)) {
    d <- max(abs(coef(fastglm(X_meth, y_meth, family = binomial(), method = m,
                               firth = TRUE)) - ref_coef))
    cat(sprintf("  method %d (%s) vs LLT: max |diff| = %.2e\n",
                m, method_names[m + 1], d))
}
#>   method 0 (ColPivQR) vs LLT: max |diff| = 5.83e-16
#>   method 1 (QR) vs LLT: max |diff| = 6.38e-16
#>   method 3 (LDLT) vs LLT: max |diff| = 1.11e-16
#>   method 4 (FullPivQR) vs LLT: max |diff| = 6.66e-16
#>   method 5 (SVD) vs LLT: max |diff| = 2.90e-16

mb_methods <- microbenchmark(
    ColPivQR = fastglm(X_meth, y_meth, family = binomial(), method = 0, firth = TRUE),
    QR       = fastglm(X_meth, y_meth, family = binomial(), method = 1, firth = TRUE),
    LLT      = fastglm(X_meth, y_meth, family = binomial(), method = 2, firth = TRUE),
    LDLT     = fastglm(X_meth, y_meth, family = binomial(), method = 3, firth = TRUE),
    FullPivQR= fastglm(X_meth, y_meth, family = binomial(), method = 4, firth = TRUE),
    SVD      = fastglm(X_meth, y_meth, family = binomial(), method = 5, firth = TRUE),
    times = 5L
)
cat(sprintf("\n--- Firth logistic, n=%d, p=%d ---\n", n_meth, p_meth))
#> 
#> --- Firth logistic, n=10000, p=10 ---
print(.fmt(mb_methods))
#>        expr    min (ms) median (ms)   mean (ms)    max (ms)
#> 1  ColPivQR    5.824173    6.964875    7.052623    7.998116
#> 2        QR    5.723559    6.578983    6.470136    7.566837
#> 3       LLT    4.114678    4.943903    5.083926    5.922204
#> 4      LDLT    4.015253    5.763944    5.633712    7.315876
#> 5 FullPivQR 2266.558556 2275.998642 2288.749527 2328.341866
#> 6       SVD   11.065531   13.015122   12.881757   14.022533
```

All six methods agree to machine precision. The LLT and LDLT Cholesky methods are the fastest because they only decompose the `p x p` cross-product, while the QR and SVD methods work with the full `n x p` matrix.

## Firth on sparse and streaming designs

Firth bias reduction is now also supported on sparse (`dgCMatrix`), `big.matrix`, and streaming callback designs. The sparse and streaming paths use *lagged leverages* from the previous IRLS iteration; at convergence, the result matches the exact-leverage dense path.


``` r
set.seed(123)
n_sp <- 5000;  p_sp <- 6
X_sp <- cbind(1, matrix(rnorm(n_sp * (p_sp - 1)), n_sp, p_sp - 1))
Xs_sp <- as(X_sp, "CsparseMatrix")
Xb_sp <- as.big.matrix(X_sp)
y_sp <- rbinom(n_sp, 1, plogis(X_sp %*% c(-0.3, 0.4, -0.2, 0.1, 0.05, -0.15)))

chunk_size_sp <- 500
chunks_sp <- function(k) {
    idx <- ((k - 1) * chunk_size_sp + 1):(k * chunk_size_sp)
    list(X = X_sp[idx, , drop = FALSE], y = y_sp[idx])
}

mb_backends <- microbenchmark(
    dense      = fastglm(X_sp,  y_sp, family = binomial(), method = 2, firth = TRUE),
    sparse     = fastglm(Xs_sp, y_sp, family = binomial(), method = 2, firth = TRUE),
    big.matrix = fastglm(Xb_sp, y_sp, family = binomial(), method = 2, firth = TRUE),
    streaming  = fastglm_streaming(chunks_sp, n_chunks = n_sp / chunk_size_sp,
                                   family = binomial(), method = 2, firth = TRUE),
    times = 5L
)
cat(sprintf("\n--- Firth logistic across backends (n=%d, p=%d) ---\n", n_sp, p_sp))
#> 
#> --- Firth logistic across backends (n=5000, p=6) ---
print(.fmt(mb_backends))
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1      dense 1.264399    2.091492  1.896914 2.365290
#> 2     sparse 3.416571    4.240179  5.136956 7.754166
#> 3 big.matrix 1.602649    1.866033  2.484387 3.958509
#> 4  streaming 1.776735    1.856193  1.872568 2.055453

# Verify agreement
ref_sp <- coef(fastglm(X_sp, y_sp, family = binomial(), method = 2, firth = TRUE))
sp_c   <- coef(fastglm(Xs_sp, y_sp, family = binomial(), method = 2, firth = TRUE))
bm_c   <- coef(fastglm(Xb_sp, y_sp, family = binomial(), method = 2, firth = TRUE))
st_c   <- coef(fastglm_streaming(chunks_sp, n_chunks = n_sp / chunk_size_sp,
                                  family = binomial(), method = 2, firth = TRUE))
cat(sprintf("  sparse  vs dense: max |diff| = %.2e\n", max(abs(sp_c - ref_sp))))
#>   sparse  vs dense: max |diff| = 2.09e-10
cat(sprintf("  big.mat vs dense: max |diff| = %.2e\n", max(abs(bm_c - ref_sp))))
#>   big.mat vs dense: max |diff| = 2.09e-10
cat(sprintf("  stream  vs dense: max |diff| = %.2e\n", max(abs(st_c - ref_sp))))
#>   stream  vs dense: max |diff| = 3.16e-11
```

All backends converge to the same Firth-penalized MLE to high precision ($< 10^{-7}$). The dense path is fastest when the matrix fits in RAM; the `big.matrix` and streaming paths are designed for datasets that exceed available memory.

# Hurdle models

`fastglm_hurdle` against `pscl::hurdle` across a range of sample sizes, Poisson count component:


``` r
set.seed(6)
run_one_hurdle <- function(n) {
    x1 <- rnorm(n);  x2 <- rnorm(n)
    lam    <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
    is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
    yt     <- integer(n)
    for (i in seq_len(n)) {
        repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
    }
    y  <- ifelse(is_pos == 1, yt, 0L)
    df <- data.frame(y = y, x1 = x1, x2 = x2)
    microbenchmark(
        fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson"),
        pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "poisson"),
        times = 5L
    )
}

hurdle_tim <- list()
for (n in c(1e3, 1e4, 5e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_hurdle(n)
    print(.fmt(mb))
    hurdle_tim[[length(hurdle_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_hurdle 0.506268    0.530540  1.275371 4.253299
#> 2    pscl_hurdle 4.691671    4.916761  5.036055 5.681165
#> 
#> --- n = 10000 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  2.825146    3.026579  3.105086  3.699758
#> 2    pscl_hurdle 43.152705   43.511332 45.518069 54.041649
#> 
#> --- n = 50000 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  14.37255    15.69447  17.12776  24.45412
#> 2    pscl_hurdle 269.05057   272.69416 273.68238 279.32353
#> 
#> --- n = 1e+05 ---
#>             expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_hurdle  30.55332    32.89479  33.47465  38.67702
#> 2    pscl_hurdle 473.65229   475.67302 484.73010 505.84525
hurdle_tim <- do.call(rbind, hurdle_tim)
.plot_scaling(hurdle_tim, "Hurdle Poisson: median time vs n")
```

![plot of chunk unnamed-chunk-14](benchmarks-figs/unnamed-chunk-14-1.png)

NB hurdle (with simultaneous `theta` MLE) on a single moderately large example:


``` r
set.seed(7)
n <- 1e4
x1 <- rnorm(n);  x2 <- rnorm(n)
lam <- exp(0.8 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1))
yt <- integer(n)
for (i in seq_len(n)) {
    repeat {
        v <- MASS::rnegbin(1, mu = lam[i], theta = 1.5)
        if (v > 0) { yt[i] <- v; break }
    }
}
y  <- ifelse(is_pos == 1, yt, 0L)
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_hurdle_nb <- microbenchmark(
    fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "negbin"),
    pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "negbin"),
    times = 3L
)
print(.fmt(mb_hurdle_nb))
#>             expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_hurdle 20.10173    20.32481  21.79809 24.96773
#> 2    pscl_hurdle 48.44277    48.83231  48.75877 49.00123
```

# Zero-inflated models

Zero-inflated Poisson against `pscl::zeroinfl()` across sample sizes:


``` r
set.seed(8)
run_one_zi <- function(n) {
    x1 <- rnorm(n);  x2 <- rnorm(n)
    eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
    eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
    z     <- rbinom(n, 1, plogis(eta_z))
    y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
    df    <- data.frame(y = y, x1 = x1, x2 = x2)
    microbenchmark(
        fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson"),
        pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson"),
        times = 5L
    )
}

zi_tim <- list()
for (n in c(1e3, 1e4, 5e4, 1e5)) {
    cat("\n--- n =", n, "---\n")
    mb <- run_one_zi(n)
    print(.fmt(mb))
    zi_tim[[length(zi_tim) + 1]] <- .collect(mb, n)
}
#> 
#> --- n = 1000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_zi 2.177797    2.263118  2.311711 2.557580
#> 2    pscl_zi 8.709425    8.853253  8.976368 9.485883
#> 
#> --- n = 10000 ---
#>         expr min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi 23.18952    23.96737  24.01417  24.75879
#> 2    pscl_zi 94.43276   107.85345 104.04964 112.38129
#> 
#> --- n = 50000 ---
#>         expr min (ms) median (ms) mean (ms) max (ms)
#> 1 fastglm_zi 109.7370    113.3301  113.4383 118.1588
#> 2    pscl_zi 525.3278    539.9472  550.7139 604.0981
#> 
#> --- n = 1e+05 ---
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi  214.1678    215.4779  216.9463  223.3263
#> 2    pscl_zi 1014.7380   1045.3144 1046.7203 1108.0405
zi_tim <- do.call(rbind, zi_tim)
.plot_scaling(zi_tim, "Zero-inflated Poisson: median time vs n")
```

![plot of chunk unnamed-chunk-16](benchmarks-figs/unnamed-chunk-16-1.png)

Zero-inflated NB on a single moderately large example:


``` r
set.seed(9)
n  <- 1e4
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.8 + 0.4 * x1 - 0.3 * x2; lam <- exp(eta_c)
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z <- rbinom(n, 1, plogis(eta_z))
y <- ifelse(z == 1, 0L, MASS::rnegbin(n, mu = lam, theta = 2.0))
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_zi_nb <- microbenchmark(
    fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "negbin",
                            em.tol = 1e-8, em.maxit = 200L),
    pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "negbin"),
    times = 3L
)
print(.fmt(mb_zi_nb))
#>         expr  min (ms) median (ms) mean (ms)  max (ms)
#> 1 fastglm_zi  52.92965    61.05474  60.96088  68.89825
#> 2    pscl_zi 141.41302   151.03523 147.94690 151.39246
```

# Summary

A compact summary of the speed advantage *fastglm* delivers, expressed as the ratio of reference-implementation median time to *fastglm* median time. Larger is better. The reference for each model class is the canonical R implementation; for the standard GLMs we report the ratio against the fastest among `glm.fit`, `glm2`, and `speedglm` so the comparison is conservative.


``` r
.speedup <- function(df, fast_label, ref_labels) {
    out <- lapply(split(df, df$n), function(sub) {
        fast <- sub$median[sub$expr == fast_label]
        if (length(fast) == 0) return(NULL)
        ref  <- min(sub$median[sub$expr %in% ref_labels])
        data.frame(n = sub$n[1], speedup = ref / fast)
    })
    do.call(rbind, out)
}

su_logit  <- .speedup(logit_tim,  "fastglm_chol",
                      c("glm.fit", "glm2_fit2", "speedglm"))
su_pois   <- .speedup(pois_tim,   "fastglm_chol",
                      c("glm.fit", "glm2_fit2", "speedglm"))
su_nb     <- .speedup(nb_tim,     "fastglm_nb",     "glm.nb")
firth_avg_tim <- aggregate(median ~ n + expr, data = firth_scale_tim, FUN = mean)
su_firth  <- .speedup(firth_avg_tim, "fastglm",  "brglm2")
su_hurdle <- .speedup(hurdle_tim, "fastglm_hurdle", "pscl_hurdle")
su_zi     <- .speedup(zi_tim,     "fastglm_zi",     "pscl_zi")

su_all <- rbind(
    transform(su_logit,  model = "logit"),
    transform(su_pois,   model = "poisson"),
    transform(su_nb,     model = "neg-binomial"),
    transform(su_firth,  model = "Firth (avg)"),
    transform(su_hurdle, model = "hurdle"),
    transform(su_zi,     model = "zero-inflated")
)

op <- par(mar = c(4.2, 4.5, 3, 1), las = 1)
models <- unique(su_all$model)
cols   <- seq_along(models)
plot(NA, log = "xy",
     xlim = range(su_all$n), ylim = range(su_all$speedup),
     xlab = "n (log scale)",
     ylab = "speedup over reference (x, log scale)",
     main = "fastglm speedup vs canonical reference, by model class")
abline(h = 1, col = "gray70", lty = 2)
grid(equilogs = FALSE, col = "gray85")
for (k in seq_along(models)) {
    sub <- su_all[su_all$model == models[k], ]
    sub <- sub[order(sub$n), ]
    lines (sub$n, sub$speedup, col = cols[k], lwd = 2)
    points(sub$n, sub$speedup, col = cols[k], pch = 19)
}
legend("topleft", legend = models, col = cols, pch = 19, lty = 1, lwd = 2,
       bty = "n")
```

![plot of chunk unnamed-chunk-18](benchmarks-figs/unnamed-chunk-18-1.png)

``` r
par(op)
```

Across all model classes the same picture holds:

- *fastglm* matches the canonical reference implementation to floating-point precision (or to within the EM tolerance for zero-inflation), so the numerical answer is the same.
- The runtime gap grows with sample size, since R-side overhead in the reference implementations is fixed-per-iteration. By `n = 1e5` the speedup is generally an order of magnitude or more.
- For models with an outer iteration (NB joint MLE, hurdle/ZI with NB), the gap is widest, since the entire outer loop is in C++ in *fastglm* and entirely in R for `MASS::glm.nb`, `pscl::hurdle`, `pscl::zeroinfl`.

# References

- Enea, M. (2009). Fitting linear models and generalized linear models with large data sets in R. In *Statistical Methods for the Analysis of Large Data-sets*, Italian Statistical Society. *speedglm* package documentation.

- Firth, D. (1993). Bias reduction of maximum likelihood estimates. *Biometrika*, 80(1), 27&ndash;38. <doi:10.1093/biomet/80.1.27>

- Heinze, G. and Schemper, M. (2002). A solution to the problem of separation in logistic regression. *Statistics in Medicine*, 21(16), 2409&ndash;2419. <doi:10.1002/sim.1047>

- Kosmidis, I. and Firth, D. (2009). Bias reduction in exponential family nonlinear models. *Biometrika*, 96(4), 793&ndash;804. <doi:10.1093/biomet/asp055>

- Marschner, I. C. (2011). glm2: Fitting generalized linear models with convergence problems. *The R Journal*, 3(2), 12&ndash;15. <doi:10.32614/RJ-2011-012>

- Venables, W. N. and Ripley, B. D. (2002). *Modern Applied Statistics with S* (4th ed.). Springer.

- Zeileis, A., Kleiber, C., and Jackman, S. (2008). Regression models for count data in R. *Journal of Statistical Software*, 27(8), 1&ndash;25. <doi:10.18637/jss.v027.i08>

---
title: "Quick Usage Guide to the 'fastglm' Package"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Usage Guide to the 'fastglm' Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

The *fastglm* package is intended to be a **fast** *and* **stable** alternative to the `glm()` and `glm2::glm2()` functions for fitting generalized linear models. *fastglm* is compatible with `R`'s `family` objects (see `?family`). The package can be installed via

```{r, eval = FALSE, echo=TRUE}
pak::pak("fastglm")
# pak::pak("jaredhuling/fastglm") #development version
```

and loaded via:

```{r, echo=TRUE}
library(fastglm)
```

## Example

The main package function is `fastglm()`. Currently, `fastglm()` does not allow for formula-based data input and is restricted to matrices. We use the following example to demonstrate the usage of `fastglm()`:

```{r, echo = TRUE}
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)

y <- cbind(esoph$ncases, esoph$ncontrols)

gfit1 <- fastglm(x = x, y = y, family = binomial(link = "cloglog"))

summary(gfit1)
```

Alternatively, to use the built-in function of `glm()` for processing formulas and data frames, we can use `glm()` with `method = fastglm_fit` to call the *fastglm* fitting functionality:

```{r}
gfit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) +
                 unclass(alcgp), data = esoph,
             family = binomial(link = "cloglog"),
             method = fastglm_fit)

summary(gfit2, refit = FALSE)
```

Setting `refit = FALSE` in `summary()` uses the standard errors computed by the internal fitting functions of *fastglm*. The default, `refit = TRUE`, re-fits the model using `glm()` with `method = "glm.fit"` and one IRLS iteration with the coefficients computed previously as the starting values.

## Computational stability 

*fastglm* does not compromise computational stability for speed. In fact, for many situations where `glm()` and even `glm2::glm2()` do not converge, `fastglm()` does converge.

As an example, consider the following data scenario, where the response distribution is (mildly) misspecified, but the link function is quite badly misspecified. In such scenarios, the standard IRLS algorithm tends to have convergence issues. The *glm2* package was designed to handle such cases, however, it still can have convergence issues. *fastglm* uses a similar step-halving technique as `glm2::glm2()`, but it starts at better initialized values and thus tends to have better convergence properties in practice. 

```{r, fig.show='hold'}
set.seed(1)
x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- exp(0.25 * x[,1] - 0.25 * x[,3] + 0.5 * x[,4] - 0.5 * x[,5] + rnorm(10000)) + 0.1


# fastglm
system.time(gfit1 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         method = fastglm_fit))

# glm
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt")))

# glm2
system.time(gfit3 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         method = glm2::glm.fit2))

## Note that fastglm() returns estimates with the
## largest likelihood
logLik(gfit1)
logLik(gfit2)
logLik(gfit3)

coef(gfit1)[1:5]
coef(gfit2)[1:5]
coef(gfit3)[1:5]

## check convergence of fastglm
gfit1$converged
## number of IRLS iterations
gfit1$iter

## now check convergence for glm()
gfit2$converged
gfit2$iter

## check convergence for glm2()
gfit3$converged
gfit3$iter


## increasing number of IRLS iterations for glm() does not help that much
system.time(gfit2 <- glm(y ~ x, family = Gamma(link = "sqrt"),
                         control = list(maxit = 100)))

gfit2$converged
gfit2$iter

logLik(gfit1)
logLik(gfit2)

```

---
title: "Negative-Binomial, Hurdle, and Zero-Inflation with 'fastglm'"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Negative-Binomial, Hurdle, and Zero-Inflation with 'fastglm'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

Beyond `family`-driven GLMs, *fastglm* has fitting functions for three count-data model types that arise often in practice but require model-specific machinery on top of the standard IRLS solver:

- **`fastglm_nb()`** — negative-binomial regression with the dispersion `theta` estimated jointly with the regression coefficients, in the spirit of `MASS::glm.nb()`.
- **`fastglm_hurdle()`** — a two-part count model with a binary zero / non-zero component and a zero-truncated Poisson or NB count component, the same model as `pscl::hurdle()`.
- **`fastglm_zi()`** — a zero-inflated Poisson or NB regression, with a binary inflation component layered on the original (untruncated) count distribution, the same model as `pscl::zeroinfl()`.

All three reuse the same C++ IRLS solver as `fastglm()` itself; the outer iteration (joint NB MLE, the EM loop in zero-inflation, the inner theta-Brent for unknown-`theta` NB hurdle / ZI) likewise lives in C++. Cameron and Trivedi (1998) is the standard textbook reference for the count-data theory underlying this vignette, and Zeileis, Kleiber, and Jackman (2008) gives the *pscl* implementation we benchmark against.

```{r}
library(fastglm)
suppressPackageStartupMessages({
    library(MASS)        # glm.nb, rnegbin
    library(pscl)        # hurdle, zeroinfl, bioChemists
    library(microbenchmark)
})
```

# Negative-binomial regression

`fastglm_nb()` is a drop-in alternative to `MASS::glm.nb()` (Venables and Ripley, 2002) for the NB2 model with `Var(Y) = mu + mu^2 / theta`. Both `theta` and `beta` are estimated by maximum likelihood, alternating an IRLS update for `beta` at fixed `theta` with a 1-D Brent root-find for `theta` at fixed `beta`. Both loops run in C++.

A small-sample comparison on `MASS::quine` (school absences):

```{r}
data(quine)
X <- model.matrix(~ Eth + Sex + Age + Lrn, data = quine)
y <- quine$Days

fit_f <- fastglm_nb(X, y)
fit_m <- MASS::glm.nb(Days ~ Eth + Sex + Age + Lrn, data = quine)

c(theta_fastglm = fit_f$theta, theta_glm.nb = fit_m$theta)

max(abs(unname(coef(fit_f)) - unname(coef(fit_m))))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_m)))
```

Coefficients and `theta` agree to roughly 1e-8 on the same MLE.

A timing comparison, on a moderately sized simulated NB(`theta = 2`) example:

```{r}
set.seed(1)
n  <- 5e4
X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y  <- MASS::rnegbin(n, mu = mu, theta = 2)
df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])

mb_nb <- microbenchmark(
    fastglm_nb = fastglm_nb(X, y),
    glm.nb     = MASS::glm.nb(y ~ x1 + x2 + x3, data = df),
    times = 5L
)
print(mb_nb)
```

The native NB family kernel is also exposed directly through `negbin(theta, link)` for the case when `theta` is known (or estimated separately). Holding `theta` fixed at the joint MLE recovers the `fastglm_nb()` regression coefficients exactly, since `fastglm_nb()` is just IRLS at the converged `theta`:

```{r}
fit_joint <- fastglm_nb(X, y)
fit_known <- fastglm(X, y, family = negbin(theta = fit_joint$theta, link = "log"))
max(abs(unname(coef(fit_known)) - unname(coef(fit_joint))))
```

# Hurdle models

Hurdle models, introduced by Mullahy (1986), factorize a count distribution into two independent pieces:

- a binary regression for `1(y > 0)` over the full sample (the *zero* part);
- a zero-truncated Poisson or NB regression for `y > 0` over the positive subset (the *count* part).

Because the two parts share no parameters, the joint likelihood factorizes and they can be fit independently. *fastglm*'s C++ driver fits both parts using the same IRLS solver as `fastglm()`, with new `FAM_POIS_TRUNC_*` / `FAM_NB_TRUNC_*` family codes that handle the truncation correction stably (`expm1`, `log1p` near `mu = 0`).

The formula uses the *Formula* package convention `y ~ x1 + x2 | z1 + z2`; the right-hand side after `|` specifies the zero-part design (it defaults to the count-part design if absent).

```{r}
data(bioChemists, package = "pscl")
fit_f <- fastglm_hurdle(art ~ ., data = bioChemists, dist = "poisson")
fit_p <- pscl::hurdle (art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))
```

`coef()` and `vcov()` accept a `model = c("full", "count", "zero")` argument so each part can be inspected separately:

```{r}
coef(fit_f, model = "count")
coef(fit_f, model = "zero")
```

For NB count parts, the dispersion is estimated by an inner Brent MLE that runs *between* outer IRLS iterations; the outer-loop tolerance is controlled by `outer.tol` / `outer.maxit`.

A timing comparison on a 4000-observation simulated example:

```{r}
set.seed(11)
n  <- 4000
x1 <- rnorm(n);  x2 <- rnorm(n)
lam    <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
yt     <- integer(n)
for (i in seq_len(n)) {
    repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
}
y  <- ifelse(is_pos == 1, yt, 0L)
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_hurdle <- microbenchmark(
    fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_hurdle)
```

# Zero-inflated models

Zero inflation, introduced by Lambert (1992), differs from a hurdle in that the two components share a latent variable: a `y = 0` outcome could come from the inflation component (with probability `pi_i`) *or* from the count component (with probability `1 - pi_i` and a count-side zero). The likelihood for `y = 0` is

$$
\Pr(Y_i = 0) = \pi_i + (1 - \pi_i) f(0; \mu_i)
$$

and for `y > 0` it is `(1 - pi_i) f(y; mu_i)`. The two-component mixture rules out a closed-form factorization, so *fastglm* uses an EM algorithm:

- **E-step**: posterior `tau_i = P(Z_i = 1 | y_i)` for each observation, computed in log-space with `logsumexp` to avoid catastrophic cancellation when `pi_i` is near 0 or 1.
- **M-step**: a binomial logit / probit / cloglog / log fit with response `tau` (continuous), weights 1; followed by a Poisson or NB fit on the *original* `y` with prior weights `1 - tau`. For NB, an inner Brent MLE re-estimates `theta` after each count-side step.

The final observed-information `vcov` comes from a numerical Jacobian of the analytical observed score at the EM fixed point, which is stable and cheap (block-diagonal per `(gamma, beta, theta)`). The complete EM driver — E-step, both M-steps, the inner theta MLE, and the score Jacobian — runs in C++.

The formula syntax matches `fastglm_hurdle()` exactly:

```{r}
fit_f <- fastglm_zi(art ~ ., data = bioChemists, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 300L)
fit_p <- pscl::zeroinfl(art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))
```

EM is iterative, so the agreement is slightly looser than for hurdle (where the two parts are closed-form independent fits), but coefficients and the joint log-likelihood still match to about 1e-5.

A timing comparison on a simulated zero-inflated Poisson:

```{r}
set.seed(21)
n  <- 3000
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z     <- rbinom(n, 1, plogis(eta_z))
y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
df    <- data.frame(y = y, x1 = x1, x2 = x2)

mb_zi <- microbenchmark(
    fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_zi)
```

# References

- Cameron, A. C. and Trivedi, P. K. (1998). *Regression Analysis of Count Data*. Cambridge University Press.

- Mullahy, J. (1986). Specification and testing of some modified count data models. *Journal of Econometrics*, 33(3), 341–365. <doi:10.1016/0304-4076(86)90002-3>

- Lambert, D. (1992). Zero-inflated Poisson regression, with an application to defects in manufacturing. *Technometrics*, 34(1), 1–14. <doi:10.2307/1269547>

- Zeileis, A., Kleiber, C., and Jackman, S. (2008). Regression models for count data in R. *Journal of Statistical Software*, 27(8), 1–25. <doi:10.18637/jss.v027.i08>

- Venables, W. N. and Ripley, B. D. (2002). *Modern Applied Statistics with S* (4th ed.). Springer.

---
title: "Overview of the 'fastglm' Package"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overview of the 'fastglm' Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

The *fastglm* package is intended as a **fast** *and* **stable** alternative to `stats::glm()` for fitting generalized linear models. It is compatible with `R`'s `family` objects (see `?family`) and produces fitted objects whose downstream methods (`summary()`, `vcov()`, `predict()`, `coef()`, `residuals()`, `logLik()`) behave exactly like those for a `glm`. This vignette walks through the main functionality of the package at a higher level than the other vignettes.

```{r}
library(fastglm)
```

# Fitting a GLM

The main package function is `fastglm()`. It takes a numeric design matrix `x`, a response `y`, and an `R` `family` object:

```{r}
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

fit <- fastglm(x, y, family = binomial(link = "cloglog"))
summary(fit)
```

`fastglm()` does not accept a formula directly, it operates on a pre-built design matrix. To use a formula and a data frame, pass `fastglm_fit` as the fitting function to base `glm()`:

```{r}
fit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) + unclass(alcgp),
            data = esoph, family = binomial(link = "cloglog"),
            method = fastglm_fit)
```

# Decomposition methods

The IRLS algorithm reduces every iteration to a weighted least-squares problem. *fastglm* supports six different matrix decompositions for solving that WLS step, all from *RcppEigen* (Bates and Eddelbuettel, 2013); the choice trades off speed against numerical stability and rank-revealing behavior:

| `method` | decomposition |
|---|---|
| `0` | column-pivoted Householder QR (default; rank-revealing) |
| `1` | unpivoted Householder QR |
| `2` | LLT Cholesky |
| `3` | LDLT Cholesky |
| `4` | full-pivoted Householder QR |
| `5` | bidiagonal divide-and-conquer SVD |

The default (`method = 0`) is the safe choice: it is rank-revealing, so it handles aliased / collinear columns gracefully. The Cholesky methods (`2` and `3`) are roughly 3–4× faster but assume full column rank.

```{r}
set.seed(123)
n <- 5000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rbinom(n, 1, plogis(x %*% rnorm(p) * 0.05))

system.time(f0 <- fastglm(x, y, family = binomial()))                 # default
system.time(f2 <- fastglm(x, y, family = binomial(), method = 2))     # LLT
system.time(f3 <- fastglm(x, y, family = binomial(), method = 3))     # LDLT
```

# Stability

*fastglm* does not compromise computational stability for speed. It uses a step-halving safeguard following Marschner (2011) and starts from better-initialized values than `glm()` or `glm2::glm2()`, so it tends to converge in cases where the standard IRLS algorithm fails. See `vignette("fastglm", package = "fastglm")` for a Gamma-with-`sqrt`-link example where `glm()` does not converge but `fastglm()` does.

# Inference: `vcov()`, robust SE, predictions

The fitted object stores the unscaled covariance directly (no refitting), and the standard `vcov()` / `summary()` machinery works as expected:

```{r}
fit <- fastglm(x, y, family = binomial(), method = 2)

V    <- vcov(fit)
dim(V)

## standard errors agree with the diagonal of vcov()
all.equal(sqrt(diag(V)), unname(coef(summary(fit))[, "Std. Error"]))
```

Heteroskedasticity-consistent and cluster-robust covariance matrices are
available via `sandwich::vcovHC()` and `sandwich::vcovCL()` — `fastglm`
registers methods on those generics, so loading `sandwich` is all that is
required:

```{r}
library(sandwich)

V_hc <- vcovHC(fit, type = "HC0")
V_hc[1:3, 1:3]

cluster <- rep(seq_len(20), length.out = n)
V_cl    <- vcovCL(fit, cluster = cluster, type = "HC1")
V_cl[1:3, 1:3]
```

Results are numerically identical to `sandwich::vcovHC()` and
`sandwich::vcovCL()` applied to a `glm` fit (Zeileis, Köll, and Graham, 2020)
to floating-point precision. They work for sparse and `big.matrix` fits as
well, since the full design matrix is preserved on the fitted object (as a
reference, not copied).

`predict()` supports `se.fit = TRUE`:

```{r}
xnew <- x[1:5, , drop = FALSE]
predict(fit, newdata = xnew, type = "response", se.fit = TRUE)
```

# Sparse, big.matrix, and streaming designs

For designs that are sparse, that live on disc, or that have to be built from a parquet / *arrow* / *DuckDB* source, *fastglm* provides three large-data paths that share a common streaming kernel and produce identical results:

- `Matrix::dgCMatrix`: pass directly to `fastglm()`. Useful for one-hot encoded categoricals and high-dimensional sparse designs.
- `bigmemory::big.matrix`: pass directly to `fastglm()`. The matrix is read in row-blocks and never fully materialised in memory.
- `fastglm_streaming(chunk_callback, n_chunks, family)`: a user-supplied closure yields one row-block per call. The right path for fitting on a parquet dataset, *DuckDB* query, or any external columnar store.

See `vignette("large-data-fastglm", package = "fastglm")` for a detailed walk through all three paths. A short example:

```{r}
n <- 4000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.4, -0.2, 0.3)))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))
```

# Native families

For the most commonly-used `family`/`link` combinations, *fastglm* dispatches `variance()`, `mu.eta()`, `linkinv()`, and `dev.resids()` to inline C++ implementations rather than calling back into `R` once per IRLS iteration. The covered combinations are:

- gaussian (identity, log, inverse)
- binomial (logit, probit, cloglog, log)
- poisson (log, identity, sqrt)
- Gamma (log, inverse, identity)
- inverse.gaussian (1/mu^2, log, identity, inverse)

Detection is automatic, if the `family` object matches one of the above, the native path is used; otherwise *fastglm* falls back to the `R`-callback path used in earlier versions. The native path is meaningfully faster on large `n` because it eliminates the per-iteration round-trip to `R` for each of the four family functions:

```{r, eval = TRUE}
library(microbenchmark)

set.seed(7)
n <- 50000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rpois(n, exp(x %*% rnorm(p) * 0.05 + 1))

## force the R-callback path by giving the family object an unrecognized name
disguise <- function(fam) { fam$family <- paste0(fam$family, "*"); fam }

mb <- microbenchmark(
    native   = fastglm(x, y, family = poisson(),            method = 2),
    callback = fastglm(x, y, family = disguise(poisson()),  method = 2),
    times = 5L
)
print(mb)
```

The *fastglm* package switches transparently between the two paths; user code does not change.

# Negative binomial, hurdle, zero-inflation, and Firth logistic

A handful of regressions that come up often in practice are not standard GLMs but build on the same IRLS machinery. *fastglm* provides dedicated fitting functions for these:

- **`fastglm_nb()`** — negative-binomial regression with the dispersion `theta` estimated jointly with `beta`. Plays the same role as `MASS::glm.nb()` but runs the joint `(beta, theta)` MLE entirely in C++ (IRLS for `beta`, Brent for `theta`).
- **`firth = TRUE`** — a flag on `fastglm()` / `fastglm_fit()` that activates Firth's bias-reducing penalty on the score. Useful for binary logistic regression under separation or in small samples; analogous to `logistf::logistf()`.
- **`fastglm_hurdle()`** — a two-part count model: a binary regression for whether `y > 0`, plus a zero-truncated Poisson or NB regression on the positive subset. The two parts factorize, and both are fit by the same C++ IRLS solver. Same model as `pscl::hurdle()`.
- **`fastglm_zi()`** — a zero-inflated Poisson or NB regression: a binary inflation component overlaid on the original count distribution, fit by an EM algorithm in C++ with closed-form posterior responsibilities and an analytical observed-information `vcov`. Same model as `pscl::zeroinfl()`.

All four take the standard *fastglm* tuning options (`method`, `tol`, `maxit`) and return objects with the usual `coef()`, `vcov()`, `predict()`, `logLik()` methods. `fastglm_hurdle()` and `fastglm_zi()` use the `Formula` package's two-RHS syntax `y ~ x1 + x2 | z1 + z2` to specify different designs for the count and zero parts. See `vignette("count-firth-fastglm", package = "fastglm")` for examples and timing comparisons against the canonical reference packages.

# Three R entry points

There are three `R`-side functions to fit a GLM with *fastglm*. They all call into the same C++ solver, but differ in how much of base R's machinery they wrap around it:

- **`fastglm()`** is the top-level S3 function. It returns an object of class `"fastglm"` with deviance, AIC, null deviance, residuals, etc. populated for interactive use.
- **`fastglm_fit()`** is designed to be passed as `method = fastglm_fit` to base `glm()`. The returned object inherits from `"fastglmFit"` so that `glm()`'s formula/data handling, `update()`, `predict()`, etc. all work on top.
- **`fastglmPure()`** is the lowest-level wrapper: no dispersion, no AIC, no null-deviance computation. Use this when calling *fastglm* from another package and you only need the coefficients.

# References

- Marschner, I. C. (2011). glm2: Fitting generalized linear models with convergence problems. *The R Journal*, 3(2), 12–15. <doi:10.32614/RJ-2011-012>

- Bates, D. and Eddelbuettel, D. (2013). Fast and elegant numerical linear algebra using the RcppEigen package. *Journal of Statistical Software*, 52(5), 1–24. <doi:10.18637/jss.v052.i05>

- Zeileis, A., Köll, S., and Graham, N. (2020). Various versatile variances: An object-oriented implementation of clustered covariances in R. *Journal of Statistical Software*, 95(1), 1–36. <doi:10.18637/jss.v095.i01>

---
title: "Negative-Binomial, Hurdle, Zero-Inflation, and Firth Logistic Regression with 'fastglm'"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Negative-Binomial, Hurdle, Zero-Inflation, and Firth Logistic Regression with 'fastglm'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

Beyond `family`-driven GLMs, *fastglm* has fitting functions for four model types/situations that arise often in practice but require model-specific machinery on top of the standard IRLS solver:

- **`fastglm_nb()`** — negative-binomial regression with the dispersion `theta` estimated jointly with the regression coefficients, in the spirit of `MASS::glm.nb()`.
- **`firth = TRUE`** — Firth's (1993) bias-reducing penalty for logistic regression, useful under separation or in small samples (the `logistf::logistf()` use case).
- **`fastglm_hurdle()`** — a two-part count model with a binary zero / non-zero component and a zero-truncated Poisson or NB count component, the same model as `pscl::hurdle()`.
- **`fastglm_zi()`** — a zero-inflated Poisson or NB regression, with a binary inflation component layered on the original (untruncated) count distribution, the same model as `pscl::zeroinfl()`.

All four reuse the same C++ IRLS solver as `fastglm()` itself; the outer iteration (joint NB MLE, the EM loop in zero-inflation, the inner theta-Brent for unknown-`theta` NB hurdle / ZI) likewise lives in C++. Cameron and Trivedi (1998) is the standard textbook reference for the count-data theory underlying this vignette, and Zeileis, Kleiber, and Jackman (2008) gives the *pscl* implementation we benchmark against.

```{r}
library(fastglm)
suppressPackageStartupMessages({
    library(MASS)        # glm.nb, rnegbin, sex2-style separation data
    library(pscl)        # hurdle, zeroinfl, bioChemists
    library(logistf)     # canonical Firth implementation
    library(microbenchmark)
})
```

# Negative-binomial regression

`fastglm_nb()` is a drop-in alternative to `MASS::glm.nb()` (Venables and Ripley, 2002) for the NB2 model with `Var(Y) = mu + mu^2 / theta`. Both `theta` and `beta` are estimated by maximum likelihood, alternating an IRLS update for `beta` at fixed `theta` with a 1-D Brent root-find for `theta` at fixed `beta`. Both loops run in C++.

A small-sample comparison on `MASS::quine` (school absences):

```{r}
data(quine)
X <- model.matrix(~ Eth + Sex + Age + Lrn, data = quine)
y <- quine$Days

fit_f <- fastglm_nb(X, y)
fit_m <- MASS::glm.nb(Days ~ Eth + Sex + Age + Lrn, data = quine)

c(theta_fastglm = fit_f$theta, theta_glm.nb = fit_m$theta)

max(abs(unname(coef(fit_f)) - unname(coef(fit_m))))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_m)))
```

Coefficients and `theta` agree to roughly 1e-8 on the same MLE.

A timing comparison, on a moderately sized simulated NB(`theta = 2`) example:

```{r}
set.seed(1)
n  <- 5e4
X  <- cbind(1, matrix(rnorm(n * 3), n, 3))
mu <- exp(X %*% c(0.5, 0.4, -0.2, 0.3))
y  <- MASS::rnegbin(n, mu = mu, theta = 2)
df <- data.frame(y = y, x1 = X[, 2], x2 = X[, 3], x3 = X[, 4])

mb_nb <- microbenchmark(
    fastglm_nb = fastglm_nb(X, y),
    glm.nb     = MASS::glm.nb(y ~ x1 + x2 + x3, data = df),
    times = 5L
)
print(mb_nb)
```

The native NB family kernel is also exposed directly through `negbin(theta, link)` for the case when `theta` is known (or estimated separately). Holding `theta` fixed at the joint MLE recovers the `fastglm_nb()` regression coefficients exactly, since `fastglm_nb()` is just IRLS at the converged `theta`:

```{r}
fit_joint <- fastglm_nb(X, y)
fit_known <- fastglm(X, y, family = negbin(theta = fit_joint$theta, link = "log"))
max(abs(unname(coef(fit_known)) - unname(coef(fit_joint))))
```

# Firth bias-reduced logistic regression

Logistic regression breaks down under (quasi-) separation: the MLE is at infinity and the Wald standard errors blow up. Firth (1993) modifies the score by `(1/2) ∂ log|I(beta)| / ∂beta`, which both removes the leading `O(1/n)` bias and produces finite estimates under separation. *fastglm* exposes the penalty as a flag,

```{r, eval = FALSE}
fastglm(x, y, family = binomial(), firth = TRUE)
```

For binomial logit the penalty collapses to a per-iteration adjustment of the working response by `h_i (0.5 - mu_i) / (mu_i (1 - mu_i))` where `h_i` is the leverage; no new family code is needed, and the cost per IRLS iteration is one extra streaming pass for the leverages.

The canonical demonstration uses the `sex2` dataset shipped with *logistf* (Heinze and Schemper, 2002):

```{r}
data(sex2, package = "logistf")
X <- model.matrix(~ age + oc + vic + vicl + vis + dia, data = sex2)
y <- sex2$case

fit_f <- fastglm(X, y, family = binomial(), firth = TRUE)
fit_l <- logistf(case ~ age + oc + vic + vicl + vis + dia, data = sex2)

cbind(fastglm = unname(coef(fit_f)), logistf = unname(coef(fit_l)))
max(abs(unname(coef(fit_f)) - unname(coef(fit_l))))
```

The two implementations land on the same penalized MLE to about 2e-7. The difference in runtime is substantial because *fastglm*'s Firth-fitting code leverages our existing C++ solver:

```{r}
mb_firth <- microbenchmark(
    fastglm = fastglm(X, y, family = binomial(), firth = TRUE),
    logistf = logistf(case ~ age + oc + vic + vicl + vis + dia, data = sex2),
    times = 10L
)
print(mb_firth)
```

`firth = TRUE` is currently restricted to `family = binomial(link = "logit")` on a dense matrix `x`; probit / cloglog use a different score correction and will be added separately.

# Hurdle models

Hurdle models, introduced by Mullahy (1986), factorize a count distribution into two independent pieces:

- a binary regression for `1(y > 0)` over the full sample (the *zero* part);
- a zero-truncated Poisson or NB regression for `y > 0` over the positive subset (the *count* part).

Because the two parts share no parameters, the joint likelihood factorizes and they can be fit independently. *fastglm*'s C++ driver fits both parts using the same IRLS solver as `fastglm()`, with new `FAM_POIS_TRUNC_*` / `FAM_NB_TRUNC_*` family codes that handle the truncation correction stably (`expm1`, `log1p` near `mu = 0`).

The formula uses the *Formula* package convention `y ~ x1 + x2 | z1 + z2`; the right-hand side after `|` specifies the zero-part design (it defaults to the count-part design if absent).

```{r}
data(bioChemists, package = "pscl")
fit_f <- fastglm_hurdle(art ~ ., data = bioChemists, dist = "poisson")
fit_p <- pscl::hurdle (art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))
```

`coef()` and `vcov()` accept a `model = c("full", "count", "zero")` argument so each part can be inspected separately:

```{r}
coef(fit_f, model = "count")
coef(fit_f, model = "zero")
```

For NB count parts, the dispersion is estimated by an inner Brent MLE that runs *between* outer IRLS iterations; the outer-loop tolerance is controlled by `outer.tol` / `outer.maxit`.

A timing comparison on a 4000-observation simulated example:

```{r}
set.seed(11)
n  <- 4000
x1 <- rnorm(n);  x2 <- rnorm(n)
lam    <- exp(0.7 + 0.4 * x1 - 0.3 * x2)
is_pos <- rbinom(n, 1, plogis(-0.4 + 0.5 * x1 + 0.2 * x2))
yt     <- integer(n)
for (i in seq_len(n)) {
    repeat { v <- rpois(1, lam[i]); if (v > 0) { yt[i] <- v; break } }
}
y  <- ifelse(is_pos == 1, yt, 0L)
df <- data.frame(y = y, x1 = x1, x2 = x2)

mb_hurdle <- microbenchmark(
    fastglm_hurdle = fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_hurdle    = pscl::hurdle (y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_hurdle)
```

# Zero-inflated models

Zero inflation, introduced by Lambert (1992), differs from a hurdle in that the two components share a latent variable: a `y = 0` outcome could come from the inflation component (with probability `pi_i`) *or* from the count component (with probability `1 - pi_i` and a count-side zero). The likelihood for `y = 0` is

$$
\Pr(Y_i = 0) = \pi_i + (1 - \pi_i) f(0; \mu_i)
$$

and for `y > 0` it is `(1 - pi_i) f(y; mu_i)`. The two-component mixture rules out a closed-form factorization, so *fastglm* uses an EM algorithm:

- **E-step**: posterior `tau_i = P(Z_i = 1 | y_i)` for each observation, computed in log-space with `logsumexp` to avoid catastrophic cancellation when `pi_i` is near 0 or 1.
- **M-step**: a binomial logit / probit / cloglog / log fit with response `tau` (continuous), weights 1; followed by a Poisson or NB fit on the *original* `y` with prior weights `1 - tau`. For NB, an inner Brent MLE re-estimates `theta` after each count-side step.

The final observed-information `vcov` comes from a numerical Jacobian of the analytical observed score at the EM fixed point, which is stable and cheap (block-diagonal per `(gamma, beta, theta)`). The complete EM driver — E-step, both M-steps, the inner theta MLE, and the score Jacobian — runs in C++.

The formula syntax matches `fastglm_hurdle()` exactly:

```{r}
fit_f <- fastglm_zi(art ~ ., data = bioChemists, dist = "poisson",
                    em.tol = 1e-10, em.maxit = 300L)
fit_p <- pscl::zeroinfl(art ~ ., data = bioChemists, dist = "poisson")

max(abs(unname(coef(fit_f, "count")) - unname(fit_p$coefficients$count)))
max(abs(unname(coef(fit_f, "zero"))  - unname(fit_p$coefficients$zero)))
abs(as.numeric(logLik(fit_f)) - as.numeric(logLik(fit_p)))
```

EM is iterative, so the agreement is slightly looser than for hurdle (where the two parts are closed-form independent fits), but coefficients and the joint log-likelihood still match to about 1e-5.

A timing comparison on a simulated zero-inflated Poisson:

```{r}
set.seed(21)
n  <- 3000
x1 <- rnorm(n);  x2 <- rnorm(n)
eta_c <- 0.7 + 0.4 * x1 - 0.3 * x2
eta_z <- -0.4 + 0.5 * x1 + 0.2 * x2
z     <- rbinom(n, 1, plogis(eta_z))
y     <- ifelse(z == 1, 0L, rpois(n, exp(eta_c)))
df    <- data.frame(y = y, x1 = x1, x2 = x2)

mb_zi <- microbenchmark(
    fastglm_zi = fastglm_zi(y ~ x1 + x2, data = df, dist = "poisson"),
    pscl_zi    = pscl::zeroinfl(y ~ x1 + x2, data = df, dist = "poisson"),
    times = 5L
)
print(mb_zi)
```

# References

- Firth, D. (1993). Bias reduction of maximum likelihood estimates. *Biometrika*, 80(1), 27–38. <doi:10.1093/biomet/80.1.27>

- Heinze, G. and Schemper, M. (2002). A solution to the problem of separation in logistic regression. *Statistics in Medicine*, 21(16), 2409–2419. <doi:10.1002/sim.1047>

- Cameron, A. C. and Trivedi, P. K. (1998). *Regression Analysis of Count Data*. Cambridge University Press.

- Mullahy, J. (1986). Specification and testing of some modified count data models. *Journal of Econometrics*, 33(3), 341–365. <doi:10.1016/0304-4076(86)90002-3>

- Lambert, D. (1992). Zero-inflated Poisson regression, with an application to defects in manufacturing. *Technometrics*, 34(1), 1–14. <doi:10.2307/1269547>

- Zeileis, A., Kleiber, C., and Jackman, S. (2008). Regression models for count data in R. *Journal of Statistical Software*, 27(8), 1–25. <doi:10.18637/jss.v027.i08>

- Venables, W. N. and Ripley, B. D. (2002). *Modern Applied Statistics with S* (4th ed.). Springer.

---
title: "Overview of the 'fastglm' Package"
author: "Jared Huling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overview of the 'fastglm' Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  echo = TRUE,
  eval = TRUE
)
```

The *fastglm* package is intended as a **fast** *and* **stable** alternative to `stats::glm()` for fitting generalized linear models. It is compatible with `R`'s `family` objects (see `?family`) and produces fitted objects whose downstream methods --- `summary()`, `vcov()`, `predict()`, `coef()`, `residuals()`, `logLik()` --- behave exactly like those for a `glm`. This vignette walks through the main functionality of the package at a higher level than the other vignettes; it is intended as a single entry point for a new user.

```{r}
library(fastglm)
```

# Fitting a GLM

The main package function is `fastglm()`. It takes a numeric design matrix `x`, a response `y`, and an `R` `family` object:

```{r}
data(esoph)
x <- model.matrix(~ agegp + unclass(tobgp) + unclass(alcgp),
                  data = esoph)
y <- cbind(esoph$ncases, esoph$ncontrols)

fit <- fastglm(x, y, family = binomial(link = "cloglog"))
summary(fit)
```

`fastglm()` does not accept a formula directly, it operates on a pre-built design matrix. To use a formula and a data frame, pass `fastglm_fit` as the fitting function to base `glm()`:

```{r}
fit2 <- glm(cbind(ncases, ncontrols) ~ agegp + unclass(tobgp) + unclass(alcgp),
            data = esoph, family = binomial(link = "cloglog"),
            method = fastglm_fit)
```

# Decomposition methods

The IRLS algorithm reduces every iteration to a weighted least-squares problem. *fastglm* supports six different matrix decompositions for solving that WLS step, all from *RcppEigen* (Bates and Eddelbuettel, 2013); the choice trades off speed against numerical stability and rank-revealing behaviour:

| `method` | decomposition |
|---|---|
| `0` | column-pivoted Householder QR (default; rank-revealing) |
| `1` | unpivoted Householder QR |
| `2` | LLT Cholesky |
| `3` | LDLT Cholesky |
| `4` | full-pivoted Householder QR |
| `5` | bidiagonal divide-and-conquer SVD |

The default (`method = 0`) is the safe choice: it is rank-revealing, so it handles aliased / collinear columns gracefully. The Cholesky methods (`2` and `3`) are roughly 3–4× faster but assume full column rank.

```{r}
set.seed(123)
n <- 5000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rbinom(n, 1, plogis(x %*% rnorm(p) * 0.05))

system.time(f0 <- fastglm(x, y, family = binomial()))                 # default
system.time(f2 <- fastglm(x, y, family = binomial(), method = 2))     # LLT
system.time(f3 <- fastglm(x, y, family = binomial(), method = 3))     # LDLT
```

# Stability

*fastglm* does not compromise computational stability for speed. It uses a step-halving safeguard following Marschner (2011) and starts from better-initialised values than `glm()` or `glm2::glm2()`, so it tends to converge in cases where the standard IRLS algorithm fails. See `vignette("fastglm", package = "fastglm")` for a Gamma-with-`sqrt`-link example where `glm()` does not converge but `fastglm()` does.

# Inference: `vcov()`, robust SE, predictions

The fitted object stores the unscaled covariance directly (no refitting), and the standard `vcov()` / `summary()` machinery works as expected:

```{r}
fit <- fastglm(x, y, family = binomial(), method = 2)

V    <- vcov(fit)
dim(V)

## standard errors agree with the diagonal of vcov()
all.equal(sqrt(diag(V)), unname(coef(summary(fit))[, "Std. Error"]))
```

Heteroskedasticity-consistent and cluster-robust covariance matrices are
available via `sandwich::vcovHC()` and `sandwich::vcovCL()` — `fastglm`
registers methods on those generics, so loading `sandwich` is all that is
required:

```{r}
library(sandwich)

V_hc <- vcovHC(fit, type = "HC0")
V_hc[1:3, 1:3]

cluster <- rep(seq_len(20), length.out = n)
V_cl    <- vcovCL(fit, cluster = cluster, type = "HC1")
V_cl[1:3, 1:3]
```

Results are numerically identical to `sandwich::vcovHC()` and
`sandwich::vcovCL()` applied to a `glm` fit (Zeileis, Köll, and Graham, 2020)
to floating-point precision. They work for sparse and `big.matrix` fits as
well, since the full design matrix is preserved on the fitted object (as a
reference, not copied).

`predict()` supports `se.fit = TRUE`:

```{r}
xnew <- x[1:5, , drop = FALSE]
predict(fit, newdata = xnew, type = "response", se.fit = TRUE)
```

# Sparse, big.matrix, and streaming designs

For designs that are sparse, that live on disc, or that have to be built from a parquet / *arrow* / *DuckDB* source, *fastglm* provides three large-data paths that share a common streaming kernel and produce identical results:

- `Matrix::dgCMatrix`: pass directly to `fastglm()`. Useful for one-hot encoded categoricals and high-dimensional sparse designs.
- `bigmemory::big.matrix`: pass directly to `fastglm()`. The matrix is read in row-blocks and never fully materialised in memory.
- `fastglm_streaming(chunk_callback, n_chunks, family)`: a user-supplied closure yields one row-block per call. The right path for fitting on a parquet dataset, *DuckDB* query, or any external columnar store.

See `vignette("large-data-fastglm", package = "fastglm")` for a detailed walk through all three paths. A short example:

```{r}
n <- 4000
X <- cbind(1, matrix(rnorm(n * 3), n, 3))
y <- rbinom(n, 1, plogis(X %*% c(0.2, 0.4, -0.2, 0.3)))

chunk_size <- 1000
chunks <- function(k) {
    idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
    list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)

max(abs(coef(fit_stream) - coef(fit_full)))
```

# Native families

For the most commonly-used `family`/`link` combinations, *fastglm* dispatches `variance()`, `mu.eta()`, `linkinv()`, and `dev.resids()` to inline C++ implementations rather than calling back into `R` once per IRLS iteration. The covered combinations are:

- gaussian (identity, log, inverse)
- binomial (logit, probit, cloglog, log)
- poisson (log, identity, sqrt)
- Gamma (log, inverse, identity)
- inverse.gaussian (1/mu^2, log, identity, inverse)

Detection is automatic, if the `family` object matches one of the above, the native path is used; otherwise *fastglm* falls back to the `R`-callback path used in earlier versions. The native path is meaningfully faster on large `n` because it eliminates the per-iteration round-trip to `R` for each of the four family functions:

```{r, eval = TRUE}
library(microbenchmark)

set.seed(7)
n <- 50000; p <- 30
x <- matrix(rnorm(n * p), n, p)
y <- rpois(n, exp(x %*% rnorm(p) * 0.05 + 1))

## force the R-callback path by giving the family object an unrecognised name
disguise <- function(fam) { fam$family <- paste0(fam$family, "*"); fam }

mb <- microbenchmark(
    native   = fastglm(x, y, family = poisson(),            method = 2),
    callback = fastglm(x, y, family = disguise(poisson()),  method = 2),
    times = 5L
)
print(mb)
```

The *fastglm* package switches transparently between the two paths; user code does not change.

# Negative binomial, hurdle, zero-inflation, and Firth logistic

A handful of regressions that come up often in practice are not standard GLMs but build on the same IRLS machinery. *fastglm* provides dedicated entry points for these:

- **`fastglm_nb()`** — negative-binomial regression with the dispersion `theta` estimated jointly with `beta`. Plays the same role as `MASS::glm.nb()` but runs the joint `(beta, theta)` MLE entirely in C++ (IRLS for `beta`, Brent for `theta`).
- **`firth = TRUE`** — a flag on `fastglm()` / `fastglm_fit()` that activates Firth's bias-reducing penalty on the score. Useful for binary logistic regression under separation or in small samples; analogous to `logistf::logistf()`.
- **`fastglm_hurdle()`** — a two-part count model: a binary regression for whether `y > 0`, plus a zero-truncated Poisson or NB regression on the positive subset. The two parts factorize, and both are fit by the same C++ IRLS solver. Same model as `pscl::hurdle()`.
- **`fastglm_zi()`** — a zero-inflated Poisson or NB regression: a binary inflation component overlaid on the original count distribution, fit by an EM algorithm in C++ with closed-form posterior responsibilities and an analytical observed-information `vcov`. Same model as `pscl::zeroinfl()`.

All four take the standard *fastglm* tuning options (`method`, `tol`, `maxit`) and return objects with the usual `coef()`, `vcov()`, `predict()`, `logLik()` methods. `fastglm_hurdle()` and `fastglm_zi()` use the `Formula` package's two-RHS syntax `y ~ x1 + x2 | z1 + z2` to specify different designs for the count and zero parts. See `vignette("count-firth-fastglm", package = "fastglm")` for examples and timing comparisons against the canonical reference packages.

# Three R entry points

There are three `R`-side functions to fit a GLM with *fastglm*. They all call into the same C++ solver, but differ in how much of base R's machinery they wrap around it:

- **`fastglm()`** is the top-level S3 function. It returns an object of class `"fastglm"` with deviance, AIC, null deviance, residuals, etc. populated --- the right entry point for interactive use.
- **`fastglm_fit()`** is designed to be passed as `method = fastglm_fit` to base `glm()`. The returned object inherits from `"fastglmFit"` so that `glm()`'s formula/data handling, `update()`, `predict()`, etc. all work on top.
- **`fastglmPure()`** is the lowest-level wrapper: no dispersion, no AIC, no null-deviance computation. Use this when calling *fastglm* from another package and you only need the coefficients.

# References

- Marschner, I. C. (2011). glm2: Fitting generalized linear models with convergence problems. *The R Journal*, 3(2), 12–15. <doi:10.32614/RJ-2011-012>

- Bates, D. and Eddelbuettel, D. (2013). Fast and elegant numerical linear algebra using the RcppEigen package. *Journal of Statistical Software*, 52(5), 1–24. <doi:10.18637/jss.v052.i05>

- Zeileis, A., Köll, S., and Graham, N. (2020). Various versatile variances: An object-oriented implementation of clustered covariances in R. *Journal of Statistical Software*, 95(1), 1–36. <doi:10.18637/jss.v095.i01>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-methods.R
\name{vcovHC.fastglm}
\alias{vcovHC.fastglm}
\alias{vcovHC.fastglmFit}
\title{Heteroskedasticity-consistent (HC) variance estimators for `fastglm` objects}
\usage{
vcovHC.fastglm(object, type = c("HC3", "HC2", "HC1", "HC0"), ...)

vcovHC.fastglmFit(object, type = c("HC3", "HC2", "HC1", "HC0"), ...)
}
\arguments{
\item{object}{a fitted object of class `"fastglm"` or `"fastglmFit"`.}

\item{type}{one of `"HC0"`, `"HC1"`, `"HC2"`, `"HC3"`. Default `"HC3"` matches
`sandwich::vcovHC.glm`.}

\item{...}{not used.}
}
\value{
A `p x p` heteroskedasticity-consistent variance-covariance matrix.
}
\description{
Methods for `sandwich::vcovHC()` on objects of class `"fastglm"` and
`"fastglmFit"`. Load `sandwich` (`library(sandwich)`) before calling
`vcovHC(fit)`; otherwise no `vcovHC` generic is in scope.
}
\details{
Computes the Eicker-Huber-White sandwich estimator
`bread %*% meat %*% bread`, where `bread = (X' W X)^{-1}` (already stored as
`cov.unscaled`) and `meat = X' diag(omega_i) X`. With `s_i = w_i^2 * r_i`
the score contribution from observation `i`, the omegas are:
\describe{
  \item{`HC0`}{`omega_i = s_i^2`}
  \item{`HC1`}{`HC0` rescaled by `n / (n - p)`}
  \item{`HC2`}{`omega_i = s_i^2 / (1 - h_i)`}
  \item{`HC3`}{`omega_i = s_i^2 / (1 - h_i)^2`}
}
where `r_i` is the working residual `(y - mu) / mu.eta(eta)`,
`w_i^2 = prior.weight * mu.eta(eta)^2 / variance(mu)` is the IRLS working
weight, and `h_i = w_i^2 * x_i' (X' W X)^(-1) x_i` is the IRLS leverage.
Equivalent to `sandwich::vcovHC.glm`.

Requires the model matrix `x` stored on the fitted object (set automatically
by `fastglm()`, `fastglmPure()`, and `fastglm_fit()` since version 0.0.6).
}
\examples{
if (requireNamespace("sandwich", quietly = TRUE)) {
  x <- cbind(1, matrix(rnorm(500 * 4), ncol = 4))
  y <- rbinom(500, 1, plogis(x \%*\% c(0.2, 0.3, -0.4, 0.1, 0.2)))
  fit <- fastglm(x, y, family = binomial())
  sandwich::vcovHC(fit)
  sandwich::vcovHC(fit, type = "HC0")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_glm.R
\name{fastglm}
\alias{fastglm}
\alias{fastglm.default}
\title{Fast generalized linear model fitting}
\usage{
fastglm(x, ...)

\method{fastglm}{default}(
  x,
  y,
  family = gaussian(),
  weights = NULL,
  offset = NULL,
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  method = 0L,
  tol = 1e-08,
  maxit = 100L,
  firth = FALSE,
  ...
)
}
\arguments{
\item{x}{input model matrix. Must be a matrix object}

\item{...}{not used}

\item{y}{numeric response vector of length nobs.}

\item{family}{a description of the error distribution and link function to be used in the model. 
For \code{fastglm} this can be a character string naming a family function, a family function or the 
result of a call to a family function. For \code{fastglmPure} only the third option is supported. 
(See \code{\link[stats]{family}} for details of family functions.)}

\item{weights}{an optional vector of 'prior weights' to be used in the fitting process. Should be a numeric vector.}

\item{offset}{this can be used to specify an a priori known component to be included in the linear predictor during fitting. 
This should be a numeric vector of length equal to the number of cases}

\item{start}{starting values for the parameters in the linear predictor.}

\item{etastart}{starting values for the linear predictor.}

\item{mustart}{values for the vector of means.}

\item{method}{an integer scalar with value 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition,
2 for the LLT Cholesky, or 3 for the LDLT Cholesky}

\item{tol}{threshold tolerance for convergence. Should be a positive real number}

\item{maxit}{maximum number of IRLS iterations. Should be an integer}

\item{firth}{logical; if `TRUE` apply Firth's (1993) bias-reducing penalty
to the score function. Currently supported only for
`family = binomial(link = "logit")` on dense `x`. See `logistf::logistf()`
for the canonical reference implementation.}
}
\value{
A list with the elements
\item{coefficients}{a vector of coefficients}
\item{se}{a vector of the standard errors of the coefficient estimates}
\item{rank}{a scalar denoting the computed rank of the model matrix}
\item{df.residual}{a scalar denoting the degrees of freedom in the model}
\item{residuals}{the vector of residuals}
\item{s}{a numeric scalar - the root mean square for residuals}
\item{fitted.values}{the vector of fitted values}
}
\description{
Fast generalized linear model fitting

bigLm default
}
\examples{

x <- matrix(rnorm(10000 * 100), ncol = 100)
y <- 1 * (0.25 * x[,1] - 0.25 * x[,3] > rnorm(10000))

system.time(gl1 <- glm.fit(x, y, family = binomial()))

system.time(gf1 <- fastglm(x, y, family = binomial()))

system.time(gf2 <- fastglm(x, y, family = binomial(), method = 1))

system.time(gf3 <- fastglm(x, y, family = binomial(), method = 2))

system.time(gf4 <- fastglm(x, y, family = binomial(), method = 3))

max(abs(coef(gl1) - gf1$coef))
max(abs(coef(gl1) - gf2$coef))
max(abs(coef(gl1) - gf3$coef))
max(abs(coef(gl1) - gf4$coef))


\dontrun{
nrows <- 50000
ncols <- 50
bkFile <- "bigmat2.bk"
descFile <- "bigmatk2.desc"
bigmat <- filebacked.big.matrix(nrow=nrows, ncol=ncols, type="double",
                                backingfile=bkFile, backingpath=".",
                                descriptorfile=descFile,
                                dimnames=c(NULL,NULL))
for (i in 1:ncols) bigmat[,i] = rnorm(nrows)*i
y <- 1*(rnorm(nrows) + bigmat[,1] > 0)

system.time(gfb1 <- fastglm(bigmat, y, family = binomial(), method = 3))
}

}
\seealso{
[fastglm_fit()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-methods.R
\name{summary.fastglm}
\alias{summary.fastglm}
\title{`summary()` method for `fastglm` fitted objects}
\usage{
\method{summary}{fastglm}(
  object,
  dispersion = NULL,
  correlation = FALSE,
  symbolic.cor = FALSE,
  ...
)
}
\arguments{
\item{object}{`fastglm` fitted object}

\item{dispersion}{the dispersion parameter for the family used. Either a single numerical value or `NULL` (the default), when it is inferred from `object`.}

\item{correlation}{logical; if `TRUE`, the correlation matrix of the estimated parameters is returned.}

\item{symbolic.cor}{logical; if `TRUE`, print the correlations in a symbolic form (see `symnum`) rather than as numbers.}

\item{...}{not used}
}
\value{
A `summary.fastglm` object
}
\description{
`summary()` method for `fastglm` fitted objects
}
\examples{
x <- matrix(rnorm(10000 * 10), ncol = 10)
y <- 1 * (0.25 * x[,1] - 0.25 * x[,3] > rnorm(10000))

fit <- fastglm(x, y, family = binomial())

summary(fit)
}
\seealso{
[summary.glm()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-nb.R
\name{fastglm_nb}
\alias{fastglm_nb}
\title{Fit a negative-binomial GLM with simultaneous (beta, theta) MLE}
\usage{
fastglm_nb(
  x,
  y,
  weights = NULL,
  offset = NULL,
  start = NULL,
  init.theta = NULL,
  link = c("log", "sqrt", "identity"),
  method = 2L,
  tol = 1e-08,
  maxit = 100L,
  outer.maxit = 25L,
  outer.tol = 1e-07,
  theta.tol = 1e-08,
  theta.maxit = 100L
)
}
\arguments{
\item{x}{design matrix (numeric matrix, dgCMatrix not yet supported here).}

\item{y}{non-negative integer response vector.}

\item{weights}{optional prior weights vector of length `length(y)`.}

\item{offset}{optional offset vector of length `length(y)`.}

\item{start}{optional starting values for `beta`.}

\item{init.theta}{optional starting value for `theta`. If `NULL`, uses
the method-of-moments estimator from a Poisson pilot fit.}

\item{link}{character, one of `"log"` (default), `"sqrt"`, `"identity"`.}

\item{method}{integer; `0..5`, see [fastglm()].}

\item{tol}{convergence tolerance for the IRLS inner loop.}

\item{maxit}{maximum number of inner-loop IRLS iterations.}

\item{outer.maxit}{maximum number of `(beta, theta)` outer iterations.}

\item{outer.tol}{convergence tolerance for the outer loop on the
sup-norm of the beta update plus the relative change in `theta`.}

\item{theta.tol}{Brent tolerance for the inner theta MLE.}

\item{theta.maxit}{max iterations for the inner theta MLE.}
}
\value{
A list of class `c("fastglm_nb", "fastglm")` with the usual
  fastglm components plus `theta`, `SE.theta`, `iter.theta`, and
  `twologlik` (twice the maximised NB log-likelihood).
}
\description{
`fastglm_nb()` fits a negative-binomial regression model, jointly
maximising the likelihood over the regression coefficients `beta` and the
NB2 dispersion `theta`. It is the fastglm analogue of [MASS::glm.nb()],
built on top of the native NB family kernel introduced in 0.0.6 so that
all numerical loops -- IRLS, the inner theta MLE Brent root-find, and
the outer (beta, theta) alternation -- run entirely in C++.
}
\examples{
set.seed(1)
n <- 500
x <- cbind(1, matrix(rnorm(n * 2), n, 2))
eta <- x \%*\% c(0.3, 0.5, -0.2)
mu  <- exp(eta)
if (requireNamespace("MASS", quietly = TRUE)) {
  y   <- MASS::rnegbin(n, mu = mu, theta = 2)
  fit <- fastglm_nb(x, y)
  c(theta = fit$theta, fit$coefficients)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bigmemory_methods.R
\docType{methods}
\name{\%*\%,big.matrix,vector-method}
\alias{\%*\%,big.matrix,vector-method}
\alias{\%*\%,vector,big.matrix-method}
\title{big.matrix prod}
\usage{
\S4method{\%*\%}{big.matrix,vector}(x, y)

\S4method{\%*\%}{vector,big.matrix}(x, y)
}
\arguments{
\item{x}{big.matrix}

\item{y}{numeric vector}
}
\description{
big.matrix prod

big.matrix prod
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_glm.R
\name{fastglmPure}
\alias{fastglmPure}
\title{Fast generalized linear model fitting}
\usage{
fastglmPure(
  x,
  y,
  family = gaussian(),
  weights = rep(1, NROW(y)),
  offset = rep(0, NROW(y)),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  method = 0L,
  tol = 1e-07,
  maxit = 100L,
  firth = FALSE
)
}
\arguments{
\item{x}{input model matrix. Must be a matrix object}

\item{y}{numeric response vector of length nobs.}

\item{family}{a description of the error distribution and link function to be used in the model.
For \code{fastglmPure} this can only be the result of a call to a family function.
(See \code{\link[stats]{family}} for details of family functions.)}

\item{weights}{an optional vector of 'prior weights' to be used in the fitting process. Should be a numeric vector.}

\item{offset}{this can be used to specify an a priori known component to be included in the linear predictor during fitting. 
This should be a numeric vector of length equal to the number of cases}

\item{start}{starting values for the parameters in the linear predictor.}

\item{etastart}{starting values for the linear predictor.}

\item{mustart}{values for the vector of means.}

\item{method}{an integer scalar with value 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition,
2 for the LLT Cholesky, 3 for the LDLT Cholesky, 4 for the full pivoted QR decomposition, 5 for the Bidiagonal Divide and 
Conquer SVD}

\item{tol}{threshold tolerance for convergence. Should be a positive real number}

\item{maxit}{maximum number of IRLS iterations. Should be an integer}

\item{firth}{logical; if `TRUE` apply Firth's (1993) bias-reducing penalty
to the score function. Currently supported only for
`family = binomial(link = "logit")` on dense `x`. The penalty modifies
the working response by `h_i (0.5 - mu_i) / (mu_i (1 - mu_i))` where
`h_i` is the leverage; convergence is checked on the sup-norm of the
coefficient update. See `logistf::logistf()` for the canonical
reference implementation.}
}
\value{
A list with the elements
\item{coefficients}{a vector of coefficients}
\item{se}{a vector of the standard errors of the coefficient estimates}
\item{rank}{a scalar denoting the computed rank of the model matrix}
\item{df.residual}{a scalar denoting the degrees of freedom in the model}
\item{residuals}{the vector of residuals}
\item{s}{a numeric scalar - the root mean square for residuals}
\item{fitted.values}{the vector of fitted values}
}
\description{
Fast generalized linear model fitting
}
\examples{

set.seed(1)
x <- matrix(rnorm(1000 * 25), ncol = 25)
eta <- 0.1 + 0.25 * x[,1] - 0.25 * x[,3] + 0.75 * x[,5] -0.35 * x[,6] #0.25 * x[,1] - 0.25 * x[,3]
y <- 1 * (eta > rnorm(1000))

yp <- rpois(1000, eta ^ 2)
yg <- rgamma(1000, exp(eta) * 1.75, 1.75)

# binomial
system.time(gl1 <- glm.fit(x, y, family = binomial()))

system.time(gf1 <- fastglmPure(x, y, family = binomial(), tol = 1e-8))

system.time(gf2 <- fastglmPure(x, y, family = binomial(), method = 1, tol = 1e-8))

system.time(gf3 <- fastglmPure(x, y, family = binomial(), method = 2, tol = 1e-8))

system.time(gf4 <- fastglmPure(x, y, family = binomial(), method = 3, tol = 1e-8))

max(abs(coef(gl1) - gf1$coef))
max(abs(coef(gl1) - gf2$coef))
max(abs(coef(gl1) - gf3$coef))
max(abs(coef(gl1) - gf4$coef))

# poisson
system.time(gl1 <- glm.fit(x, yp, family = poisson(link = "log")))

system.time(gf1 <- fastglmPure(x, yp, family = poisson(link = "log"), tol = 1e-8))

system.time(gf2 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 1, tol = 1e-8))

system.time(gf3 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 2, tol = 1e-8))

system.time(gf4 <- fastglmPure(x, yp, family = poisson(link = "log"), method = 3, tol = 1e-8))

max(abs(coef(gl1) - gf1$coef))
max(abs(coef(gl1) - gf2$coef))
max(abs(coef(gl1) - gf3$coef))
max(abs(coef(gl1) - gf4$coef))

# gamma
system.time(gl1 <- glm.fit(x, yg, family = Gamma(link = "log")))

system.time(gf1 <- fastglmPure(x, yg, family = Gamma(link = "log"), tol = 1e-8))

system.time(gf2 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 1, tol = 1e-8))

system.time(gf3 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 2, tol = 1e-8))

system.time(gf4 <- fastglmPure(x, yg, family = Gamma(link = "log"), method = 3, tol = 1e-8))

max(abs(coef(gl1) - gf1$coef))
max(abs(coef(gl1) - gf2$coef))
max(abs(coef(gl1) - gf3$coef))
max(abs(coef(gl1) - gf4$coef))

}
\seealso{
[fastglm_fit()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-methods.R
\name{fastglm-sandwich}
\alias{fastglm-sandwich}
\alias{estfun.fastglm}
\alias{estfun.fastglmFit}
\alias{bread.fastglm}
\alias{bread.fastglmFit}
\title{Empirical estimating-function and bread methods for `fastglm` objects}
\usage{
estfun.fastglm(x, ...)

estfun.fastglmFit(x, ...)

bread.fastglm(x, ...)

bread.fastglmFit(x, ...)
}
\arguments{
\item{x}{a fitted object of class `"fastglm"` or `"fastglmFit"`.}

\item{...}{not used.}
}
\value{
`estfun()` returns an `n x p` matrix; `bread()` returns a
  `p x p` matrix.
}
\description{
Methods for `sandwich::estfun()` and `sandwich::bread()`, which let
`sandwich::vcovCL()`, `sandwich::vcovBS()`, and the rest of the *sandwich*
machinery work directly on `fastglm` and `fastglmFit` objects.  Load
`sandwich` (`library(sandwich)`) before calling them.
}
\details{
`estfun(x)` returns the per-observation empirical estimating-function
contributions, an `n x p` matrix whose `i`-th row is
`(y_i - mu_i) * mu.eta_i / variance(mu_i) * x_i / dispersion`. `bread(x)`
returns `(X' W X)^(-1) * n * dispersion` (the *sandwich* convention).
For poisson, binomial, and negative-binomial families the dispersion is
fixed at 1, exactly matching `sandwich::estfun.glm()` /
`sandwich::bread.glm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm_fit.R
\name{fastglm_fit}
\alias{fastglm_fit}
\alias{fastglm_control}
\alias{vcov.fastglmFit}
\alias{summary.fastglmFit}
\title{Fast generalized linear model fitting}
\usage{
fastglm_fit(
  x,
  y,
  weights = rep(1, NROW(y)),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  offset = rep(0, NROW(y)),
  family = gaussian(),
  control = list(),
  intercept = TRUE,
  singular.ok = TRUE,
  firth = FALSE
)

fastglm_control(fastmethod = 0L, tol = 1e-07, maxit = 100L, ...)

\method{vcov}{fastglmFit}(object, ...)

\method{summary}{fastglmFit}(object, ...)
}
\arguments{
\item{x}{a design matrix of dimension `n * p`. Can also be a `big.matrix` object from \pkg{bigmemory}.}

\item{y}{a vector of observations of length `n`.}

\item{weights}{an optional vector of 'prior weights' to be used in the fitting process. Should be `NULL` or a numeric vector.}

\item{start}{optional starting values for the parameters in the linear predictor.}

\item{etastart}{optional starting values for the linear predictor.}

\item{mustart}{optional starting values for the vector of means.}

\item{offset}{this can be used to specify an *a priori* known component to be included in the linear predictor during fitting. This should be `NULL` or a numeric vector of length equal to the number of cases.}

\item{family}{a description of the error distribution and link function to be used in the model. This must be a family function or the result of a call to a family function. (See [`family`] for details of family functions.)}

\item{control}{a list of parameters for controlling the fitting process. This is passed to `fastglm_control()`.}

\item{singular.ok, intercept}{See [glm.fit()].}

\item{firth}{`logical`; if `TRUE` apply Firth's (1993) bias-reducing
penalty to the score function. Currently supported only for
`family = binomial(link = "logit")` on dense `x`. See `logistf::logistf()`
for the canonical reference implementation.}

\item{fastmethod}{`integer`; the method used for fitting. Allowable values include 0 for the column-pivoted QR decomposition, 1 for the unpivoted QR decomposition, 2 for the LLT Cholesky, 3 for the LDLT Cholesky, 4 for the full pivoted QR decomposition, and 5 for the Bidiagonal Divide and Conquer SVD. Default is 0. Can also be supplied as `method` when not supplied directly as an argument from `glm()` (see Examples).}

\item{tol}{`numeric`; threshold tolerance for convergence.}

\item{maxit}{`integer`; the maximum number of IRLS iterations.}

\item{\dots}{for `vcov()` and `summary()`, other arguments passed downstream.}

\item{object}{a `fastglmFit` object; the output of a call to `glm()` with `method = fastglm_fit`.}
}
\description{
`fastglm_fit()` is a fitting method for [glm()]. It works like `glm.fit()`, i.e., by being supplied to the `method` argument of `glm()`.
}
\details{
The purpose of the functions documented on this page is to facilitate integration with existing [glm()] utilities in base R. `fastglm_fit()` is just a wrapper for [fastglmPure()] with some additional quality-of-life features. The `vcov()` and `summary()` methods use the unscaled coefficient covariance matrix returned directly from the C++ solver, so no refit is required.
}
\examples{
set.seed(1234)
n <- 1e4
x <- matrix(rnorm(n * 25), ncol = 25)
eta <- 0.1 + 0.25 * x[,1] - 0.25 * x[,3] + 0.75 * x[,5] -0.35 * x[,6]
dat <- as.data.frame(x)

# binomial
dat$y <- rbinom(n, 1, pnorm(eta))

system.time({
    gl <- glm(y ~ ., data = dat,
              family = binomial)
})

system.time({
    gf0 <- glm(y ~ ., data = dat,
               family = binomial,
               method = fastglm_fit)
})

system.time({
    gf1 <- glm(y ~ ., data = dat,
               family = binomial,
               method = fastglm_fit,
               fastmethod = 1)
})

# poisson
dat$y <- rpois(n, eta^2)

system.time({
    gl <- glm(y ~ ., data = dat,
              family = poisson)
})

system.time({
    gf0 <- glm(y ~ ., data = dat,
               family = poisson,
               method = fastglm_fit)
})

system.time({
    gf1 <- glm(y ~ ., data = dat,
               family = poisson,
               method = fastglm_fit,
               fastmethod = 1)
})

# gamma
dat$y <- rgamma(n, exp(eta) * 1.75, 1.75)

system.time({
    gl <- glm(y ~ ., data = dat,
              family = Gamma(link = "log"))
})

system.time({
    gf0 <- glm(y ~ ., data = dat,
               family = Gamma(link = "log"),
               method = fastglm_fit)
})

system.time({
    gf1 <- glm(y ~ ., data = dat,
               family = Gamma(link = "log"),
               method = fastglm_fit,
               fastmethod = 1)
})

# Different (equivalent) ways of supplying
# control arguments:
gf1 <- glm(y ~ ., data = dat,
           family = Gamma(link = "log"),
           method = fastglm_fit,
           fastmethod = 1)

gf1 <- glm(y ~ ., data = dat,
           family = Gamma(link = "log"),
           method = fastglm_fit,
           control = list(fastmethod = 1))

gf1 <- glm(y ~ ., data = dat,
           family = Gamma(link = "log"),
           method = fastglm_fit,
           control = list(method = 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/streaming.R
\name{fastglm_streaming}
\alias{fastglm_streaming}
\title{Fit a GLM by streaming row-blocks of the design matrix}
\usage{
fastglm_streaming(
  chunk_callback,
  n_chunks,
  family = gaussian(),
  start = NULL,
  method = 2L,
  tol = 1e-07,
  maxit = 100L
)
}
\arguments{
\item{chunk_callback}{a function. Called as `chunk_callback(k)` for
`k = 1, ..., n_chunks`. Must return a list with elements
\describe{
  \item{`X`}{an `n_k x p` numeric matrix (chunk of the design matrix).}
  \item{`y`}{numeric vector of length `n_k` (response).}
  \item{`weights`}{optional; numeric vector of length `n_k` of prior weights.}
  \item{`offset`}{optional; numeric vector of length `n_k` of offsets.}
}
Every chunk must have the same number of columns, in the same order. The
closure is called multiple times per IRLS iteration, so it should be
reasonably cheap (e.g. an Arrow scanner that reads from columnar files).}

\item{n_chunks}{integer; the number of chunks to iterate over.}

\item{family}{a `family` object describing the error distribution and link.
See [stats::family()].}

\item{start}{optional length-`p` numeric vector of starting coefficients.}

\item{method}{integer; `2` for LLT Cholesky (default) or `3` for LDLT
Cholesky. QR / SVD methods are not supported in streaming mode.}

\item{tol}{convergence tolerance on the relative change in deviance.}

\item{maxit}{maximum number of IRLS iterations.}
}
\value{
A list with class `"fastglm"` containing the same elements as
  [fastglm()], including `coefficients`, `cov.unscaled`, `deviance`,
  `iter`, `converged`, etc. The design matrix is *not* attached, so
  `sandwich::vcovHC()` / `sandwich::vcovCL()` will require re-streaming.
}
\description{
`fastglm_streaming()` runs an IRLS GLM fit where the design matrix is
produced one chunk at a time by a user-supplied closure. It never holds
more than a single chunk in memory at once, plus a `p x p` accumulator.
This is the front-end for fitting on data sources too large to load
completely into RAM (Arrow datasets, Parquet files, DuckDB query results,
on-disk CSV streams, etc.).
}
\details{
The IRLS loop and step-halving (Marschner 2011) run entirely in C++; the
R closure is called only to *deliver* one chunk at a time. For tier-1
families (gaussian / binomial / poisson / Gamma / inverse.gaussian on
their common links) the family functions are evaluated inline in C++, so
the only R round-trip per iteration is the chunk fetch.

Standard errors and `cov.unscaled` come from the final `(X' W X)`
Cholesky factor, exactly as in the in-memory path.
}
\section{Arrow / Parquet recipe}{


Wrap an Arrow scanner in a closure -- pull each `RecordBatch` as a chunk,
build the model matrix, and return it together with the response:
`chunks(k)` opens the dataset, scans the `k`-th batch, calls
`as.data.frame()`, then returns
`list(X = model.matrix(~ x1 + x2, data = tbl), y = tbl$y)`. Pass that
closure plus `n_chunks = length(batches)` to `fastglm_streaming()`.

The same recipe works for DuckDB (one `dbReadTable` per chunk), CSV
streamers, and custom binary formats.
}

\examples{
# Simulate a "data source" that yields the design matrix in 4 row-blocks.
set.seed(1)
n <- 1000; p <- 5
X <- cbind(1, matrix(rnorm(n * (p - 1)), n, p - 1))
y <- rbinom(n, 1, plogis(X \%*\% c(0.2, 0.5, -0.3, 0.4, -0.2)))
chunk_size <- 250
chunks <- function(k) {
  idx <- ((k - 1) * chunk_size + 1):(k * chunk_size)
  list(X = X[idx, , drop = FALSE], y = y[idx])
}

fit_stream <- fastglm_streaming(chunks, n_chunks = 4, family = binomial())
fit_full   <- fastglm(X, y, family = binomial(), method = 2)
max(abs(coef(fit_stream) - coef(fit_full)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-hurdle.R
\name{fastglm_hurdle}
\alias{fastglm_hurdle}
\title{Hurdle (two-part) Poisson / Negative-Binomial regression}
\usage{
fastglm_hurdle(
  formula,
  data,
  subset,
  na.action,
  weights,
  offset,
  dist = c("poisson", "negbin"),
  zero.dist = "binomial",
  link = c("logit", "probit", "cloglog", "log"),
  init.theta = NULL,
  tol = 1e-08,
  maxit = 100L,
  outer.tol = 1e-07,
  outer.maxit = 50L,
  theta.tol = 1e-08,
  theta.maxit = 100L,
  model = TRUE,
  y = TRUE,
  x = FALSE
)
}
\arguments{
\item{formula}{a [Formula::Formula] of the form `y ~ x1 + x2 | z1 + z2`
where the right-hand side after `|` specifies the zero-model design.
If `|` is absent, the same RHS is used for both parts.}

\item{data}{optional data frame / environment from which to draw model
matrix variables.}

\item{subset, na.action}{standard model-frame arguments.}

\item{weights}{optional non-negative prior weights.}

\item{offset}{optional vector or matrix.  If a 2-column matrix, columns
are taken as `(count_offset, zero_offset)`.}

\item{dist}{count component distribution: `"poisson"` or `"negbin"`.}

\item{zero.dist}{zero/non-zero distribution; only `"binomial"` is
currently supported.}

\item{link}{character, link for the zero/non-zero binomial.  One of
`"logit"` (default), `"probit"`, `"cloglog"`, `"log"`.}

\item{init.theta}{optional positive scalar starting value for the NB
dispersion.  If `NULL`, a method-of-moments initializer is used.}

\item{tol, maxit}{IRLS-loop convergence tolerance and iteration cap.}

\item{outer.tol, outer.maxit}{`(beta, theta)` outer-loop convergence
tolerance and iteration cap (only used when `dist = "negbin"`).}

\item{theta.tol, theta.maxit}{Brent-loop tolerance / iteration cap for
the inner theta MLE (only used when `dist = "negbin"`).}

\item{model}{logical; keep the model frame on the result.}

\item{y, x}{logical; keep the response / design matrices on the result.}
}
\value{
A list of class `c("fastglm_hurdle", "fastglm")` with elements
  `coefficients` (a list with `count` and `zero`), `se` (likewise),
  `vcov` (block-diagonal), `loglik`, `loglik_count`, `loglik_zero`,
  `theta`, `SE.theta`, `iter_*`, and `converged`.
}
\description{
Fits a hurdle model for count data with excess zeros: a binary
(zero / non-zero) regression on the full sample combined with a zero-
truncated count regression on the positive subset. The two parts have
independent likelihoods and are estimated jointly. With `dist = "negbin"`
the negative-binomial dispersion `theta` is estimated by an inner
Brent-1D MLE inside the C++ driver, alternating with the count-side
IRLS.
}
\details{
This is the fastglm analogue of [pscl::hurdle()] -- coefficients agree
to 1e-6 on standard datasets, with all numerical loops in C++.
}
\examples{
set.seed(1)
n <- 300
x1 <- rnorm(n); x2 <- rnorm(n)
eta_count <- 1.0 + 0.4*x1 - 0.2*x2
lam <- exp(eta_count)
y_count <- rpois(n, lam)
eta_zero  <- -0.5 + 0.6*x1
p_pos     <- plogis(eta_zero)
is_pos    <- rbinom(n, 1, p_pos)
# zero-truncated: resample any rare zero from positive Poisson
for (i in seq_len(n)) {
  while (y_count[i] == 0) y_count[i] <- rpois(1, lam[i])
}
y <- ifelse(is_pos == 1, y_count, 0)
df <- data.frame(y = y, x1 = x1, x2 = x2)
fit <- fastglm_hurdle(y ~ x1 + x2, data = df, dist = "poisson")
coef(fit, model = "count")
coef(fit, model = "zero")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-zi.R
\name{fastglm_zi}
\alias{fastglm_zi}
\title{Zero-inflated Poisson / Negative-Binomial regression}
\usage{
fastglm_zi(
  formula,
  data,
  subset,
  na.action,
  weights,
  offset,
  dist = c("poisson", "negbin"),
  link = c("logit", "probit", "cloglog", "log"),
  init.theta = NULL,
  em.tol = 1e-08,
  em.maxit = 100L,
  tol = 1e-09,
  maxit = 100L,
  theta.tol = 1e-08,
  theta.maxit = 100L,
  model = TRUE,
  y = TRUE,
  x = FALSE
)
}
\arguments{
\item{formula}{a [Formula::Formula] of the form `y ~ x1 + x2 | z1 + z2`
where the right-hand side after `|` specifies the inflation design.
If `|` is absent, the same RHS is used for both parts.}

\item{data}{optional data frame / environment.}

\item{subset, na.action}{standard model-frame arguments.}

\item{weights}{optional non-negative prior weights.}

\item{offset}{optional vector or matrix.  If a 2-column matrix, columns
are taken as `(count_offset, zero_offset)`.}

\item{dist}{count component distribution: `"poisson"` or `"negbin"`.}

\item{link}{character, link for the inflation binomial.  One of
`"logit"` (default), `"probit"`, `"cloglog"`, `"log"`.}

\item{init.theta}{optional positive scalar starting value for the NB
dispersion.  If `NULL`, a method-of-moments pilot is used.}

\item{em.tol, em.maxit}{EM convergence tolerance and iteration cap.}

\item{tol, maxit}{IRLS-loop convergence tolerance and iteration cap
(used inside each M-step).}

\item{theta.tol, theta.maxit}{Brent-loop tolerance / iteration cap for
the inner theta MLE (only used when `dist = "negbin"`).}

\item{model}{logical; keep the model frame on the result.}

\item{y, x}{logical; keep the response / design matrices on the result.}
}
\value{
A list of class `c("fastglm_zi", "fastglm")` with elements
  `coefficients` (a list with `count` and `zero`), `se` (likewise),
  `vcov` (full, including theta if NB), `loglik`, `theta`, `SE.theta`,
  `tau` (posterior P(Z=1|y)), `em_iter`, `converged`.
}
\description{
Fits a zero-inflated count model: a binary inflation component combined
with a Poisson or NB count component, where the count component is the
*original* (non-truncated) distribution.  Excess zeros above what the
count model implies are absorbed by the inflation component.
}
\details{
Estimation uses an EM algorithm with all numerical loops in C++.  The
E-step computes the posterior `tau_i = P(Z_i = 1 | y_i)` analytically;
M-steps fit the inflation logit/probit/cloglog/log and the Poisson/NB
count regression via the existing native fastglm IRLS (with
weights `w_i (1 - tau_i)` on the count side).  For NB, an inner Brent
MLE re-estimates `theta` after the count-side beta step.  Final vcov
comes from the numerical Jacobian of the analytical observed score at
the EM fixed point (block-structured for `(gamma, beta, theta)`).

This is the fastglm analogue of [pscl::zeroinfl()] -- coefficients
agree to 1e-5 on standard datasets (slightly looser than hurdle to
allow for the EM iteration), with all numerical loops in C++.
}
\examples{
set.seed(1)
n <- 400
x  <- rnorm(n)
eta_count <- 0.8 + 0.4*x
lam <- exp(eta_count)
eta_zero <- -0.6 + 0.5*x
p_inflate <- plogis(eta_zero)
z <- rbinom(n, 1, p_inflate)
y <- ifelse(z == 1, 0, rpois(n, lam))
df <- data.frame(y = y, x = x)
fit <- fastglm_zi(y ~ x, data = df, dist = "poisson")
coef(fit, model = "count")
coef(fit, model = "zero")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-methods.R
\name{vcov.fastglm}
\alias{vcov.fastglm}
\title{`vcov()` method for `fastglm` fitted objects}
\usage{
\method{vcov}{fastglm}(object, ...)
}
\arguments{
\item{object}{a fitted object of class inheriting from `"fastglm"`.}

\item{...}{additional arguments (currently unused).}
}
\value{
The estimated variance-covariance matrix of the fitted coefficients.
For rank-deficient fits, rows and columns corresponding to aliased
coefficients are filled with `NA`.
}
\description{
`vcov()` method for `fastglm` fitted objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin.R
\name{negbin}
\alias{negbin}
\title{Negative binomial family with known dispersion}
\usage{
negbin(theta, link = "log")
}
\arguments{
\item{theta}{the (known) dispersion parameter; must be positive.}

\item{link}{character; one of `"log"` (default), `"sqrt"`, or `"identity"`.}
}
\value{
A `family` object with class `"family"` and
  `family$family == "Negative Binomial(theta)"`. The object also carries
  the slot `family$theta`, which `family_code()` and the dispatch layer
  use to detect the native NB fast path.
}
\description{
A built-in negative-binomial family for use with [fastglm()] and
[fastglmPure()] when `theta` (the NB2 dispersion) is known. Equivalent to
`MASS::negative.binomial(theta, link)` but without taking a hard dependency
on the **MASS** package. The variance is `mu + mu^2 / theta`; as
`theta -> Inf` it reduces to Poisson.
}
\details{
For joint estimation of `theta` together with `beta`, use the dedicated
(forthcoming) `fastglm_nb()` entry point. The `negbin()` family here is
intended for use cases where `theta` has been pre-specified or estimated
separately.
}
\examples{
set.seed(1)
n <- 500
x  <- cbind(1, matrix(rnorm(n * 2), n, 2))
mu <- exp(x \%*\% c(0.3, 0.4, -0.2))
y  <- rpois(n, lambda = mu * rgamma(n, shape = 2, rate = 2))   # NB(theta=2)

fit <- fastglm(x, y, family = negbin(theta = 2, link = "log"))
coef(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastglm-methods.R
\name{predict.fastglm}
\alias{predict.fastglm}
\title{Obtains predictions and optionally estimates standard errors of those predictions from a fitted generalized linear model object.}
\usage{
\method{predict}{fastglm}(
  object,
  newdata = NULL,
  type = c("link", "response"),
  se.fit = FALSE,
  dispersion = NULL,
  ...
)
}
\arguments{
\item{object}{a fitted object of class inheriting from `"fastglm"`.}

\item{newdata}{a matrix to be used for prediction.}

\item{type}{the type of prediction required. The default is on the scale of the linear predictors;
the alternative "\code{response}" is on the scale of the response variable. Thus for a default binomial
model the default predictions are of log-odds (probabilities on logit scale) and \code{type = "response"}
 gives the predicted probabilities. The "\code{terms}" option returns a matrix giving the fitted values of each
 term in the model formula on the linear predictor scale.

The value of this argument can be abbreviated.}

\item{se.fit}{logical switch indicating if standard errors are required.}

\item{dispersion}{the dispersion of the GLM fit to be assumed in computing the standard errors.
If omitted, that returned by \code{summary} applied to the object is used.}

\item{...}{further arguments passed to or from other methods.}
}
\description{
Obtains predictions and optionally estimates standard errors of those predictions from a fitted generalized linear model object.
}
